'
' ####################
' #####  PROLOG  #####
' ####################
'
' A console program template
'
PROGRAM "TransLib"
VERSION "0.0001"
'
	IMPORT "xst"				' Standard library : required by most programs
'	IMPORT "xsx"				' Extended standard library
'	IMPORT "xio"				' Console input/ouput library

'	IMPORT "xst_s.lib"
' IMPORT "xsx_s.lib"
' IMPORT "xio_s.lib"

' IMPORT "gdi32"			' gdi32.dll
'	IMPORT "user32"		  ' user32.dll
	IMPORT "kernel32"	  ' kernel32.dll
'	IMPORT "shell32"		' shell32.dll
'	IMPORT "msvcrt"		  ' msvcrt.dll
	IMPORT "gmp"

'
EXPORT
' This structure stores the key constants.  These must be recomputed every time the precision is changed
TYPE CONSTANTS
	XLONG	.cBitsPrec
	XLONG	.cDigits
	XLONG	.pPi
	XLONG	.pPiBy2
	XLONG	.p3PiBy2
	XLONG	.p2Pi
	XLONG	.pRadToDeg		' 180/pi
	XLONG	.pRadToGrad		' 200/pi
	XLONG	.pE
	XLONG	.pLn2
	XLONG	.pLn10
	XLONG	.pSqrt2
	XLONG	.pInvSqrt2
END TYPE
' note that all functions return $$TRUE if undefined
DECLARE FUNCTION TransLib ()
DECLARE FUNCTION TransLib_InitConstants (CONSTANTS @constants)
DECLARE FUNCTION TransLib_ATan (rop, op)
DECLARE FUNCTION TransLib_ASin (rop, op)
DECLARE FUNCTION TransLib_ACos (rop, op)
DECLARE FUNCTION TransLib_ACot (rop, op)
DECLARE FUNCTION TransLib_ASec (rop, op)
DECLARE FUNCTION TransLib_ACosec (rop, op)
DECLARE FUNCTION TransLib_Ln (rop, op)
DECLARE FUNCTION TransLib_Power (rop, op1, op2)
DECLARE FUNCTION TransLib_Sin (rop, op)
DECLARE FUNCTION TransLib_Cos (rop, op)
DECLARE FUNCTION TransLib_Tan (rop, op)
DECLARE FUNCTION TransLib_Cot (rop, op)
DECLARE FUNCTION TransLib_Sec (rop, op)
DECLARE FUNCTION TransLib_Cosec (rop, op)
DECLARE FUNCTION TransLib_Sinh (rop, op)
DECLARE FUNCTION TransLib_Cosh (rop, op)
DECLARE FUNCTION TransLib_Tanh (rop, op)
DECLARE FUNCTION TransLib_Coth (rop, op)
DECLARE FUNCTION TransLib_Sech (rop, op)
DECLARE FUNCTION TransLib_Cosech (rop, op)
DECLARE FUNCTION TransLib_ASinh (rop, op)
DECLARE FUNCTION TransLib_ACosh (rop, op)
DECLARE FUNCTION TransLib_ATanh (rop, op)
DECLARE FUNCTION TransLib_ACoth (rop, op)
DECLARE FUNCTION TransLib_ASech (rop, op)
DECLARE FUNCTION TransLib_ACosech (rop, op)
DECLARE FUNCTION TransLib_IsEq (op1, op2)
DECLARE FUNCTION TransLib_Exp (rop, op)
END EXPORT
DECLARE FUNCTION I_ATan (rop, op)
DECLARE FUNCTION I_AGM (rop, op1, op2)
DECLARE FUNCTION I_Sin (rop, op)
DECLARE FUNCTION Delete_Float (pNumber)
DECLARE FUNCTION Allocate_Float ()
DECLARE FUNCTION formatFloat$ (pNumber)
'
'
' ######################
' #####  Entry ()  #####
' ######################
'
FUNCTION TransLib ()
	SHARED CONSTANTS const
  
	RETURN $$FALSE
END FUNCTION
'
' ####################################
' #####  TransLib_InitConstants  #####
' ####################################
' Initialises the constants
' constants = the constants structure, the client must allocate and deallocate the numbers
' returns $$TRUE on success or $$FALSE on fail
FUNCTION TransLib_InitConstants (CONSTANTS constants)
	SHARED CONSTANTS const
	mpf_t floats[]
	
	DIM floats[2]
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_init (&floats[i])
	NEXT
	t0 = &floats[0]
	t1 = &floats[1]
	maxError = &floats[2]
	
	const = constants
	
	' calculate pi
	__gmpf_set_si (t0, 1)
	__gmpf_div_ui (t0, t0, 2)
	TransLib_ATan (t0, t0)
	__gmpf_set_si (t1, 1)
	__gmpf_div_ui (t1, t1, 3)
	TransLib_ATan (t1, t1)
	__gmpf_add (t0, t0, t1)
	__gmpf_mul_ui (const.pPi, t0, 4)
	
	' calculate the constants derived from pi
	__gmpf_div_ui (const.pPiBy2, const.pPi, 2)
	__gmpf_mul_ui (const.p3PiBy2, const.pPiBy2, 3)
	__gmpf_mul_ui (const.p2Pi, const.pPi, 2)
	__gmpf_ui_div (const.pRadToDeg, 180, const.pPi)
	__gmpf_ui_div (const.pRadToGrad, 200, const.pPi)
	
	' calculate ln2
	__gmpf_set_si (maxError, 1)
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)
	__gmpf_set_si (const.pLn2, 3)		' The overall result
	__gmpf_div_ui (const.pLn2, const.pLn2, 4)
	__gmpf_set (t0, const.pLn2)		' The last term computed
	__gmpf_set (t1, const.pLn2)		' The absolute value of the last term computed, used for comparisons
	i = 1
	DO WHILE __gmpf_cmp (t1, maxError) >= 0
		__gmpf_mul_ui (t0, t0, i)
		__gmpf_neg (t0, t0)		' t1 = t1*-k
		__gmpf_div_ui (t0, t0, (8*i)+4)
		
		__gmpf_add (const.pLn2, const.pLn2, t0)
		__gmpf_abs (t1, t0)
		INC i
	LOOP
	
	' calculate e
	__gmpf_set_si (const.pE, 1)
	TransLib_Exp (const.pE, const.pE)
	
	' and ln10
	__gmpf_set_si (const.pLn10, 10)
	TransLib_Ln (const.pLn10, const.pLn10)

	' calculate sqr(2) and derived constants
	__gmpf_sqrt_ui (const.pSqrt2, 2)
	__gmpf_ui_div (const.pInvSqrt2, 1, const.pSqrt2)
	
	' Calculate the number of digits
	__gmpf_set_si (t0, 1)
	__gmpf_mul_2exp (t0, t0, const.cBitsPrec)
	TransLib_Ln (t0, t0)
	__gmpf_div (t0, t0, const.pLn10)
	const.cDigits = __gmpf_get_si (t0) + 10
	
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_clear (&floats[i])
	NEXT
END FUNCTION
'
' ####################
' #####  I_ATan  #####
' ####################
' For -1 < op < 1
FUNCTION I_ATan (rop, op)
	SHARED CONSTANTS const
	mpf_t floats[]
	
	DIM floats[11]
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_init (&floats[i])
	NEXT
	
	al1 = &floats[0]
	al2 = &floats[1]
	bl1 = &floats[2]
	bl2 = &floats[3]
	lastRop = &floats[4]
	t0 = &floats[5]
	t1 = &floats[6]
	t2 = &floats[7]
	a = &floats[8]
	A = &floats[9]
	B = &floats[10]
	maxError = &floats[11]

	__gmpf_set_si (maxError, 1)
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)

	' prime the pump
	__gmpf_set_si (al2, 0)
	__gmpf_set (al1, op)
	__gmpf_set_si (bl2, 1)
	__gmpf_set_si (bl1, 1)
	__gmpf_set_si (rop, 0)

	__gmpf_set (lastRop, op)
	i = 1
	b = 1
	
	__gmpf_sub (t2, rop, lastRop)
	__gmpf_abs (t2, t2)
	DO WHILE __gmpf_cmp (t2, maxError) >= 0
		__gmpf_set (lastRop, rop)
		' compute a
		__gmpf_mul_ui (a, op, i)	' a = x*i
		__gmpf_mul (a, a, a)		' a = a^2 ((x*i)^2)
		
		' compute b
		b = b+2
		
		' compute A
		__gmpf_mul_ui (t0, al1, b)
		__gmpf_mul (t1, al2, a)
		__gmpf_add (A, t0, t1)			' A = al1*b + al2*a
		'? "BPI4"
		__gmpf_set (al2, al1)
		__gmpf_set (al1, A)
		
		' compute B
		__gmpf_mul_ui (t0, bl1, b)
		__gmpf_mul (t1, bl2, a)
		__gmpf_add (B, t0, t1)			' A = bl1*b + bl2*a
		__gmpf_set (bl2, bl1)
		__gmpf_set (bl1, B)
		
		' compute result
		__gmpf_div (rop, A, B)
		__gmpf_sub (t2, rop, lastRop)
		__gmpf_abs (t2, t2)
		INC i
	LOOP
	
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_clear (&floats[i])
	NEXT
END FUNCTION
'
' ###########################
' #####  TransLib_ATan  #####
' ###########################
FUNCTION TransLib_ATan (rop, op)
	mpf_t rt0
	mpf_t rrx
	
	t0 = &rt0
	rx = &rrx
	
	__gmpf_init (t0)
	__gmpf_init_set (rx, op)
	
	IF (__gmpf_cmp_si (rx, -1) > 0) && (__gmpf_cmp_si (rx, 1) < 0) THEN
		I_ATan (rop, rx)
	ELSE
		__gmpf_mul (t0, rx, rx)
		__gmpf_add_ui (t0, t0, 1)
		__gmpf_sqrt (t0, t0)
		__gmpf_add_ui (t0, t0, 1)
		__gmpf_div (t0, rx, t0)
		I_ATan (rop, t0)
		__gmpf_mul_ui (rop, rop, 2)
		' 2ATan ( x/(1+Sqr(1+x^2)) )
	END IF
	
	__gmpf_clear (t0)
	__gmpf_clear (rx)
END FUNCTION
'
' ###########################
' #####  TransLib_ASin  #####
' ###########################
FUNCTION TransLib_ASin (rop, op)
	SHARED CONSTANTS const
	mpf_t rrx
	
	relNeg1 = __gmpf_cmp_si (op, -1) 
	rel1 = __gmpf_cmp_si (op, 1)
	SELECT CASE 0
		CASE relNeg1
			__gmpf_set (rop, const.pPiBy2)
			__gmpf_neg (rop, rop)
			RETURN
		CASE rel1
			__gmpf_set (rop, const.pPiBy2)
			RETURN
		CASE ELSE
			IF (relNeg1 <= 0)||(rel1 >= 0) THEN RETURN $$TRUE
	END SELECT
	
	
	rx = &rrx
	
	__gmpf_init_set (rx, op)
	
	__gmpf_mul (rop, rx, rx)
	__gmpf_ui_sub (rop, 1, rop)
	__gmpf_sqrt (rop, rop)
	__gmpf_div (rop, rx, rop)
	TransLib_ATan (rop, rop)
	' ATan (x/Sqr(1-x^2))
	
	__gmpf_clear (rx)
END FUNCTION
'
' ###########################
' #####  TransLib_ACos  #####
' ###########################
FUNCTION TransLib_ACos (rop, op)
	SHARED CONSTANTS const
	
	ret = TransLib_ASin (rop, op)
	__gmpf_sub (rop, const.pPiBy2, rop)
	' pi/2 - ASin (x)
	
	RETURN ret
END FUNCTION
'
' ###########################
' #####  TransLib_ACot  #####
' ###########################
FUNCTION TransLib_ACot (rop, op)
	SHARED CONSTANTS const
	
	ret = TransLib_ATan (rop, op)
	__gmpf_sub (rop, const.pPiBy2, rop)
	' pi/2 - ATan (x)
	RETURN ret
END FUNCTION
'
' ###########################
' #####  TransLib_ASec  #####
' ###########################
FUNCTION TransLib_ASec (rop, op)
	SHARED CONSTANTS const
	
	IF (__gmpf_cmp_si (op, -1) >= 0)&&(__gmpf_cmp_si (op, 1) <= 0) THEN RETURN $$TRUE
	__gmpf_ui_div (rop, 1, op)
	TransLib_ASin (rop, op)
	__gmpf_sub (rop, const.pPiBy2, rop)
	' pi/2 - ASin (1/x)
END FUNCTION
'
' #############################
' #####  TransLib_ACosec  #####
' #############################
FUNCTION TransLib_ACosec (rop, op)
	SHARED CONSTANTS const
	
	IF (__gmpf_cmp_si (op, -1) >= 0)&&(__gmpf_cmp_si (op, 1) <= 0) THEN RETURN $$TRUE
	__gmpf_ui_div (rop, 1, op)
	TransLib_ASin (rop, op)
	' ASin (1/x)
END FUNCTION
'
' ###################
' #####  I_Exp  #####
' ###################
' Computes e^x.  Used internally to compute other functions
FUNCTION TransLib_Exp (rop, op)
	SHARED CONSTANTS const
	mpf_t floats[]
	
	DIM floats[4]
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_init (&floats[i])
	NEXT
	m = &floats[0]
	rn = &floats[1]
	u = &floats[2]
	t0 = &floats[3]
	maxError = &floats[4]
	
	IF __gmpf_cmp_ui (op, 0) < 0 THEN
		__gmpf_neg (op, op)
		inv = $$TRUE
	ELSE
		inv = $$FALSE
	END IF
	
	' computer n
	__gmpf_div (rn, op, const.pLn2)
	__gmpf_floor (rn, rn)
	n = __gmpf_get_si (rn)
	
	' compute u
	__gmpf_mul (u, rn, const.pLn2)
	__gmpf_sub (u, op, u)
	
	' compute m
	__gmpf_set_si (maxError, 1)
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)
	__gmpf_set_si (m, 1)
	__gmpf_set_si (t0, 1)	' The previous term in the series
	i = 1
	DO WHILE __gmpf_cmp (t0, maxError) >= 0
		__gmpf_mul (t0, t0, u)
		__gmpf_div_ui (t0, t0, i)
		
		__gmpf_add (m, m, t0)
		INC i
	LOOP
	
	' put it all together
	__gmpf_mul_2exp (rop, m, n)
	
	IF inv THEN __gmpf_ui_div (rop, 1, rop)
	
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_clear (&floats[i])
	NEXT
END FUNCTION
'
' ###################
' #####  I_AGM  #####
' ###################
' Computes the arithmetic geometric mean of op1 and op2.  Used for computing logs
FUNCTION I_AGM (rop, op1, op2)
	SHARED CONSTANTS const
	mpf_t floats[]
	
	DIM floats[4]
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_init (&floats[i])
	NEXT
	a = &floats[0]
	g = &floats[1]
	diff = &floats[2]
	a1 = &floats[3]
	maxError = &floats[4]
	
	__gmpf_add (a, op1, op2)
	__gmpf_div_ui (a, a, 2)		' a = (x+y)/2
	__gmpf_mul (g, op1, op2)
	__gmpf_sqrt (g, g)				' g = Sqr(xy)
	
	__gmpf_set_si (maxError, 1)
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)
	__gmpf_sub (diff, a, g)
	__gmpf_abs (diff, diff)
	DO WHILE __gmpf_cmp (diff, maxError) >= 0
		__gmpf_add (a1, a, g)
		__gmpf_div_ui (a1, a1, 2)
		__gmpf_mul (g, a, g)
		__gmpf_sqrt (g, g)
		__gmpf_set (a, a1)
		
		__gmpf_sub (diff, a, g)
		__gmpf_abs (diff, diff)
	LOOP
	__gmpf_set (rop, a)
	
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_clear (&floats[i])
	NEXT
END FUNCTION
'
' #########################
' #####  TransLib_Ln  #####
' #########################
' Natural log
FUNCTION TransLib_Ln (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	mpf_t b

	IF __gmpf_cmp_si (op, 0) <= 0 THEN RETURN $$TRUE
	
	s = &a
	t0 = &b
	__gmpf_init (s)
	__gmpf_init (t0)
	minS = Allocate_Float ()
	
	__gmpf_set_ui (minS, 1)
	__gmpf_mul_2exp (minS, minS, const.cBitsPrec>>1)
	__gmpf_set (s, op)
	m = 0
	DO WHILE __gmpf_cmp (s, minS) <= 0
		__gmpf_mul_2exp (s, s, 1)
		INC m
	LOOP
	__gmpf_ui_div (s, 4, s)
	
	' compute first term, pi/2AGM(1, 4/s)
	__gmpf_set_ui (t0, 1)
	I_AGM (rop, t0, s)
	__gmpf_mul_ui (rop, rop, 2)
	__gmpf_div (rop, const.pPi, rop)
	
	' compute the second term, mln2
	__gmpf_mul_ui (t0, const.pLn2, m)
	
	' put it all together
	__gmpf_sub (rop, rop, t0)
	
	Delete_Float (minS)
	__gmpf_clear (s)
	__gmpf_clear (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Power  #####
' ############################
' computes op1^op2
FUNCTION TransLib_Power (rop, op1, op2)
	SHARED CONSTANTS const

	t0 = Allocate_Float ()
	t1 = Allocate_Float ()
	rx = Allocate_Float ()
	ry = Allocate_Float ()

	__gmpf_set (rx, op1)
	__gmpf_set (ry, op2)	
	' make sure this operation is defined
	order = __gmpf_cmp_ui (rx, 0)
	SELECT CASE TRUE
		CASE order = 0
			IF __gmpf_cmp_ui (ry, 0) = 0 THEN
				__gmpf_set_ui (rop, 1)
			ELSE
				__gmpf_set_ui (rop, 0)
			END IF
		CASE order < 0
			IF __gmpf_integer_p (ry) THEN
				' Compute the power
				__gmpf_neg (rx, rx)
				TransLib_Power (rop, rx, ry)
				' Negate the answer as required
				__gmpf_div_ui (t0, ry, 2)
				__gmpf_floor (t1, t0)
				IF __gmpf_cmp (t0, t1) != 0 THEN __gmpf_neg (rop, rop)
			ELSE
				GOSUB CleanUp
				RETURN $$TRUE
			END IF
		CASE ELSE
			IF __gmpf_integer_p (ry) && (__gmpf_cmp_ui (ry, 0) >= 0) THEN
				IF __gmpf_cmp_ui (rx, 2) = 0 THEN
					__gmpf_mul_2exp (rop, rx, __gmpf_get_ui (ry)-1)
				ELSE
					__gmpf_pow_ui (rop, rx, __gmpf_get_ui (ry))
				END IF
			ELSE
				__gmpf_set (t0, ry)
				SELECT CASE TRUE
					CASE __gmpf_cmp (rx, const.pE) = 0
						' do nothing
					CASE __gmpf_cmp_ui (rx, 2) = 0
						__gmpf_mul (t0, t0, const.pLn2)
					CASE __gmpf_cmp_ui (rx, 10) = 0
						__gmpf_mul (t0, t0, const.pLn10)
					CASE ELSE
						TransLib_Ln (t1, rx)
						__gmpf_mul (t0, t0, t1)
				END SELECT
				TransLib_Exp (rop, t0)
			END IF
	END SELECT
	
	GOSUB CleanUp
	RETURN $$FALSE
	
	SUB CleanUp
		Delete_Float (t0)
		Delete_Float (t1)
		Delete_Float (rx)
		Delete_Float (ry)
	END SUB
END FUNCTION
'
' ###################
' #####  I_Sin  #####
' ###################
' Calculates the sine of an angle between 0 and pi/2
FUNCTION I_Sin (rop, op)
	SHARED CONSTANTS const
	mpf_t floats[]
	
	DIM floats[3]
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_init (&floats[i])
	NEXT
	t0 = &floats[0]
	t1 = &floats[1]
	opSqaured = &floats[2]
	maxError = &floats[3]
	
	__gmpf_set_si (maxError, 1)
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)
	__gmpf_mul (opSqaured, op, op)
	__gmpf_set (rop, op)	' the overall result
	__gmpf_set (t0, rop)	' the previous term
	__gmpf_abs (t1, t0)		' the absolute value of the last term
	i = 2
	DO WHILE __gmpf_cmp (t1, maxError) >= 0
		__gmpf_mul (t0, t0, opSqaured)
		__gmpf_div_ui (t0, t0, i)
		INC i
		__gmpf_div_ui (t0, t0, i)
		
		__gmpf_neg (t0, t0)
		__gmpf_add (rop, rop, t0)
		
		__gmpf_abs (t1, t0)
		INC i
	LOOP
	
	FOR i = 0 TO UBOUND(floats[])
		__gmpf_clear (&floats[i])
	NEXT
END FUNCTION
'
' ##########################
' #####  TransLib_Sin  #####
' ##########################
' Computes the sine of any number
FUNCTION TransLib_Sin (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	' Use periodicity to simplify to angle between 0 and 2pi
	__gmpf_set (t0, op)
	IF __gmpf_cmp_ui (t0, 0) < 0 THEN
		DO WHILE __gmpf_cmp_ui (t0, 0) < 0
			__gmpf_add (t0, t0, const.p2Pi)
		LOOP
	ELSE
		DO WHILE __gmpf_cmp (t0, const.p2Pi) >= 0
			__gmpf_sub (t0, t0, const.p2Pi)
		LOOP
	END IF
	
	' now find out which quadrant we are dealing with
	SELECT CASE TRUE
		CASE __gmpf_cmp (t0, const.pPiBy2) < 0
			I_Sin (rop, t0)
		CASE __gmpf_cmp (t0, const.pPi) < 0
			__gmpf_sub (t0, const.pPi, t0)
			I_Sin (rop, t0)
		CASE __gmpf_cmp (t0, const.p3PiBy2) < 0
			__gmpf_sub (t0, t0, const.pPi)
			I_Sin (rop, t0)
			__gmpf_neg (rop, rop)
		CASE ELSE
			__gmpf_sub (t0, const.p2Pi, t0)
			I_Sin (rop, t0)
			__gmpf_neg (rop, rop)
	END SELECT
	
	__gmpf_clear (t0)
END FUNCTION
'
' ##########################
' #####  TransLib_Cos  #####
' ##########################
FUNCTION TransLib_Cos (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	__gmpf_sub (t0, const.pPiBy2, op)
	TransLib_Sin (rop, t0)
	
	__gmpf_clear (t0)
END FUNCTION
'
' ##########################
' #####  TransLib_Tan  #####
' ##########################
FUNCTION TransLib_Tan (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	__gmpf_set (t0, op)
	TransLib_Cos (rop, t0)
	IF __gmpf_cmp_ui (rop, 0) = 0 THEN
		__gmpf_clear (t0)
		RETURN $$TRUE
	END IF
	TransLib_Sin (t0, t0)
	__gmpf_div (rop, t0, rop)
	
	__gmpf_clear (t0)
END FUNCTION
'
' ##########################
' #####  TransLib_Cot  #####
' ##########################
FUNCTION TransLib_Cot (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	__gmpf_set (t0, op)
	TransLib_Sin (rop, t0)
	IF __gmpf_cmp_ui (rop, 0) = 0 THEN
		__gmpf_clear (t0)
		RETURN $$TRUE
	END IF
	TransLib_Cos (t0, t0)
	__gmpf_div (rop, t0, rop)
	
	__gmpf_clear (t0)
END FUNCTION
'
' ##########################
' #####  TransLib_Sec  #####
' ##########################
FUNCTION TransLib_Sec (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	__gmpf_set (t0, op)
	TransLib_Cos (rop, t0)
	IF __gmpf_cmp_ui (rop, 0) = 0 THEN
		__gmpf_clear (t0)
		RETURN $$TRUE
	END IF
	__gmpf_ui_div (rop, 1, rop)
	
	__gmpf_clear (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Cosec  #####
' ############################
FUNCTION TransLib_Cosec (rop, op)
	SHARED CONSTANTS const
	mpf_t a
	
	t0 = &a
	__gmpf_init (t0)
	
	__gmpf_set (t0, op)
	TransLib_Sin (rop, t0)
	IF __gmpf_cmp_ui (rop, 0) = 0 THEN
		__gmpf_clear (t0)
		RETURN $$TRUE
	END IF
	__gmpf_ui_div (rop, 1, rop)
	
	__gmpf_clear (t0)
END FUNCTION
'
' ###########################
' #####  TransLib_Sinh  #####
' ###########################
FUNCTION TransLib_Sinh (rop, op)
	t0 = Allocate_Float ()
	
	TransLib_Exp (t0, op)
	__gmpf_ui_div (rop, 1, t0)
	__gmpf_sub (t0, t0, rop)
	__gmpf_div_ui (rop, t0, 2)
	
	Delete_Float (t0)
END FUNCTION
'
' ###########################
' #####  TransLib_Cosh  #####
' ###########################
FUNCTION TransLib_Cosh (rop, op)
	t0 = Allocate_Float ()
	
	TransLib_Exp (t0, op)
	__gmpf_ui_div (rop, 1, t0)
	__gmpf_add (t0, t0, rop)
	__gmpf_div_ui (rop, t0, 2)
	
	Delete_Float (t0)
END FUNCTION
'
' ###########################
' #####  TransLib_Tanh  #####
' ###########################
FUNCTION TransLib_Tanh (rop, op)
	t0 = Allocate_Float ()
	
	__gmpf_mul_2exp (t0, op, 1)
	TransLib_Exp (t0, t0)
	__gmpf_sub_ui (rop, t0, 1)
	__gmpf_add_ui (t0, t0, 1)
	__gmpf_div (rop, rop, t0)
	
	Delete_Float (t0)
END FUNCTION
'
' ###########################
' #####  TransLib_Coth  #####
' ###########################
FUNCTION TransLib_Coth (rop, op)
	t0 = Allocate_Float ()
	
	IF __gmpf_cmp_si (op, 0) = 0 THEN RETURN $$TRUE
	
	__gmpf_mul_2exp (t0, op, 1)
	TransLib_Exp (t0, t0)
	__gmpf_add_ui (rop, t0, 1)
	__gmpf_sub_ui (t0, t0, 1)
	__gmpf_div (rop, rop, t0)
	
	Delete_Float (t0)
END FUNCTION
'
' ###########################
' #####  TransLib_Sech  #####
' ###########################
FUNCTION TransLib_Sech (rop, op)
	t0 = Allocate_Float ()
	
	TransLib_Exp (t0, op)
	__gmpf_ui_div (rop, 1, t0)
	__gmpf_add (t0, t0, rop)
	__gmpf_ui_div (rop, 2, t0)
	
	Delete_Float (t0)
END FUNCTION
'
' #############################
' #####  TransLib_Cosech  #####
' #############################
FUNCTION TransLib_Cosech (rop, op)
	t0 = Allocate_Float ()
	
	IF __gmpf_cmp_si (op, 0) = 0 THEN RETURN $$TRUE
	
	TransLib_Exp (t0, op)
	__gmpf_ui_div (rop, 1, t0)
	__gmpf_sub (t0, t0, rop)
	__gmpf_ui_div (rop, 2, t0)
	
	Delete_Float (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Asinh  #####
' ############################
FUNCTION TransLib_ASinh (rop, op)
	t0 = Allocate_Float ()
	
	__gmpf_set (t0, op)
	__gmpf_mul (rop, t0, t0)
	__gmpf_add_ui (rop, rop, 1)
	__gmpf_sqrt (rop, rop)
	__gmpf_add (rop, rop, t0)
	TransLib_Ln (rop, rop)
	
	Delete_Float (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Acosh  #####
' ############################
FUNCTION TransLib_ACosh (rop, op)
	t0 = Allocate_Float ()
	
	IF __gmpf_cmp_si (op, 1) < 0 THEN RETURN $$TRUE
	
	__gmpf_set (t0, op)
	__gmpf_mul (rop, t0, t0)
	__gmpf_sub_ui (rop, rop, 1)
	__gmpf_sqrt (rop, rop)
	__gmpf_add (rop, rop, t0)
	TransLib_Ln (rop, rop)
	
	Delete_Float (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Atanh  #####
' ############################
FUNCTION TransLib_ATanh (rop, op)
	t0 = Allocate_Float ()
	IF (__gmpf_cmp_si (op, -1) <= 0)||(__gmpf_cmp_si (op, 1) >= 0) THEN RETURN $$TRUE

	__gmpf_add_ui (rop, op, 1)
	__gmpf_ui_sub (t0, 1, op)
	__gmpf_div (rop, rop, t0)
	TransLib_Ln (rop, rop)
	__gmpf_div_2exp (rop, rop, 1)
	
	Delete_Float (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Acoth  #####
' ############################
FUNCTION TransLib_ACoth (rop, op)
	t0 = Allocate_Float ()
	IF (__gmpf_cmp_si (op, -1) >= 0)&&(__gmpf_cmp_si (op, 1) <= 0) THEN RETURN $$TRUE

	__gmpf_add_ui (rop, op, 1)
	__gmpf_sub_ui (t0, op, 1)
	__gmpf_div (rop, rop, t0)
	TransLib_Ln (rop, rop)
	__gmpf_div_2exp (rop, rop, 1)
	
	Delete_Float (t0)
END FUNCTION
'
' ############################
' #####  TransLib_Asech  #####
' ############################
FUNCTION TransLib_ASech (rop, op)
	t0 = Allocate_Float ()
	
	IF (__gmpf_cmp_si (op, 0) <= 0)||(__gmpf_cmp_si (op, 1) > 0) THEN RETURN $$TRUE
	
	__gmpf_ui_div (t0, 1, op)
	__gmpf_set (t0, op)
	__gmpf_mul (rop, t0, t0)
	__gmpf_sub_ui (rop, rop, 1)
	__gmpf_sqrt (rop, rop)
	__gmpf_add (rop, rop, t0)
	TransLib_Ln (rop, rop)
	
	Delete_Float (t0)
END FUNCTION
'
' ##############################
' #####  TransLib_Acosech  #####
' ##############################
FUNCTION TransLib_ACosech (rop, op)
	t0 = Allocate_Float ()
	
	IF __gmpf_cmp_si (op, 0) = 0  THEN RETURN $$TRUE
	
	__gmpf_ui_div (t0, 1, op)
	__gmpf_mul (rop, t0, t0)
	__gmpf_add_ui (rop, rop, 1)
	__gmpf_sqrt (rop, rop)
	__gmpf_add (rop, rop, t0)
	TransLib_Ln (rop, rop)
	
	Delete_Float (t0)
END FUNCTION
'
' ##########################
' #####  Delete_Float  #####
' ##########################
' Frees a floating point number correctly
' pNumber = the number to delete
' returns $$TRUE on success or $$FALSE on fail
FUNCTION Delete_Float (pNumber)
	__gmpf_clear (pNumber)
	GlobalFree (pNumber)
	RETURN $$TRUE
END FUNCTION
'
' ############################
' #####  Allocate_Float  #####
' ############################
' Allocates a new floating point number
' returns a pointer to the number
FUNCTION Allocate_Float ()
	ret = GlobalAlloc ($$GMEM_FIXED, SIZE(mpf_t))
	__gmpf_init (ret)
	RETURN ret
END FUNCTION
'
' ##########################
' #####  formatFloat$  #####
' ##########################
' For debugging mainly
FUNCTION formatFloat$ (pNumber)
	stri$ = CSTRING$(__gmpf_get_str (0, &exp, 10, 0, pNumber))
	IF MID$(stri$, 1, 1) = "-" THEN
		stri$ = LCLIP$(stri$)
		neg = $$TRUE
	ELSE
		neg = $$FALSE
	END IF
	
	IF UBOUND(stri$) > 77 THEN
		i = UBOUND(stri$)
		IF (stri${i}-'0') > 5 THEN
			stri$ = RCLIP$(stri$)
			DEC i
			DO WHILE i > -1
				IF (stri${i}-'0') < 9 THEN
					UBYTEAT(&stri$,i) = stri${i}+1
					EXIT DO
				ELSE
					stri$ = RCLIP$(stri$)
					DEC i
				END IF
			LOOP
		END IF
	END IF
	
	out$ = ""
	IF UBOUND(stri$) < 0 && exp = 0 THEN
		out$ = "0"
	ELSE
		IF exp < 1 THEN out$ = out$+"0"		' Write a leading 0 if we need one
		IF exp < 0 THEN										' Write leading 0's after the point
			out$ = out$+"."
			FOR i = 0 TO (ABS(exp)-1)
				out$ = out$+"0"
			NEXT
			fracDigit = 0		' The 0 based index of the first fractional digit
		ELSE				' Write out the non fractional part
			FOR i = 0 TO (exp-1)
				IF i > UBOUND(stri$) THEN out$ = out$+"0" ELSE out$ = out$+CHR$(stri${i})
			NEXT
			fracDigit = exp
			IF fracDigit <= UBOUND(stri$) THEN out$ = out$+"."
		END IF
		FOR i = fracDigit TO UBOUND(stri$)
			out$ = out$+CHR$(stri${i})
		NEXT
	END IF
	
	IF neg THEN RETURN "-"+out$ ELSE RETURN out$
END FUNCTION
'
' ###########################
' #####  TransLib_IsEq  #####
' ###########################
' Check for equality
FUNCTION TransLib_IsEq (op1, op2)
	SHARED CONSTANTS const
	maxError = Allocate_Float ()
	t0 = Allocate_Float ()
	
	__gmpf_set_si (maxError, 1)
	__gmpf_get_d_2exp (&exp1, op1)
	__gmpf_get_d_2exp (&exp2, op2)
	IF exp1 > exp2 THEN
		__gmpf_mul_2exp (maxError, maxError, exp1)
	ELSE
		__gmpf_mul_2exp (maxError, maxError, exp2)
	END IF
	__gmpf_div_2exp (maxError, maxError, const.cBitsPrec)
	
	__gmpf_sub (t0, op1, op2)
	__gmpf_abs (t0, t0)
	
	IF __gmpf_cmp (t0, maxError) <= 0 THEN ret = $$TRUE ELSE ret = $$FALSE
	Delete_Float (maxError)
	Delete_Float (t0)
	
	RETURN ret
END FUNCTION
END PROGRAM