'
' ####################
' #####  PROLOG  #####
' ####################
'
' An advanced expression evaluation library
'
PROGRAM "calculate"
VERSION "0.0001"
'
	IMPORT "xst"				' Standard library : required by most programs
	IMPORT "xsx_s"				' Extended standard library
	IMPORT "xma_s"
'	IMPORT "xio"				' Console input/ouput library

'	IMPORT "xst_s.lib"
' IMPORT "xsx_s.lib"
' IMPORT "xio_s.lib"

	IMPORT "gdi32"			' gdi32.dll
	IMPORT "user32"		  ' user32.dll
	IMPORT "kernel32"	  ' kernel32.dll
'	IMPORT "shell32"		' shell32.dll
'	IMPORT "msvcrt"		  ' msvcrt.dll
	IMPORT "gmp"
	IMPORT "TransLib"
	IMPORT "ADT"
	
m4_include(`accessors.m4')

EXPORT
' Rouding modes
$$ROUND_NONE				= 0
$$ROUND_FIXED				= 1
$$ROUND_SCIENTIFIC	= 2
' Angle modes
$$ANGLE_RADIANS			= 0
$$ANGLE_DEGREES			= 1
$$ANGLE_GRADS				= 2
' Out modes
$$OUT_DEC						= 0
$$OUT_HEX						= 1
$$OUT_OCT						= 2
$$OUT_BIN						= 3
END EXPORT

TYPE CALCSTATE
	XLONG	.maxDigits
	XLONG	.lastAnswer
	XLONG	.roundMode
	XLONG	.roundParameter
	XLONG	.angleMode
	XLONG	.outMode
	XLONG	.debug
	XLONG	.iLastAns
	gmp_randstate_t	.randState
END TYPE

' Operators
$$OP_INVALID		= 000
$$OP_LITERAL		= 001
$$OP_VAR				= 002
$$OP_PI					= 003
$$OP_E					= 004
$$OP_I					= 005
$$OP_RAND				= 006
$$OP_FUNCCALL		= 007
$$OP_MAKEVECTOR	= 008
$$OP_PARAMETER	= 009
$$OP_BOUNDPARAM	= 010
$$OP_FUNCLIT		= 011
$$OP_RECURSE		= 012
$$OP_LASTANS		= 013

$$OP_ADD				= 101
$$OP_SUB				= 102
$$OP_OR					= 103
$$OP_XOR				= 104

$$OP_MUL				= 201
$$OP_DIV				= 202
$$OP_UPSHIFT		= 203
$$OP_DOWNSHIFT	= 204
$$OP_AND				= 205
$$OP_C					= 206
$$OP_P					= 207
$$OP_DOTPROD		= 208

$$OP_RAISE			= 301
$$OP_ROOT				= 302

$$OP_NEGATE			= 401
$$OP_FACT				= 402
$$OP_NOT				= 403
$$OP_SQR				= 404
$$OP_CBR				= 405
$$OP_SIN				= 406
$$OP_COS				= 407
$$OP_TAN				= 408
$$OP_COT				= 409
$$OP_SEC				= 410
$$OP_COSEC			= 411
$$OP_ASIN				= 412
$$OP_ACOS				= 413
$$OP_ATAN				= 414
$$OP_ACOT				= 415
$$OP_ASEC				= 416
$$OP_ACOSEC			= 417
$$OP_SINH				= 418
$$OP_COSH				= 419
$$OP_TANH				= 420
$$OP_COTH				= 421
$$OP_SECH				= 422
$$OP_COSECH			= 423
$$OP_ASINH			= 424
$$OP_ACOSH			= 425
$$OP_ATANH			= 426
$$OP_ACOTH			= 427
$$OP_ASECH			= 428
$$OP_ACOSECH		= 429
$$OP_INT				= 430
$$OP_FRAC				= 431
$$OP_LOGE				= 432
$$OP_EXP				= 433
$$OP_RE					= 434
$$OP_IM					= 435
$$OP_CONJ				= 436

$$OP_LOGAND			= 501
$$OP_LOGOR			= 502
$$OP_RELEQ			= 503
$$OP_RELGT			= 504
$$OP_RELLT			= 505
$$OP_RELGTE			= 506
$$OP_RELLTE			= 507
$$OP_RELNOTEQ		= 508
$$OP_SHORTC_T		= 509		' The short curcuit markers are used to implement short circuit evaluation
$$OP_SHORTC_F		= 510		' They check to see if they match the stack, and if they do they stop evaluation

$$CMD_STOR			= 601
$$CMD_CLEAR			= 602

$$OP_VECTORDIM	= 713
$$OP_GENVEC			= 714
$$OP_VECCOM			= 715
$$OP_CONCAT			= 716

TYPE TOKEN
	XLONG	.op
	XLONG	.value
	XLONG	.value2		' Used for some ops which require 2 values
END TYPE

TYPE VECTOR
	XLONG				.size			' The number of components
	XLONG				.type			' The type of the components, just a rough guide really
	LINKEDLIST	.values		' The actual values for the vector
END TYPE

TYPE PARAMETER
XLONG				.type
	XLONG				.name
END TYPE

TYPE FUNCCASE
	LINKEDLIST	.condExpr
	LINKEDLIST	.realExpr
END TYPE

TYPE FUNC
	XLONG				.definition
	XLONG				.isClosure
	LINKEDLIST	.paramTypes
	LINKEDLIST	.cases
	LINKEDLIST	.caseElse
END TYPE

TYPE COMPLEX
	XLONG				.real
	XLONG				.imaginary
END TYPE

$$TYPE_INVALID	= 0
$$TYPE_INT			= 1		' We never have a symbol of this type, it's just a restriction for function output
$$TYPE_NUMBER		= 2
$$TYPE_COMPLEX	= 3
$$TYPE_VECTOR		= 4
$$TYPE_FUNCTION	= 5
$$TYPE_BOOL			= 6
TYPE SYMBOL
	XLONG			.type	' the type of the symbol
	XLONG			.data
	FUNCADDR	.delete (XLONG)	' The function to delete the data associated with this symbol
END TYPE

EXPORT
' Exceptions
$$EXCEPTION_NONE = 0
$$EXCEPTION_UNDEFINED	= 1
$$EXCEPTION_PINF	= 2
$$EXCEPTION_NINF	= 3
$$EXCEPTION_INTERNAL = 4
$$EXCEPTION_TYPES = 5
$$EXCEPTION_NOTNUM = 6
$$EXCEPTION_PARAMETERS = 7

DECLARE FUNCTION calculate_init ()
DECLARE FUNCTION calculate_parse (expression$, @errorString$)
DECLARE FUNCTION calculate_execute (id, result, @exception, @resultString$)
DECLARE FUNCTION calculate_setPrecision (bits)
DECLARE FUNCTION calculate_setRounding (roundMode, parameter)
DECLARE FUNCTION calculate_setAngle (angleMode)
DECLARE FUNCTION calculate_setOutMode (outMode)
DECLARE FUNCTION calculate_clearExpression (id)
DECLARE FUNCTION calculate_getSymbols$ ()
DECLARE FUNCTION calculate_getPrecision ()
DECLARE FUNCTION calculate_getRounding (@roundParam)
DECLARE FUNCTION calculate_getOutMode ()
DECLARE FUNCTION calculate_getAngle ()
DECLARE FUNCTION calculate_getDependantVars (iExpr, @vars[])
DECLARE FUNCTION calculate_getVarID (var$)

END EXPORT
DECLARE FUNCTION ReadToken$ ()
DECLARE FUNCTION PeekToken$ ()
DECLARE FUNCTION isNum (char)
DECLARE FUNCTION isHexit (char)
DECLARE FUNCTION isAlpha (char)
DECLARE FUNCTION isAlphaNum (char)
DECLARE FUNCTION isWhitespace (char)
DECLARE FUNCTION AddError (error$)
DECLARE FUNCTION parse_line (LINKEDLIST @out)
DECLARE FUNCTION parse_command (LINKEDLIST @out)
DECLARE FUNCTION parse_expression (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_nexpr (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_product (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_factor (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_term (LINKEDLIST @out, @exprType)
DECLARE FUNCTION FormatFloat$ (pNumber)
DECLARE FUNCTION Deparse$ (LINKEDLIST expression)
DECLARE FUNCTION TakeToken ()
DECLARE FUNCTION SymbolTable_Init ()
DECLARE FUNCTION SymbolTable_Insert (name$, SYMBOL symbol)
DECLARE FUNCTION SymbolTable_Retreive (name$, SYMBOL @symbol)
DECLARE FUNCTION SymbolTable_Update (name$, SYMBOL symbol)
DECLARE FUNCTION SymbolTable_Delete (name$)
DECLARE FUNCTION SymbolTable_DeleteAll ()
DECLARE FUNCTION isVarNameValid (var$)
DECLARE FUNCTION evaluate (LINKEDLIST expression, @exception, @iData, @type, iThisFunc, iCase)
DECLARE FUNCTION eval_push (type, iData)
DECLARE FUNCTION eval_pop (@type, @iData, FUNCADDR @Delete)
DECLARE FUNCTION eval_peek (@type, @iData)
DECLARE FUNCTION Delete_Float (pNumber)
DECLARE FUNCTION Allocate_Float ()
DECLARE FUNCTION op_add (iData1, type1, iData2, type2)
DECLARE FUNCTION op_sub (iData1, type1, iData2, type2)
DECLARE FUNCTION op_mul (iData1, type1, iData2, type2)
DECLARE FUNCTION op_div (iData1, type1, iData2, type2)
DECLARE FUNCTION op_or (iData1, type1, iData2, type2)
DECLARE FUNCTION op_xor (iData1, type1, iData2, type2)
DECLARE FUNCTION op_upshift (iData1, type1, iData2, type2)
DECLARE FUNCTION op_downshift (iData1, type1, iData2, type2)
DECLARE FUNCTION op_and (iData1, type1, iData2, type2)
DECLARE FUNCTION op_c (iData1, type1, iData2, type2)
DECLARE FUNCTION op_p (iData1, type1, iData2, type2)
DECLARE FUNCTION op_raise (iData1, type1, iData2, type2)
DECLARE FUNCTION op_root (iData1, type1, iData2, type2)
DECLARE FUNCTION op_negate (iData1, type1)
DECLARE FUNCTION op_fact (iData1, type1)
DECLARE FUNCTION op_not (iData1, type1)
DECLARE FUNCTION op_sqr (iData1, type1)
DECLARE FUNCTION op_cbr (iData1, type1)
DECLARE FUNCTION op_sin (iData1, type1)
DECLARE FUNCTION op_cos (iData1, type1)
DECLARE FUNCTION op_tan (iData1, type1)
DECLARE FUNCTION op_cot (iData1, type1)
DECLARE FUNCTION op_sec (iData1, type1)
DECLARE FUNCTION op_cosec (iData1, type1)
DECLARE FUNCTION op_asin (iData1, type1)
DECLARE FUNCTION op_acos (iData1, type1)
DECLARE FUNCTION op_atan (iData1, type1)
DECLARE FUNCTION op_acot (iData1, type1)
DECLARE FUNCTION op_asec (iData1, type1)
DECLARE FUNCTION op_acosec (iData1, type1)
DECLARE FUNCTION op_sinh (iData1, type1)
DECLARE FUNCTION op_cosh (iData1, type1)
DECLARE FUNCTION op_tanh (iData1, type1)
DECLARE FUNCTION op_coth (iData1, type1)
DECLARE FUNCTION op_sech (iData1, type1)
DECLARE FUNCTION op_cosech (iData1, type1)
DECLARE FUNCTION op_asinh (iData1, type1)
DECLARE FUNCTION op_acosh (iData1, type1)
DECLARE FUNCTION op_atanh (iData1, type1)
DECLARE FUNCTION op_acoth (iData1, type1)
DECLARE FUNCTION op_asech (iData1, type1)
DECLARE FUNCTION op_acosech (iData1, type1)
DECLARE FUNCTION op_int (iData1, type1)
DECLARE FUNCTION op_frac (iData1, type1)
DECLARE FUNCTION op_ln (iData1, type1)
DECLARE FUNCTION Delete_Complex (iData)
DECLARE FUNCTION CopySymbol (iData, type)
DECLARE FUNCTION parse_vexpr (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_vterm (LINKEDLIST @out, @exprType)
DECLARE FUNCTION Delete_Vector (iData)
DECLARE FUNCTION op_dotprod (iData1, type1, iData2, type2)
DECLARE FUNCTION parse_fLiteral (@out)
DECLARE FUNCTION parse_condition (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_condTerm (LINKEDLIST @out, @exprType)
DECLARE FUNCTION parse_fexpr (LINKEDLIST @out, @exprType)
DECLARE FUNCTION SearchCompileTimeParamStack (name$, @index, @level, @type)
DECLARE FUNCTION Delete_Function (iFunc)
DECLARE FUNCTION Delete_Expression (LINKEDLIST list)
DECLARE FUNCTION op_closure (iFunc, level)
DECLARE FUNCTION op_eq (iData1, type1, iData2, type2, @exception)
DECLARE FUNCTION op_gt (iData1, type1, iData2, type2, @exception)
DECLARE FUNCTION op_lt (iData1, type1, iData2, type2, @exception)
DECLARE FUNCTION op_exp (iData1, type1)
DECLARE FUNCTION op_re (iData1, type1)
DECLARE FUNCTION op_im (iData1, type1)
DECLARE FUNCTION op_conj (iData1, type1)
DECLARE FUNCTION Delete_RuntimeParams (level)
DECLARE FUNCTION GetFuncCase (FUNC func, LINKEDLIST @iExpr, @iCase)
DECLARE FUNCTION FormatSymbol$ (iData, type)
DECLARE FUNCTION ConvertAngle (in, iAngle)
DECLARE FUNCTION StringToFloat (stri$, float)
DECLARE FUNCTION op_veccom (iData1, type1, iData2, type2)
DECLARE FUNCTION calculate_fastExecute (iExpr)
DECLARE FUNCTION calculate_getVector (@vec[])
DECLARE FUNCTION calculate_getReal (pMpf)
DECLARE FUNCTION op_concat (iData1, type1, iData2, type2)
DECLARE FUNCTION spaces$ (x)

DECLARE CFUNCTION gmp_malloc (aSize)
DECLARE CFUNCTION gmp_realloc (ptr, oldSize, newSize)
DECLARE CFUNCTION VOID gmp_free (ptr, aSize)

DeclareAccess(LINKEDLIST)
DeclareAccess(TOKEN)
DeclareAccess(VECTOR)
DeclareAccess(FUNC)
DeclareAccess(COMPLEX)
DeclareAccess(SYMBOL)
DeclareAccess(PARAMETER)
DeclareAccess(FUNCCASE)
'
'
' ###############################
' #####  calculate_init ()  #####
' ###############################
' returns $$TRUE if there is an error otherwise $$FALSE
FUNCTION calculate_init ()
	SHARED CALCSTATE calcState
	SHARED STACK evalStack
	SHARED CONSTANTS const
	SHARED LINKEDLIST	compileParamStack
	SHARED LINKEDLIST	runtimeParamStack
	STATIC init
	
	SYMBOL symbol
	
	IF init THEN RETURN $$TRUE
	
	__gmp_set_memory_functions (&gmp_malloc (), &gmp_realloc (), &gmp_free ())
	
	IF ADT () THEN RETURN $$TRUE
	
	LINKEDLIST_Init()
	TOKEN_Init()
	VECTOR_Init()
	FUNC_Init()
	COMPLEX_Init ()
	SYMBOL_Init()
	PARAMETER_Init()
	FUNCCASE_Init()
	
	SymbolTable_Init ()
	Stack_Init (@evalStack)
	
	LinkedList_Init (@compileParamStack)
	LinkedList_Init (@runtimeParamStack)
	
	calculate_setPrecision (128)
	
	__gmp_randinit_default (&calcState.randState)
	
	symbol.type = $$TYPE_NUMBER
	symbol.data = Allocate_Float ()
	__gmpf_set_ui (symbol.data, 0)
	symbol.delete = &Delete_Float ()
	calcState.iLastAns = SYMBOL_New (symbol)
	
	init = $$TRUE
  RETURN $$FALSE
END FUNCTION
'
' #############################
' #####  calculate_parse  #####
' #############################
' Parses an expression
' expression$ = the expression to parse
' errors$ = the string variable to store the errors
' returns the id of the parsed expression or 0 on errors
FUNCTION calculate_parse (expression$, errorString$)
	SHARED CALCSTATE calcState
	SHARED in$
	SHARED pos
	SHARED nextToken
	SHARED errors$
	
	LINKEDLIST list
	
	pos = 0
	nextToken = 0
	errors$ = ""
	errorString$ = ""
	
	' Let's do a little preprocessing
	' First - strip any comments
	in$ = ""
	XstParseStringToStringArray (expression$, "\n", @s$[])
	FOR i = 0 TO UBOUND(s$[])
		in$ = in$+MID$(s$[i], 1, INSTR(s$[i], "#")-1)
	NEXT
	
	IF TRIM$(in$) = "" THEN RETURN 0
	
	' Now insert a let command if it is implied
	i = 0
	IF i <= UBOUND(in$) THEN
		IF isAlpha (in${i}) THEN
			DO WHILE (i <= UBOUND(in$)) && (isAlphaNum (in${i})) : INC i :LOOP
			DO WHILE (i <= UBOUND(in$)) && (isWhitespace (in${i})) : INC i :LOOP
			IF i <= UBOUND(in$) THEN
				IF in${i} = '=' THEN
					in$ = "LET "+in$
				END IF
			END IF
		END IF
	END IF
	
	' And insert a last ans if that is implied
	IF in$ THEN 
		SELECT CASE in${0}
			CASE '+','-','*','/'
				in$ = "@ "+in$
		END SELECT
	END IF
	
	' Is this a set command
	token$ = PeekToken$ ()
	SELECT CASE UCASE$(token$)
		CASE "SET"
			TakeToken ()
			setting$ = ReadToken$ ()
			IF ReadToken$ () != "=" THEN
				AddError ("Expected = in SET command")
				INC cErrors
			ELSE
				value$ = ReadToken$ ()
				SELECT CASE UCASE$(setting$)
					CASE "ANGLEMODE", "ANGLE", "ANGLES", "A"
						SELECT CASE UCASE$(value$)
							CASE "RADIANS", "RAD", "RADS", "R"
								calculate_setAngle ($$ANGLE_RADIANS)
							CASE "DEGREES", "DEG", "DEGS", "D"
								calculate_setAngle ($$ANGLE_DEGREES)
							CASE "GRADS", "GRAD", "G"
								calculate_setAngle ($$ANGLE_GRADS)
							CASE ELSE
								GOSUB ValueError
						END SELECT
					CASE "OUTMODE", "OUT", "O"
						SELECT CASE UCASE$(value$)
							CASE "BIN", "2", "B"
								calculate_setOutMode ($$OUT_BIN)
							CASE "OCT", "8", "O"
								calculate_setOutMode ($$OUT_OCT)
							CASE "DEC", "10", "D"
								calculate_setOutMode ($$OUT_DEC)
							CASE "HEX", "16", "H"
								calculate_setOutMode ($$OUT_HEX)
							CASE ELSE
								GOSUB ValueError
						END SELECT
					CASE "ROUNDMODE", "ROUNDING", "ROUND", "R"
						SELECT CASE UCASE$(value$)
							CASE "NONE", "N", "0"
								calculate_setRounding ($$ROUND_NONE, 0)
							CASE "FIXED", "FIX", "F", "DP"
								token$ = ReadToken$ ()
								IF token$ = "," THEN token$ = ReadToken$ ()
								param = XLONG(token$)
								IF param > 0 THEN
									calculate_setRounding ($$ROUND_FIXED, param)
								ELSE
									GOSUB ValueError
								END IF
							CASE "SCIENTIFIC", "SCI", "S", "SF"
								token$ = ReadToken$ ()
								IF token$ = "," THEN token$ = ReadToken$ ()
								param = XLONG(token$)
								IF param > 0 THEN
									calculate_setRounding ($$ROUND_SCIENTIFIC, param)
								ELSE
									GOSUB ValueError
								END IF
							CASE ELSE
								GOSUB ValueError
						END SELECT
					CASE "PRECISION", "PREC", "P"
						prec = XLONG(value$)
						IF prec > 0 THEN
							calculate_setPrecision (prec)
						ELSE
							GOSUB ValueError
						END IF
					CASE "DEBUG"
						SELECT CASE UCASE$(value$)
							CASE "ON"
								calcState.debug = $$TRUE
							CASE "OFF"
								calcState.debug = $$FALSE
							CASE ELSE
								GOSUB ValueError
						END SELECT
					CASE ELSE
						AddError ("\""+setting$+"\" is not a valid setting")
						INC cErrors
				END SELECT
			END IF
			
			IFZ cErrors THEN
				errorString$ = "Settings changed"
			END IF
			ret = 0
		CASE "INCLUDE","RUN"
			TakeToken ()
			fileName$ = TRIM$(MID$(in$, pos+1))
			fh = OPEN (fileName$, $$RD)
			IF fh = -1 THEN
				AddError ("Cannot find file \""+fileName$+"\"")
				INC cErrors
			ELSE
				lineNum = 0
				
				line$ = ""
				DO WHILE (line$ = "")||(MID$(TRIM$(line$),1,1)="#")
					line$ = INFILE$ (fh)
					INC lineNum
				LOOP
				
				bracketsP = 0
				bracketsS = 0
				bracketsC = 0
				
				overallError$ = ""
				DO WHILE line$
					FOR i = 0 TO UBOUND(line$)
						SELECT CASE line${i}
							CASE '('	: INC bracketsP
							CASE ')'	: DEC bracketsP
							CASE '['	: INC bracketsS
							CASE ']'	: DEC bracketsS
							CASE '{'	: INC bracketsC
							CASE '}'	: DEC bracketsC
						END SELECT
					NEXT
					
					toExecute$ = toExecute$+line$
					IFZ bracketsP|bracketsS|bracketsC THEN
						iExpr = calculate_parse (toExecute$, @errorString$)
						IFZ iExpr THEN
							overallError$ = overallError$ +"\nOn line "+STRING$(lineNum)+": " + errorString$
							INC cErrors
						ELSE
							calculate_execute (iExpr, 0, @exception, @errorString$)
							IF (exception != $$EXCEPTION_NONE)&&(exception != $$EXCEPTION_NOTNUM) THEN
								overallError$ = overallError$ +"\nOn line "+STRING$(lineNum)+": " + errorString$
								INC cErrors
							END IF
							calculate_clearExpression (iExpr)
						END IF
						
						toExecute$ = ""
					END IF
					
					line$ = ""				
					DO WHILE (line$ = "")||(MID$(TRIM$(line$),1,1)="#")
						IF EOF(fh) THEN EXIT DO 2
						line$ = INFILE$ (fh)
						INC lineNum
					LOOP
				LOOP
				CLOSE(fh)
				errors$ = overallError$
				errorString$ = "Successfully included \""+fileName$+"\""
			END IF
		CASE ELSE
			LinkedList_Init (@list)
			
			cErrors = parse_line (@list)
			ret = LINKEDLIST_New (list)
			
			IF calcState.debug THEN PRINT Deparse$ (list)
	END SELECT
	
	IF cErrors THEN
		IF MID$(errors$, 1, 1) = "\n" THEN
			errorString$ = LCLIP$(errors$)
		ELSE
			errorString$ = STRING$(cErrors)+" errors:"+LCLIP$(errors$)
			LINKEDLIST_Delete (ret)
		END IF
		ret = 0
	END IF

	RETURN ret
	
	SUB ValueError
		AddError ("\""+value$+"\" is not a valid value for the "+setting$+" setting")
		INC cErrors
	END SUB
END FUNCTION
'
' ################################
' #####  calculate_execute  #####
' ################################
' Executes a parsed expression
' id = the id of the expression to evaluate
' result = a pointer to the mpf_t to store the result
' exception = the variable to store an exception
' resultString$ = the variable to store the result string.  This will be a list of errors if there are any errors
' returns the nunmber of errors
FUNCTION calculate_execute (id, result, @exception, @resultString$)
	SHARED CALCSTATE calcState
	SHARED STACK evalStack
	SHARED errors$
	
	LINKEDLIST expression
	
	COMPLEX complex
	SYMBOL symbol
	
	FUNCADDR Delete(XLONG)
	FUNCADDR Delete2(XLONG)
	
	errors$ = ""
	resultString$ = ""
	exception = $$EXCEPTION_NONE
	
	' Push a dummy value onto the stack to handle commands that naturally underflow
	x = Allocate_Float ()
	__gmpf_set_ui (x, 0)
	eval_push ($$TYPE_NUMBER, x)
	
	IF LINKEDLIST_Get (id, @expression) THEN
		cErrors = evaluate (expression, @exception, @iData, @type, 0, -1)
		
		' pop off the dummy, preventing a memmory leak
		IF eval_pop (0, @iData2, @Delete2) THEN
			@Delete2(iData2)
		END IF
		
		IF exception = $$EXCEPTION_NONE
			SELECT CASE type
				CASE $$TYPE_NUMBER, $$TYPE_INT
					IF result THEN __gmpf_set (result, iData)
					Delete = &Delete_Float ()
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData, @complex)
					IF result THEN __gmpf_set (result, complex.real)
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Complex ()
				CASE $$TYPE_VECTOR
					IF result THEN __gmpf_set_ui (result, 0)
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Vector ()
				CASE $$TYPE_FUNCTION
					IF result THEN __gmpf_set_ui (result, 0)
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Function ()
			END SELECT
		END IF
	ELSE
		AddError ("[Internal error] Invalid expression id")
		INC cErrors
	END IF
	
	IF cErrors THEN
		resultString$ = STRING$(cErrors)+" errors: "+LCLIP$(errors$)
	ELSE
		' Replace the last answer
		SYMBOL_Get (calcState.iLastAns, @symbol)
		@symbol.delete (symbol.data)
		SYMBOL_Delete (calcState.iLastAns)
		symbol.type = type
		symbol.data = CopySymbol (iData, type) 
		symbol.delete = Delete
		calcState.iLastAns = SYMBOL_New (symbol)
		
		resultString$ = FormatSymbol$ (iData, type)
	END IF
	@Delete(iData)
	
	RETURN cErrors
END FUNCTION
'
' ####################################
' #####  calculate_setPrecision  #####
' ####################################
' Sets the precision in bits
' bits = the number of bits of precision
FUNCTION calculate_setPrecision (bits)
	SHARED CALCSTATE calcState
	SHARED CONSTANTS const
	
	__gmpf_set_default_prec (bits)
	calcState.maxDigits = Log10(2.0**DOUBLE(bits))
	
	' Delete any existing constants, the allocate new ones
	maxConsts = SIZE(CONSTANTS)>>2 -1
	FOR i = 1 TO maxConsts
		IF XLONGAT(&const, i<<2) THEN Delete_Float (XLONGAT(&const, i<<2))
		XLONGAT(&const, i<<2) = Allocate_Float ()
	NEXT
	
	const.cBitsPrec = bits
	TransLib_InitConstants (@const)
'	PRINT "pi = "; FormatFloat$(const.pPi)
'	PRINT "pi/2 = ";FormatFloat$(const.pPiBy2)
'	PRINT "3pi/2 = ";FormatFloat$(const.p3PiBy2)
'	PRINT "2pi = ";FormatFloat$(const.p2Pi)
'	PRINT "radToDeg = ";FormatFloat$(const.pRadToDeg)
'	PRINT "radToGrad = ";FormatFloat$(const.pRadToGrad)
'	PRINT "E = ";FormatFloat$(const.pE)
'	PRINT "ln2 = ";FormatFloat$(const.pLn2)
'	PRINT "ln10 = ";FormatFloat$(const.pLn10)
'	PRINT "sqr(2) = ";FormatFloat$(const.pSqrt2)
'	PRINT "1/sqr(2) = ";FormatFloat$(const.pInvSqrt2)
'	INLINE$("wait")
END FUNCTION
'
' ###################################
' #####  calculate_setRounding  #####
' ###################################
' Sets the rounding mode
' roundMode = the rounding mode
' parameter = the parameter for the rounding mode (e.g, significant figures or decimal digits)
FUNCTION calculate_setRounding (roundMode, parameter)
	SHARED CALCSTATE calcState
	
	calcState.roundMode = roundMode
	calcState.roundParameter = parameter
END FUNCTION
'
' ################################
' #####  calculate_setAngle  #####
' ################################
' Sets the angle mode
FUNCTION calculate_setAngle (angleMode)
	SHARED CALCSTATE calcState
	
	calcState.angleMode = angleMode
END FUNCTION
'
' ##################################
' #####  calculate_setOutMode  #####
' ##################################
' Sets the output mode
FUNCTION calculate_setOutMode (outMode)
	SHARED CALCSTATE calcState
	
	calcState.outMode = outMode
END FUNCTION
'
' ########################
' #####  ReadToken$  #####
' ########################
' Reads the next token from the input stream
' Returns the token or "" if there are no more tokens
FUNCTION ReadToken$ ()
	SHARED PeekBuffer$
	SHARED in$
	SHARED pos
	
	$STATE_START			= 0
	$STATE_NUMBER			= 1
	$STATE_AFTERPOINT	= 2
	$STATE_AFTERE			= 3
	$STATE_DIGIT			= 4
	$STATE_HEXNUM			= 5
	$STATE_ALPHA			= 6
	
	state = $STATE_START
	hexPrefix = $$FALSE
	token$ = ""
	PeekBuffer$ = ""
	
	DO
		IF pos > UBOUND(in$) THEN
			' the input expression has ended
			SELECT CASE state
				CASE $STATE_AFTERE
					AddError ("Expected exponent")
					RETURN " "
				CASE $STATE_HEXNUM
					IF hexPrefix THEN
						RETURN "0x"+token$
					ELSE
						IF (UBOUND(token$) > 0)&&(token${UBOUND(token$)} = 'b') THEN
							GOSUB CheckBin
						ELSE
							RETURN token$
						END IF
					END IF
				CASE ELSE
					RETURN token$
			END SELECT
		ELSE
			char = in${pos}
		END IF
		
		SELECT CASE state
			CASE $STATE_START
				SELECT CASE TRUE
					CASE isNum(char)
						token$ = token$+CHR$(char)
						state = $STATE_NUMBER
					CASE isHexit(char)
						token$ = token$+CHR$(char)
						state = $STATE_HEXNUM
					CASE isAlpha(char)
						token$ = token$+CHR$(char)
						state = $STATE_ALPHA
					CASE isWhitespace(char)
						' do nothing
					CASE ELSE
						token$ = CHR$(char)
						
						' Handle multiple character symbols
						SELECT CASE char
							CASE '<'
								IF pos < UBOUND(in$) THEN
									SELECT CASE in${pos+1}
										CASE '<', '=', '>'
											token$ = token$+CHR$(in${pos+1})
											INC pos
									END SELECT
								END IF
							CASE '>'
								IF pos < UBOUND(in$) THEN
									SELECT CASE in${pos+1}
										CASE '<', '=', '>'
											token$ = token$+CHR$(in${pos+1})
											INC pos
									END SELECT
								END IF
							CASE '&'
								IF pos < UBOUND(in$) THEN
									IF in${pos+1} = '&' THEN
										token$ = token$+"&"
										INC pos
									END IF
								END IF
							CASE '|'
								IF pos < UBOUND(in$) THEN
									IF in${pos+1} = '|' THEN
										token$ = token$+"|"
										INC pos
									END IF
								END IF
							END SELECT
						
						INC pos
						RETURN token$
				END SELECT
			CASE $STATE_NUMBER
				SELECT CASE TRUE
					CASE isNum(char)
						token$ = token$+CHR$(char)
					CASE isHexit(char)
						token$ = token$+CHR$(char)
						state = $STATE_HEXNUM
					CASE char = 'h'
						INC pos
						RETURN "0x"+token$
					CASE char = 'o'
						num = $$TRUE
						FOR i = 0 TO UBOUND(token$)
							IFF isNum(token${i})&&(token${i} != '8')&&(token${i} != '9') THEN
								num = $$FALSE
								EXIT FOR
							END IF
						NEXT
						IF num THEN 
							INC pos
							RETURN "0o"+token$
						ELSE
							RETURN token$
						END IF
					CASE char = 'x'
						IF token$ = "0" THEN
							token$ = ""
							state = $STATE_HEXNUM
							hexPrefix = $$TRUE
						ELSE
							RETURN token$
						END IF
					CASE char = '.'
						token$ = token$+"."
						state = $STATE_AFTERPOINT
					CASE char = 'e'
						token$ = token$+"@"
						state = $STATE_AFTERE
					CASE ELSE
						RETURN token$
				END SELECT
			CASE $STATE_AFTERPOINT
				SELECT CASE TRUE
					CASE isNum(char)
						token$ = token$+CHR$(char)
					CASE char = 'e'
						token$ = token$+"@"
						state = $STATE_AFTERE
					CASE ELSE
						RETURN token$
				END SELECT
			CASE $STATE_AFTERE
				SELECT CASE TRUE
					CASE isNum(char)
						token$ = token$+CHR$(char)
						state = $STATE_DIGIT
					CASE (char = '+')||(char = '-')
						token$ = token$+CHR$(char)
						state = $STATE_DIGIT
					CASE ELSE
						AddError ("Expected exponent")
						RETURN " "
				END SELECT
			CASE $STATE_DIGIT
				IF isNum(char) THEN
					token$ = token$+CHR$(char)
				ELSE
					RETURN token$
				END IF
			CASE $STATE_HEXNUM
				SELECT CASE TRUE
					CASE isHexit(char)
						token$ = token$+CHR$(char)
					CASE (char = 'h')&&(!hexPrefix)
						INC pos
						RETURN "0x"+token$
					CASE (isAlpha(char))&&(!hexPrefix)
						IF isNum(token${0}) THEN
							FOR i = UBOUND(token$) TO 0 STEP -1
								IF !isNum(token${i}) THEN
									token$ = RCLIP$(token$)
									DEC pos
								END IF
							NEXT
							RETURN token$
						ELSE
							token$ = token$+CHR$(char)
							state = $STATE_ALPHA
						END IF
					CASE ELSE
						IF hexPrefix THEN
							RETURN "0x"+token$
						ELSE
							IF UBOUND(token$)>0 THEN
								SELECT CASE token${UBOUND(token$)}
									CASE 'e'
										num = $$TRUE
										FOR i = 0 TO UBOUND(token$)-1
											IF !isNum (token${i}) THEN
												num = $$FALSE
												EXIT FOR
											END IF
										NEXT
										IF num THEN
											DEC pos
											state = $STATE_AFTERE
										END IF
									CASE 'b'
										GOSUB CheckBin
									CASE ELSE
										RETURN token$
								END SELECT
							ELSE
								RETURN token$
							END IF
						END IF
				END SELECT
			CASE $STATE_ALPHA
				IF isAlphaNum (char) THEN
					token$ = token$+CHR$(char)
				ELSE
					RETURN token$
				END IF
		END SELECT
		
		INC pos
	LOOP
	
	RETURN ""
	
	SUB CheckBin
		num = $$TRUE
		FOR i = 0 TO UBOUND(token$)-1
			IFF (token${i} = '0')||(token${i} = '1') THEN
				num = $$FALSE
				EXIT FOR
			END IF
		NEXT
		IF num THEN 
			RETURN "0b"+RCLIP$(token$)
		ELSE
			DEC pos
			RETURN RCLIP$(token$)
		END IF
	END SUB
END FUNCTION
'
' ########################
' #####  PeekToken$  #####
' ########################
' Peeks at the next token, it can be read again
' Returns the same as ReadToken$
FUNCTION PeekToken$ ()
	SHARED pos
	SHARED nextToken
	SHARED PeekBuffer$
	
	IF PeekBuffer$ THEN
		RETURN PeekBuffer$
	ELSE
		savedPos = pos
		PeekBuffer$ = ReadToken$ ()
		nextToken = pos
		pos = savedPos
		
		RETURN PeekBuffer$
	END IF
END FUNCTION
FUNCTION isNum (char)
	IF (char >= '0' && char <= '9') THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
FUNCTION isHexit (char)
	IF isNum (char) || (char >= 'a' && char <= 'f') || (char >= 'A' && char <='F') THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
FUNCTION isAlpha (char)
	IF (char >= 'A' && char <= 'Z')||(char >= 'a' && char <= 'z')||(char = '_') THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
FUNCTION isAlphaNum (char)
	IF isNum(char)||isAlpha(char) THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
FUNCTION isWhitespace (char)
	IF (char = ' ')||(char = '\t')||(char = '\v')||(char = '\f')||(char = '\r')||(char = '\n') THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
'
' ######################
' #####  AddError  #####
' ######################
' Adds an error to the errors list
FUNCTION AddError (error$)
	SHARED errors$
	errors$ = errors$+"; "+error$
END FUNCTION
'
' ########################
' #####  parse_line  #####
' ########################
' The root of the parser
' out = the output from the parser
' returns then number of errors
FUNCTION parse_line (LINKEDLIST out)
	cErrors = 0
	
	token$ = UCASE$(PeekToken$ ())
	IF (token$ = "LET")||(token$ = "CLEAR") THEN
		cErrors = cErrors+parse_command (@out)
	ELSE
		cErrors = cErrors+parse_expression (@out, @exprType)
	END IF

	RETURN cErrors
END FUNCTION
'
' ###########################
' #####  parse_command  #####
' ###########################
FUNCTION parse_command (LINKEDLIST out)
	TOKEN token
	SYMBOL symbol
	cErrors = 0
	token$ = ReadToken$ ()

	SELECT CASE UCASE$(token$)
		CASE "LET"
			' Get any number of comma seperated variable names
			DIM vars[7]
			cVars = 0
			DO
				var$ = ReadToken$ ()
				IF !isVarNameValid (var$) THEN
					AddError ("You cannot use \""+var$+"\" as a variable name, it is a reserved word")
					INC cErrors
					DEC cVars		' Remove this variable from the array as it is invalid
				ELSE
					vars[cVars] = STRING_New (var$)
				END IF
				
				SELECT CASE ReadToken$ ()
					CASE ","
					CASE "="
						EXIT DO
					CASE ELSE
						AddError ("Expected =")
						INC cErrors
						EXIT DO
				END SELECT
				
				INC cVars
				IF UBOUND(vars[]) < cVars THEN REDIM vars[((UBOUND(vars[])+1)<<1)-1]
			LOOP
			
			cErrors = cErrors+parse_expression (@out, @exprType)
			FOR i = 0 TO cVars
				token.op = $$CMD_STOR
				token.value = vars[i]
				LinkedList_Append (@out, TOKEN_New (token))
			NEXT
		CASE "CLEAR"
			DO
				var$ = ReadToken$ ()
				IF SymbolTable_Retreive (var$, @symbol) THEN
					token.op = $$CMD_CLEAR
					token.value = STRING_New (var$)
					LinkedList_Append (@out, TOKEN_New (token))
				ELSE
					AddError ("\""+var$+"\" is not defined")
					INC cErrors
				END IF
				
				token$ = ReadToken$ ()
				IF token$ != "," THEN EXIT DO
			LOOP
	END SELECT
	
	RETURN cErrors
END FUNCTION
'
' ##############################
' #####  parse_expression  #####
' ##############################
FUNCTION parse_expression (LINKEDLIST out, @exprType)
	SYMBOL symbol
	cErrors = 0
	
	' Assume we are dealing with a numeric expression
	exprType = $$TYPE_NUMBER
	'cErrors = cErrors+parse_nexpr (@out, @exprType)
	cErrors = cErrors+parse_condition (@out, @exprType)
	
	RETURN cErrors
END FUNCTION
'
' #########################
' #####  parse_nexpr  #####
' #########################
FUNCTION parse_nexpr (LINKEDLIST out, @exprType)
	TOKEN token
	cErrors = 0
	
	cErrors = cErrors + parse_product (@out, @exprType)
	
	token$ = PeekToken$ ()
	DO WHILE token$
		token.value = 0
		SELECT CASE UCASE$(token$)
'			CASE ")"
'				TakeToken ()
'				RETURN cErrors
			CASE "+"
				TakeToken ()
				cErrors = cErrors + parse_product (@out, @exprType)
				token.op = $$OP_ADD
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "-"
				TakeToken ()
				cErrors = cErrors + parse_product (@out, @exprType)
				token.op = $$OP_SUB
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "OR"
				TakeToken ()
				cErrors = cErrors + parse_product (@out, @exprType)
				token.op = $$OP_OR
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "XOR"
				TakeToken ()
				cErrors = cErrors + parse_product (@out, @exprType)
				token.op = $$OP_XOR
				LinkedList_Append (@out, TOKEN_New(token))
			CASE ELSE
				EXIT DO
		END SELECT
		
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
END FUNCTION
'
' ###########################
' #####  parse_product  #####
' ###########################
FUNCTION parse_product (LINKEDLIST out, @exprType)
	TOKEN token
	cErrors = 0
	
	cErrors = cErrors + parse_factor (@out, @exprType)
	
	token$ = PeekToken$ ()
	DO WHILE token$
		token.value = 0
		SELECT CASE UCASE$(token$)
			CASE "*"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_MUL
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "/"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_DIV
				LinkedList_Append (@out, TOKEN_New(token))
			CASE ">>"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_DOWNSHIFT
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "<<"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_UPSHIFT
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "AND"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_AND
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "COM"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_C
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "PER"
				TakeToken ()
				cErrors = cErrors + parse_factor (@out, @exprType)
				token.op = $$OP_P
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "-", ",", "]", ":", "}", ">", "<", ">=", "<=", "<>", "><", "=", ";", "&&", "||"
				EXIT DO
			CASE ELSE
				' This allows for implied multiplication
				IFZ parse_term (@out, @exprType) THEN
					token.op = $$OP_MUL
					LinkedList_Append (@out, TOKEN_New(token))
				ELSE
					EXIT DO
				END IF
		END SELECT
		
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
END FUNCTION
'
' ##########################
' #####  parse_factor  #####
' ##########################
FUNCTION parse_factor (LINKEDLIST out, @exprType)
	TOKEN token
	cErrors = 0
	
	token$ = PeekToken$ ()
	GOSUB SafeParseTerm
	
	token$ = PeekToken$ ()
	DO WHILE token$
		token.value = 0
		SELECT CASE UCASE$(token$)
			CASE "^"
				TakeToken ()
				GOSUB SafeParseTerm
				token.op = $$OP_RAISE
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "RT"
				TakeToken ()
				GOSUB SafeParseTerm
				token.op = $$OP_ROOT
				LinkedList_Append (@out, TOKEN_New(token))
			CASE ELSE
				EXIT DO
		END SELECT
		
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
	SUB SafeParseTerm
		newErrors = parse_term (@out, @exprType)
		IF newErrors THEN
			AddError ("Undefined symbol \""+token$+"\"")
			TakeToken ()
		END IF
		cErrors = cErrors+newErrors
	END SUB
END FUNCTION
'
' ########################
' #####  parse_term  #####
' ########################
FUNCTION parse_term (LINKEDLIST out, @exprType)
	SHARED CALCSTATE calcState
	SHARED LINKEDLIST	compileParamStack
	TOKEN token
	TOKEN token2
	SYMBOL symbol
	
	cErrors = 0

	exprType = $$TYPE_NUMBER
	
	token$ = PeekToken$ ()
	
	SELECT CASE UCASE$(token$)
		CASE " "
			INC cErrors
		CASE ""
			AddError ("Syntax Error, expected a number")
			INC cErrors
		CASE ",", "]", ":", "}", ">", "<", ">=", "<=", "<>", "><", "=", ";", "&&", "||"
			RETURN cErrors
		CASE "("
			TakeToken ()
			'cErrors = cErrors + parse_nexpr (@out, @exprType)
			cErrors = cErrors + parse_condition (@out, @exprType)
			token$ = PeekToken$ ()
			IF token$ != ")" THEN
				AddError ("Expected )")
			ELSE
				TakeToken ()
			END IF
		CASE "[", "MKV"
			' Parse a vector expression
			cErrors = cErrors + parse_vexpr (@out, @exprType)
		CASE "DIM"
			token.op = $$OP_VECTORDIM
			GOSUB outUnop
		CASE "{", "SELF"
			' Parse a function expression
			cErrors = cErrors + parse_fexpr (@out, @exprType)
		CASE "PI"
			TakeToken ()
			token.op = $$OP_PI
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "E"
			TakeToken ()
			token.op = $$OP_E
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "I"
			TakeToken ()
			token.op = $$OP_I
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "RAND"
			TakeToken ()
			token.op = $$OP_RAND
			token.value= 0
			LinkedList_Append (@out, TOKEN_New(token))
		CASE "@", "ANS", "LASTANS"
			SYMBOL_Get (calcState.iLastAns, @symbol)
			SELECT CASE symbol.type
				CASE $$TYPE_VECTOR
					cErrors = cErrors+parse_vexpr (@out, @exprType)
				CASE $$TYPE_FUNCTION
					cErrors = cErrors+parse_fexpr (@out, @exprType)
				CASE ELSE
					TakeToken ()
					token.op = $$OP_LASTANS
					token.value = 0
					LinkedList_Append (@out, TOKEN_New (token))
			END SELECT
		CASE "-"
			token.op = $$OP_NEGATE
			GOSUB outUnop
		CASE "NOT"
			token.op = $$OP_NOT
			GOSUB outUnop
		CASE "SQR"
			token.op = $$OP_SQR
			GOSUB outUnop
		CASE "CBR"
			token.op = $$OP_CBR
			GOSUB outUnop
		CASE "SIN"
			token.op = $$OP_SIN
			GOSUB outUnop
		CASE "COS"
			token.op = $$OP_COS
			GOSUB outUnop
		CASE "TAN"
			token.op = $$OP_TAN
			GOSUB outUnop
		CASE "COT"
			token.op = $$OP_COT
			GOSUB outUnop
		CASE "SEC"
			token.op = $$OP_SEC
			GOSUB outUnop
		CASE "COSEC"
			token.op = $$OP_COSEC
			GOSUB outUnop
		CASE "ASIN"
			token.op = $$OP_ASIN
			GOSUB outUnop
		CASE "ACOS"
			token.op = $$OP_ACOS
			GOSUB outUnop
		CASE "ATAN"
			token.op = $$OP_ATAN
			GOSUB outUnop
		CASE "ACOT"
			token.op = $$OP_ACOT
			GOSUB outUnop
		CASE "ASEC"
			token.op = $$OP_ASEC
			GOSUB outUnop
		CASE "ACOSEC"
			token.op = $$OP_ACOSEC
			GOSUB outUnop
		CASE "SINH"
			token.op = $$OP_SINH
			GOSUB outUnop
		CASE "COSH"
			token.op = $$OP_COSH
			GOSUB outUnop
		CASE "TANH"
			token.op = $$OP_TANH
			GOSUB outUnop
		CASE "COTH"
			token.op = $$OP_COTH
			GOSUB outUnop
		CASE "SECH"
			token.op = $$OP_SECH
			GOSUB outUnop
		CASE "COSECH"
			token.op = $$OP_COSECH
			GOSUB outUnop
		CASE "ASINH"
			token.op = $$OP_ASINH
			GOSUB outUnop
		CASE "ACOSH"
			token.op = $$OP_ACOSH
			GOSUB outUnop
		CASE "ATANH"
			token.op = $$OP_ATANH
			GOSUB outUnop
		CASE "ACOTH"
			token.op = $$OP_ACOTH
			GOSUB outUnop
		CASE "ASECH"
			token.op = $$OP_ASECH
			GOSUB outUnop
		CASE "ACOSECH"
			token.op = $$OP_ACOSECH
			GOSUB outUnop
		CASE "LOGE","LN"
			token.op = $$OP_LOGE
			GOSUB outUnop
		CASE "INT"
			token.op = $$OP_INT
			GOSUB outUnop
		CASE "FRAC"
			token.op = $$OP_FRAC
			GOSUB outUnop
		CASE "EXP"
			token.op = $$OP_EXP
			GOSUB outUnop
		CASE "RE"
			token.op = $$OP_RE
			GOSUB outUnop
		CASE "IM"
			token.op = $$OP_IM
			GOSUB outUnop
		CASE "CONJ"
			token.op = $$OP_CONJ
			GOSUB outUnop
		CASE ELSE
			' Perhaps this is a parameter?
			IF SearchCompileTimeParamStack (token$, @index, @level, @type) THEN
				SELECT CASE type
					CASE $$TYPE_BOOL, $$TYPE_INT, $$TYPE_NUMBER, $$TYPE_COMPLEX
						TakeToken ()
						token.op = $$OP_PARAMETER
						token.value = index
						token.value2 = level
						LinkedList_Append (@out, TOKEN_New (token))
					CASE $$TYPE_VECTOR
						' Parse a vector expression
						cErrors = cErrors + parse_vexpr (@out, @exprType)
					CASE $$TYPE_FUNCTION
						' Parse a function call
						cErrors = cErrors + parse_fexpr (@out, @exprType)
				END SELECT
			ELSE
				iSymbol = SymbolTable_Retreive (token$, @symbol)
				IF iSymbol THEN
					SELECT CASE symbol.type
						CASE $$TYPE_BOOL, $$TYPE_NUMBER, $$TYPE_COMPLEX
							TakeToken ()
							token.op = $$OP_VAR
							token.value = iSymbol
							LinkedList_Append (@out, TOKEN_New(token))
						CASE $$TYPE_VECTOR
							' Parse a vector expression
							cErrors = cErrors + parse_vexpr (@out, @exprType)
						CASE $$TYPE_FUNCTION
							' Parse a function call
							cErrors = cErrors + parse_fexpr (@out, @exprType)
					END SELECT
				ELSE
					' It might be a literal number
					x = Allocate_Float ()
					IF StringToFloat (token$, x) THEN
						' This is not a valid term
						Delete_Float (x)
						RETURN cErrors+1
					ELSE
						TakeToken ()
					END IF
					token.op = $$OP_LITERAL
					token.value = x
					LinkedList_Append (@out, TOKEN_New(token))
				END IF
			END IF
	END SELECT

	IF exprType = $$TYPE_NUMBER THEN
		token$ = PeekToken$ ()
		SELECT CASE token$
			CASE "!"
				TakeToken ()
				token.op = $$OP_FACT
				token.value = 0
				LinkedList_Append (@out, TOKEN_New(token))
		END SELECT
	END IF

	RETURN cErrors
	
	SUB outUnop
		TakeToken ()
		parse_term (@out, @exprType)
		token.value = 0
		
		LinkedList_GetItem (out, out.cItems-1, @iData)
		TOKEN_Get (iData, @token2)
		IF token2.op = $$OP_FACT THEN
			LinkedList_Insert (@out, out.cItems-1, TOKEN_New(token))
		ELSE
			LinkedList_Append (@out, TOKEN_New(token))
		END IF
	END SUB
END FUNCTION
'
' ##########################
' #####  FormatFloat$  #####
' ##########################
' Formats a gmp float according to the current rounding mode
' pNumber = a pointer to the number to convert
' returns the formatted string
FUNCTION FormatFloat$ (pNumber)
	SHARED CALCSTATE calcState
	SHARED CONSTANTS const
	mpz_t Rint
	
	FUNCADDR GmpFree(XLONG)
	
	int0 = &Rint
	__gmpz_init (int0)
	
	SELECT CASE calcState.outMode
		CASE $$OUT_BIN
			__gmpz_set_f (int0, pNumber)
			out$ = NULL$(__gmpz_sizeinbase (int0, 2)+2)
			out$ = CSTRING$(__gmpz_get_str (&out$, 2, int0)) + "b"
		CASE $$OUT_HEX
			__gmpz_set_f (int0, pNumber)
			out$ = NULL$(__gmpz_sizeinbase (int0, 16)+2)
			out$ = CSTRING$(__gmpz_get_str (&out$, 16, int0)) + "h"
		CASE $$OUT_OCT
			__gmpz_set_f (int0, pNumber)
			out$ = NULL$(__gmpz_sizeinbase (int0, 8)+2)
			out$ = CSTRING$(__gmpz_get_str (&out$, 8, int0)) + "o"
		CASE $$OUT_DEC
			pStr = __gmpf_get_str (0, &exp, 10, 0, pNumber)
			stri$ = CSTRING$ (pStr)
			gmp_free (pStr, 0)
			
			out$ = ""
			IF UBOUND(stri$) < 0 && exp = 0 THEN
				out$ = "0"
			ELSE
				' Handle negatives
				IF MID$(stri$, 1, 1) = "-" THEN
					stri$ = LCLIP$(stri$)
					neg = $$TRUE
				ELSE
					neg = $$FALSE
				END IF
				
				' Rounding to deal with errors in binary representation
				IF UBOUND(stri$) >= calcState.maxDigits THEN
					stri$ = RCLIP$(stri$)
					i = UBOUND(stri$)
					IF (stri${i}-'0') >= 5 THEN
						stri$ = RCLIP$(stri$)
						DEC i
						DO WHILE i > -1
							digit = stri${i}-'0'
							IF digit < 9 THEN
								UBYTEAT(&stri$,i) = stri${i}+1
								EXIT DO
							ELSE
								stri$ = RCLIP$(stri$)
								DEC i
							END IF
						LOOP
'					ELSE
'						UBYTEAT(&stri$,i) = '0'
'						DEC i
'						IF i > -1 THEN
'							IF (stri${i}-'0') <= 5 THEN
'								UBYTEAT(&stri$,i) = '0'
'							END IF
'						END IF
					END IF
				END IF
				
				' trim off any trailing 0's
				i = UBOUND(stri$)
				DO WHILE stri${i} = '0'
					stri$ = RCLIP$(stri$)
					DEC i
					IF i < 0 THEN EXIT DO
				LOOP
				SELECT CASE calcState.roundMode
					CASE $$ROUND_NONE
						IF exp < 1 THEN out$ = out$+"0"		' Write a leading 0 if we need one
						IF exp < 0 THEN										' Write leading 0's after the point
							out$ = out$+"."
							FOR i = 0 TO (ABS(exp)-1)
								out$ = out$+"0"
							NEXT
							fracDigit = 0		' The 0 based index of the first fractional digit
						ELSE				' Write out the non fractional part
							FOR i = 0 TO (exp-1)
								IF i > UBOUND(stri$) THEN out$ = out$+"0" ELSE out$ = out$+CHR$(stri${i})
							NEXT
							fracDigit = exp
							IF fracDigit <= UBOUND(stri$) THEN out$ = out$+"."
						END IF
						FOR i = fracDigit TO UBOUND(stri$)
							out$ = out$+CHR$(stri${i})
						NEXT
					CASE $$ROUND_FIXED
						' The toughest rounding mode
						
						digitsToGo = calcState.roundParameter
						IF exp < 1 THEN out$ = out$+"0"		' Write a leading 0 if we need one
						IF exp < 0 THEN										' Write leading 0's after the point
							out$ = out$+"."
							FOR i = 0 TO (ABS(exp)-1)
								IF digitsToGo = 0 THEN EXIT FOR
								out$ = out$+"0"
								DEC digitsToGo
							NEXT
							fracDigit = 0		' The 0 based index of the first fractional digit
						ELSE				' Write out the non fractional part
							FOR i = 0 TO (exp-1)
								IF i > UBOUND(stri$) THEN out$ = out$+"0" ELSE out$ = out$+CHR$(stri${i})
							NEXT
							fracDigit = exp
							'IF fracDigit <= UBOUND(stri$) THEN out$ = out$+"."
							out$ = out$+"."
						END IF
						FOR i = fracDigit TO UBOUND(stri$)
							IF digitsToGo = 0 THEN EXIT FOR
							out$ = out$+CHR$(stri${i})
							DEC digitsToGo
						NEXT
						
						IF (digitsToGo = 0) THEN GOSUB RoundUp
						
						FOR i = 1 TO digitsToGo
							' add 0's to tidy up the number
							out$ = out$+"0"
						NEXT
					CASE $$ROUND_SCIENTIFIC
						' This is refreshingly easy
						out$ = CHR$(stri${0})+"."
						FOR i = 1 TO calcState.roundParameter-1
							IF i > UBOUND(stri$) THEN char$ = "0" ELSE char$ = CHR$(stri${i})
							out$ = out$+char$
						NEXT
						
						' At this point we'll do some rounding
						GOSUB RoundUp
						
						out$ = out$+"e"+STRING$(exp-1)
						
						i = calcState.roundParameter+1	' This is probably uneccassary
				END SELECT
			END IF
			
			IF neg THEN out$ = "-"+out$
	END SELECT
	
	__gmpz_clear (int0)
	
	RETURN out$
	
	SUB RoundUp
		IF i <= UBOUND(stri$) THEN
			' we need to do a little rounding
			IF stri${i} > '4' THEN
				i = UBOUND(out$)
				
				DO WHILE (out${i} = '9')||(out${i} = '.')
					IF out${i} != '.' THEN UBYTEAT(&out$,i) = '0'
					DEC i
					IF i = -1 THEN
						SELECT CASE calcState.roundMode
							CASE $$ROUND_FIXED
								out$ = "0"+out$
							CASE $$ROUND_SCIENTIFIC
								out$ = "0."+MID$(out$, 1, 1)+MID$(out$, LEN(out$)-1)
								INC exp
						END SELECT
						i = 0
					END IF
				LOOP
				
				UBYTEAT(&out$, i) = out${i}+1
			END IF
		END IF
	END SUB
END FUNCTION
'
' ######################
' #####  Deparse$  #####
' ######################
' Deparses an expression
' iExpression = the expression to deparse
' returns the deparsed expression
FUNCTION Deparse$ (LINKEDLIST expression)
	SHARED CONSTANTS const
	TOKEN token
	FUNC func
	FUNCCASE case
	
	out$ = ""
	hWalk = LinkedList_StartWalk (expression)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		TOKEN_Get (iData, @token)
		SELECT CASE token.op
			CASE $$OP_LITERAL
				out$ = out$ + " " + FormatFloat$ (token.value)
			CASE $$OP_VAR
				out$ = out$ + " var["+STRING$(token.value)+"]"
			CASE $$OP_PI
				out$ = out$ + " pi"
			CASE $$OP_E
				out$ = out$ + " e"
			CASE $$OP_I
				out$ = out$ + " i"
			CASE $$OP_RAND
				out$ = out$ + " RAND"
			CASE $$OP_MAKEVECTOR
				out$ = out$ + " mv("+STRING$(token.value)+")"
			CASE $$OP_FUNCCALL
				out$ = out$ + " call("+STRING$(token.value)+")"
			CASE $$OP_PARAMETER
				out$ = out$ + " parameter[index: "+STRING$(token.value)+", level: "+STRING$(token.value2)+"]"
			CASE $$OP_FUNCLIT
				out$ = out$ + " func("
				FUNC_Get (token.value, @func)
				hParamsWalk = LinkedList_StartWalk (func.paramTypes)
				DO WHILE LinkedList_Walk (hParamsWalk, @iParamData)
					SELECT CASE iParamData
						CASE $$TYPE_INT
							out$ = out$ + "int, "
						CASE $$TYPE_NUMBER
							out$ = out$ + "real, "
						CASE $$TYPE_COMPLEX
							out$ = out$ + "complex, "
						CASE $$TYPE_VECTOR
							out$ = out$ + "vector, "
						CASE $$TYPE_FUNCTION
							out$ = out$ + "function, "
					END SELECT
				LOOP
				LinkedList_EndWalk (hParamsWalk)
				out$ = RCLIP$(out$, 2)+")"
				
				hCasesWalk = LinkedList_StartWalk (func.cases)
				DO WHILE LinkedList_Walk (hCasesWalk, @iCasesData)
					FUNCCASE_Get (iCasesData, @case)
					
					out$ = out$ + " case: " + Deparse$ (case.condExpr) + " DO " + Deparse$ (case.realExpr)
				LOOP
				LinkedList_EndWalk (hCasesWalk)
				
				out$ = out$ + " case else: " + Deparse$ (func.caseElse)
			CASE $$OP_RECURSE
				out$ = out$ + " SELF"
			CASE $$OP_VECTORDIM
				out$ = out$ + " VECDIM"
			CASE $$OP_GENVEC
				out$ = out$ + " GENVEC"
			CASE $$OP_VECCOM
				out$ = out$ + " VECCOMP"
			CASE $$OP_LASTANS
				out$ = out$ + " @"
			CASE $$OP_LOGAND
				out$ = out$ + " &&"
			CASE $$OP_LOGOR
				out$ = out$ + " ||"
			CASE $$OP_SHORTC_T
				out$ = out$ + " SHORT_T"
			CASE $$OP_SHORTC_F
				out$ = out$ + " SHORT_F"
			CASE $$OP_RELEQ
				out$ = out$ + " ="
			CASE $$OP_RELGT
				out$ = out$ + " >"
			CASE $$OP_RELLT
				out$ = out$ + " <"
			CASE $$OP_RELGTE
				out$ = out$ + " >="
			CASE $$OP_RELLTE
				out$ = out$ + " <="
			CASE $$OP_RELNOTEQ
				out$ = out$ + " ><"
			CASE $$OP_ADD
				out$ = out$ + " +"
			CASE $$OP_SUB
				out$ = out$ + " -"
			CASE $$OP_OR
				out$ = out$ + " OR"
			CASE $$OP_XOR
				out$ = out$ + " XOR"
			CASE $$OP_MUL
				out$ = out$ + " *"
			CASE $$OP_DIV
				out$ = out$ + " /"
			CASE $$OP_UPSHIFT
				out$ = out$ + " <<"
			CASE $$OP_DOWNSHIFT
				out$ = out$ + " >>"
			CASE $$OP_AND
				out$ = out$ + " AND"
			CASE $$OP_C
				out$ = out$ + " C"
			CASE $$OP_P
				out$ = out$ + " P"
			CASE $$OP_DOTPROD
				out$ = out$ + " ."
			CASE $$OP_RAISE
				out$ = out$ + " ^"
			CASE $$OP_ROOT
				out$ = out$ + " Rt"
			CASE $$OP_NEGATE
				out$ = out$ + " NEG"
			CASE $$OP_FACT
				out$ = out$ + " !"
			CASE $$OP_NOT
				out$ = out$ + " NOT"
			CASE $$OP_SQR
				out$ = out$ + " SQR"
			CASE $$OP_CBR
				out$ = out$ + " CBR"
			CASE $$OP_SIN
				out$ = out$ + " SIN"
			CASE $$OP_COS
				out$ = out$ + " COS"
			CASE $$OP_TAN
				out$ = out$ + " TAN"
			CASE $$OP_COT
				out$ = out$ + " COT"
			CASE $$OP_SEC
				out$ = out$ + " SEC"
			CASE $$OP_COSEC
				out$ = out$ + " COSEC"
			CASE $$OP_ASIN
				out$ = out$ + " ASIN"
			CASE $$OP_ACOS
				out$ = out$ + " ACOS"
			CASE $$OP_ATAN
				out$ = out$ + " ATAN"
			CASE $$OP_ACOT
				out$ = out$ + " ACOT"
			CASE $$OP_ASEC
				out$ = out$ + " ASEC"
			CASE $$OP_ACOSEC
				out$ = out$ + " ACOSEC"
			CASE $$OP_SINH
				out$ = out$ + " SINH"
			CASE $$OP_COSH
				out$ = out$ + " COSH"
			CASE $$OP_TANH
				out$ = out$ + " TANH"
			CASE $$OP_COTH
				out$ = out$ + " COTH"
			CASE $$OP_SECH
				out$ = out$ + " SECH"
			CASE $$OP_COSECH
				out$ = out$ + " COSECH"
			CASE $$OP_ASINH
				out$ = out$ + " ASINH"
			CASE $$OP_ACOSH
				out$ = out$ + " ACOSH"
			CASE $$OP_ATANH
				out$ = out$ + " ATANH"
			CASE $$OP_ACOTH
				out$ = out$ + " ACOTH"
			CASE $$OP_ASECH
				out$ = out$ + " ASECH"
			CASE $$OP_ACOSECH
				out$ = out$ + " ACOSECH"
			CASE $$OP_INT
				out$ = out$ + " INT"
			CASE $$OP_FRAC
				out$ = out$ + " FRAC"
			CASE $$OP_LOGE
				out$ = out$ + " LOGE"
			CASE $$CMD_STOR
				STRING_Get (token.value, @var$)
				out$ = out$ + " STOR "+var$
			CASE $$CMD_CLEAR
				STRING_Get (token.value, @var$)
				out$ = out$ + " CLEAR "+var$
			CASE ELSE
				out$ = out$ + " op("+STRING$(token.op)+", "+STRING$(token.value)+")"
		END SELECT
	LOOP
	LinkedList_EndWalk (hWalk)
	
	RETURN LCLIP$(out$)
END FUNCTION
'
' #######################################
' #####  calculate_clearExpression  #####
' #######################################
' Deletes a previously parsed expression
' id = the id of the expression to clear
' returns $$TRUE on success or $$FALSE on fail
FUNCTION calculate_clearExpression (id)
	LINKEDLIST list
	TOKEN token
	
	IFF LINKEDLIST_Get (id, @list) THEN RETURN $$FALSE
	Delete_Expression (list)
	LINKEDLIST_Delete (id)
	
	RETURN $$TRUE
END FUNCTION
'
' #######################
' #####  TakeToken  #####
' #######################
' Takes the last token
FUNCTION TakeToken ()
	SHARED nextToken
	SHARED pos
	SHARED PeekBuffer$
	
	pos = nextToken
	PeekBuffer$ = ""
END FUNCTION
'
' ##############################
' #####  SymbolTable_Init  #####
' ##############################
' Initialises the symbol table
' returns $$TRUE on error otherwise $$FALSE
FUNCTION SymbolTable_Init ()
	SHARED ASSOCARRAY symbolTable
	
	AssocArray_Init (@symbolTable)
	
	RETURN $$FALSE
END FUNCTION
'
' ################################
' #####  SymbolTable_Insert  #####
' ################################
' Inserts a symbol into the symbol table
' name$ = the name of the symbol
' symbol = the symbol to insert
' returns the id of the symbol or 0 if it could not be added
FUNCTION SymbolTable_Insert (name$, SYMBOL symbol)
	SHARED ASSOCARRAY symbolTable
	
	iData = SYMBOL_New(symbol)
	ret = AssocArray_Insert (@symbolTable, UCASE$(name$), iData)
	IF ret THEN
		RETURN iData
	ELSE
		SYMBOL_Delete (iData)
		RETURN 0
	END IF
END FUNCTION
'
' ##################################
' #####  SymbolTable_Retreive  #####
' ##################################
' Retreives a symbol from the symbol table
' name$ = the name of the symbol
' symbol = the variable to store the symbol
' returns the id of the returned symbol or 0 on fail
FUNCTION SymbolTable_Retreive (name$, SYMBOL symbol)
	SHARED ASSOCARRAY symbolTable
	
	AssocArray_Find (symbolTable, UCASE$(name$), @iData)
	SYMBOL_Get (iData, @symbol)
	RETURN iData
END FUNCTION
'
' ################################
' #####  SymbolTable_Delete  #####
' ################################
' Deletes a symbol from the symbol table
' name$ = the name of the symbol to delete
' returns $$TRUE on success or $$FALSE on fail
FUNCTION SymbolTable_Delete (name$)
	SHARED ASSOCARRAY symbolTable
	SYMBOL symbol
	
	ret = AssocArray_Delete (@symbolTable, UCASE$(name$), @iData)
	
	' Delete the symbol safely
	SYMBOL_Get (iData, @symbol)
	@symbol.delete (symbol.data)
	SYMBOL_Delete (iData)
	RETURN ret
END FUNCTION
'
' ###################################
' #####  SymbolTable_DeleteAll  #####
' ###################################
' Deletes all the symbols in the symbol table
' returns $$TRUE on success or $$FALSE on fail
FUNCTION SymbolTable_DeleteAll ()
	SHARED ASSOCARRAY symbolTable
	
	RETURN AssocArray_Clear (@symbolTable)
END FUNCTION
'
' ############################
' #####  isVarNameValid  #####
' ############################
' Checks to see if a variable name is allowed
' name$ = the name to check
' returns $$TRUE if the name is valid otherwise $$FALSE
FUNCTION isVarNameValid (var$)
	IF var$ = "" THEN RETURN $$FALSE
	IF !isAlpha(var${0}) THEN RETURN $$FALSE
	FOR i = 1 TO UBOUND(var$)
		IF !isAlphaNum(var${i}) THEN RETURN $$FALSE
	NEXT
	SELECT CASE UCASE$(var$)
		CASE ""
			RETURN $$FALSE
		CASE "XOR", "OR", "AND", "COM", "PER", "RT", "NOT", "SQR", "CBR", "SIN", "COS", "TAN", "COT", "SEC", "COSEC", _
		"ASIN", "ACOS", "ATAN", "ACOT", "ASEC", "ACOSEC", "SINH", "COSH", "TANH", "COTH", "SECH", "COSECH", "ASINH", _
		"ACOSH", "ATANH", "ACOTH", "ASECH", "ACOSECH", "LOGE", "LN", "INT", "FRAC", "I", "E", "PI", "INCLUDE", "RUN", _
		"SET", "ANS", "LASTANS", "DIM", "MKV", "SELF", "CONJ", "RE", "IM", "EXP"
			RETURN $$FALSE
		CASE ELSE
			RETURN $$TRUE
	END SELECT
END FUNCTION
'
' ################################
' #####  SymbolTable_Update  #####
' ################################
' Updates a symbol in the symbol table
' name$ = the name of the symbol to update
' symbol = the new value for the symbol
' returns $$TRUE on success or $$FALSE on fail
FUNCTION SymbolTable_Update (name$, SYMBOL symbol)
	SHARED ASSOCARRAY symbolTable
	
	IF AssocArray_Find (symbolTable, UCASE$(name$), @iData) THEN
		SYMBOL_Update (iData, symbol)
		RETURN $$TRUE
	ELSE
		RETURN $$FALSE
	END IF
END FUNCTION
'
' ######################
' #####  evaluate  #####
' ######################
' Evaluates an expression
' expression = the expression to evaluate
' result = the variable to store the result
' type = the varaible to store the type of the result
' returns the number of errors generated
FUNCTION evaluate (LINKEDLIST expression, @exception, @result, @type, iThisFunc, iCase)
	SHARED CONSTANTS const
	SHARED CALCSTATE calcState
	SHARED STACK evalStack
	SHARED LINKEDLIST	runtimeParamStack
	
	FUNCADDR Delete1 (XLONG)
	FUNCADDR Delete2 (XLONG)
	
	TOKEN token
	SYMBOL symbol1
	SYMBOL symbol2
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	VECTOR vector1
	VECTOR vector2
	VECTOR vectorR
	FUNC func
	FUNCCASE case
	LINKEDLIST stack
	LINKEDLIST expr
	
	hWalk = LinkedList_StartWalk (expression)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		Delete1 = FUNCADDR(0)
		Delete2 = FUNCADDR(0)
		iData1 = 0
		iData2 = 0
		recursion = $$FALSE
		
		TOKEN_Get (iData, @token)
		'PRINT token.op
		
		SELECT CASE token.op
			CASE $$OP_LITERAL
				eval_push ($$TYPE_NUMBER, CopySymbol (token.value, $$TYPE_NUMBER))
			CASE $$OP_FUNCLIT
				IF runtimeParamStack.cItems THEN
					' This is an inner function, we'd better bind it with the closure op
					symbol1.type = $$TYPE_FUNCTION
					symbol1.data = op_closure (CopySymbol (token.value, $$TYPE_FUNCTION), 1)
					symbol1.delete = &Delete_Function ()
					
					FUNC_Get (symbol1.data, @func)
					
					' Append it to the parameter stack so it gets cleaned up
					closure = SYMBOL_New (symbol1)
					eval_push ($$TYPE_FUNCTION, symbol1.data)
					LinkedList_GetItem (@runtimeParamStack, 0, @iData)
					LINKEDLIST_Get (iData, @stack)
					LinkedList_Append (@stack, closure)
					LINKEDLIST_Update (iData, stack)
				ELSE
					' No need to copy here
					eval_push ($$TYPE_FUNCTION, token.value)
				END IF
			CASE $$OP_VAR, $$OP_BOUNDPARAM
				SYMBOL_Get (token.value, @symbol1)
				IF symbol1.type = $$TYPE_FUNCTION THEN
					' Don't copy functions
					eval_push (symbol1.type, symbol1.data)
				ELSE
					eval_push (symbol1.type, CopySymbol (symbol1.data, symbol1.type))
				END IF
			CASE $$OP_LASTANS
				SYMBOL_Get (calcState.iLastAns, @symbol1)
				eval_push (symbol1.type, CopySymbol (symbol1.data, symbol1.type))
			CASE $$OP_PARAMETER
				LinkedList_GetItem (runtimeParamStack, token.value2, @iData)
				LINKEDLIST_Get (iData, @stack)
				LinkedList_GetItem (stack, token.value, @iData)
				SYMBOL_Get (iData, @symbol1)
				IF symbol1.type = $$TYPE_FUNCTION THEN
					' Don't copy functions
					eval_push (symbol1.type, symbol1.data)
				ELSE
					eval_push (symbol1.type, CopySymbol (symbol1.data, symbol1.type))
				END IF
			CASE $$OP_RECURSE
				' push the special recursion function
				eval_push ($$TYPE_FUNCTION, 0)
			CASE $$OP_PI
				x = Allocate_Float ()
				__gmpf_set (x, const.pPi)
				eval_push ($$TYPE_NUMBER, x)
			CASE $$OP_E
				x = Allocate_Float ()
				__gmpf_set (x, const.pE)
				eval_push ($$TYPE_NUMBER, x)
			CASE $$OP_I
				complexR.real = Allocate_Float ()
				complexR.imaginary = Allocate_Float ()
				__gmpf_set_ui (complexR.real, 0)
				__gmpf_set_ui (complexR.imaginary, 1)
				eval_push ($$TYPE_COMPLEX, COMPLEX_New (complexR))
			CASE $$OP_RAND
				x = Allocate_Float ()
				__gmpf_urandomb (x, &calcState.randState, calculate_getPrecision ())
				eval_push ($$TYPE_NUMBER, x)
			CASE $$OP_FUNCCALL
				' Pop off the parameters
				LinkedList_Init (@stack)
				FOR i = 1 TO token.value
					IFF eval_pop (@type1, @iData1, @Delete1) THEN
						exception = $$EXCEPTION_INTERNAL
						EXIT FOR
					END IF
					
'					PRINT spaces$(#cSpaces);"PARAM: ";FormatSymbol$ (iData1, type1)
					symbol1.type = type1
					symbol1.data = iData1
					symbol1.delete = Delete1
					LinkedList_Insert (@stack, 0, SYMBOL_New (@symbol1))
				NEXT
				LinkedList_Insert (@runtimeParamStack, 0, LINKEDLIST_New (stack))
				
				' Pop off the function
				IFF eval_pop (@type2, @iData2, @Delete2) THEN
					exception = $$EXCEPTION_INTERNAL
				END IF
				IF type2 = $$TYPE_FUNCTION THEN
					IFZ iData2 THEN
						' PRINT "This is: ";iThisFunc
						' PRINT FormatSymbol$ (iThisFunc, $$TYPE_FUNCTION)
						iData2 = iThisFunc
						recursion = $$TRUE
					END IF
					FUNC_Get (iData2, @func)
					
					IF func.paramTypes.cItems != stack.cItems THEN
						exception = $$EXCEPTION_PARAMETERS
						Delete1 = FUNCADDR(0)
						Delete2 = FUNCADDR(0)
					ELSE
						' Check the parameter types
						hFWalk = LinkedList_StartWalk (func.paramTypes)
						hFWalk2 = LinkedList_StartWalk (stack)
						DO WHILE LinkedList_Walk (hFWalk, @expectedType) && LinkedList_Walk (hFWalk2, @iData)
							SYMBOL_Get (iData, @symbol1)
							actualType = symbol1.type
							SELECT CASE expectedType
								CASE $$TYPE_INT
									SELECT CASE actualType
										CASE $$TYPE_INT
										CASE $$TYPE_NUMBER
											IFF __gmpf_integer_p (symbol1.data) THEN exception = $$EXCEPTION_TYPES
										CASE $$TYPE_COMPLEX
											COMPLEX_Get (symbol1.data, @complex1)
											IFF (__gmpf_cmp_ui (complex1.imaginary, 0) = 0)&&(__gmpf_integer_p (complex1.real)) THEN exception = $$EXCEPTION_TYPES
										CASE ELSE
											exception = $$EXCEPTION_TYPES
									END SELECT
								CASE $$TYPE_NUMBER
									SELECT CASE actualType
										CASE $$TYPE_INT, $$TYPE_NUMBER
										CASE $$TYPE_COMPLEX
											COMPLEX_Get (symbol1.data, @complex1)
											IFF (__gmpf_cmp_ui (complex1.imaginary, 0) = 0) THEN exception = $$EXCEPTION_TYPES
										CASE ELSE
											exception = $$EXCEPTION_TYPES
									END SELECT
								CASE $$TYPE_COMPLEX
									IFF (actualType = $$TYPE_INT)||(actualType = $$TYPE_NUMBER)||(actualType = $$TYPE_COMPLEX) THEN exception = $$EXCEPTION_TYPES
								CASE $$TYPE_VECTOR
									IF actualType != $$TYPE_VECTOR THEN exception = $$EXCEPTION_TYPES
								CASE $$TYPE_FUNCTION
									IF actualType != $$TYPE_FUNCTION THEN exception = $$EXCEPTION_TYPES
							END SELECT
						LOOP
						LinkedList_EndWalk (hFWalk)
						LinkedList_EndWalk (hFWalk2)
						
						IF exception = $$EXCEPTION_NONE THEN
							IF recursion THEN
								IF LinkedList_IsLastNode (hWalk) THEN
									' Tail recursion - if the recursive call uses the same case we can optimise it
									exception = GetFuncCase (func, @expr, @iNextCase)
									IF exception = $$EXCEPTION_NONE THEN
										IF iNextCase = iCase THEN
											Delete_RuntimeParams (1)			' Delete the old parameters
											LinkedList_ResetWalk (hWalk)	' Loop back to the start
											DO DO
										ELSE
											' we can't optimise it, too bad
											GOSUB CallFunction
										END IF
									END IF
								ELSE
'									PRINT "RECURSION"
									exception = GetFuncCase (func, @expr, @iNextCase)
'									PRINT exception
									' This is definately not tail recursion
									GOSUB CallFunction
								END IF
							ELSE
								' Evaluate the function using the appropriate case
								exception = GetFuncCase (func, @expr, @iNextCase)
								IF exception = $$EXCEPTION_NONE THEN
									GOSUB CallFunction
								END IF
							END IF
							
							' We have already deleted all the parameters
							Delete1 = FUNCADDR(0)
							Delete2 = FUNCADDR(0)
						END IF
					END IF
				ELSE
					IF token.value = 1 THEN
						' Treat this as multiplication
						op_mul (iData1, type1, iData2, type2)
					END IF
				END IF
				
			CASE $$OP_MAKEVECTOR
				LinkedList_Init (@vectorR.values)
				vectorR.type = $$TYPE_INVALID
				vectorR.size = token.value
				FOR i = 1 TO token.value
					@Delete1 (iData1)
					IFF eval_pop (@type1, @iData1, @Delete1) THEN
						exception = $$EXCEPTION_INTERNAL
						EXIT FOR
					END IF
					
					' Make sure the element types are compatible
					SELECT CASE vectorR.type
						CASE $$TYPE_INT
							SELECT CASE type1
								CASE $$TYPE_INT
								CASE $$TYPE_NUMBER, $$TYPE_COMPLEX, $$TYPE_INVALID
									vectorR.type = type1
								CASE ELSE
									exception = $$EXCEPTION_TYPES
									EXIT FOR
							END SELECT
						CASE $$TYPE_NUMBER
							SELECT CASE type1
								CASE $$TYPE_INT, $$TYPE_NUMBER
								CASE $$TYPE_COMPLEX, $$TYPE_INVALID
									vectorR.type = type1
								CASE ELSE
									exception = $$EXCEPTION_TYPES
									EXIT FOR
							END SELECT
						CASE $$TYPE_COMPLEX
							SELECT CASE type1
								CASE $$TYPE_INT, $$TYPE_NUMBER, $$TYPE_COMPLEX
								CASE $$TYPE_INVALID
									vectorR.type = type1
								CASE ELSE
									exception = $$EXCEPTION_TYPES
									EXIT FOR
							END SELECT
					END SELECT
					
					symbol1.type = type1
					SELECT CASE type1
						CASE $$TYPE_NUMBER
							symbol1.data = CopySymbol (iData1, type1)
							symbol1.delete = &Delete_Float ()
						CASE $$TYPE_COMPLEX
							symbol1.data = CopySymbol (iData1, type1)
							symbol1.delete = &Delete_Complex ()
						CASE ELSE
							exception = $$EXCEPTION_TYPES
							EXIT FOR
					END SELECT
					
					LinkedList_Insert (@vectorR.values, 0, SYMBOL_New (symbol1))
				NEXT
				
				eval_push ($$TYPE_VECTOR, VECTOR_New (vectorR))
			CASE $$OP_VECTORDIM
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
'					PRINT "TYPE1: ";type1, iData1
					IF type1 = $$TYPE_VECTOR THEN
						VECTOR_Get (iData1, @vector1)
						x = Allocate_Float ()
						__gmpf_set_si (x, vector1.size)
						eval_push ($$TYPE_NUMBER, x)
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				END IF
			CASE $$OP_GENVEC
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					IF __gmpf_integer_p (iData2) THEN
						vectorR.size = __gmpf_get_si (iData2)
						vectorR.type = $$TYPE_INVALID
						
						IF vectorR.size > 0 THEN 
							SELECT CASE type1
								CASE $$TYPE_NUMBER, $$TYPE_COMPLEX
									vectorR.type = type1
									symbol1.type = type1
									IF type1 = $$TYPE_NUMBER THEN symbol1.delete = &Delete_Float () ELSE symbol1.delete = &Delete_Complex ()
									
									LinkedList_Init (@vectorR.values)
									FOR i = 1 TO vectorR.size
										symbol1.data = CopySymbol (iData1, type1)
										LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
									NEXT
									
									eval_push ($$TYPE_VECTOR, VECTOR_New (vectorR))
								CASE $$TYPE_FUNCTION
									LinkedList_Init (@expr)
									
									' Build the expression that calls the function
									symbol2.type = $$TYPE_FUNCTION
									symbol2.delete = &Delete_Function ()
									symbol2.data = iData1
									function = SYMBOL_New (symbol2)
									
									token.op = $$OP_VAR
									token.value = function
									
									' prepare the parameter tempory variable
									symbol2.type = $$TYPE_NUMBER
									symbol2.delete = &Delete_Float ()
									symbol2.data = Allocate_Float ()
									parameter = SYMBOL_New (symbol2)
									
									' Continue building the expression
									LinkedList_Append (@expr, TOKEN_New(token))
									token.op = $$OP_VAR
									token.value = parameter
									LinkedList_Append (@expr, TOKEN_New(token))
									token.op = $$OP_FUNCCALL
									token.value = 1
									LinkedList_Append (@expr, TOKEN_New(token))
									
									' GO!
									LinkedList_Init (@vectorR.values)
									FOR i = 1 TO vectorR.size
										__gmpf_set_si (symbol2.data, i)
										
										evaluate (expr, @exception, @iData, @type, iThisFunc, iCase)
										symbol1.type = type
										IF type = $$TYPE_NUMBER THEN symbol1.delete = &Delete_Float () ELSE symbol1.delete = &Delete_Complex ()
										
										IF exception != $$EXCEPTION_NONE THEN
											LinkedList_DeleteAll (@vectorR.values)
											LinkedList_Uninit (@vectorR.values)
											EXIT FOR
										END IF
										
										SELECT CASE type
											CASE vectorR.type
												symbol1.data = iData
												LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
											CASE $$TYPE_NUMBER
												IF vectorR.type = $$TYPE_INVALID THEN vectorR.type = $$TYPE_NUMBER
												symbol1.data = iData
												LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
											CASE $$TYPE_COMPLEX
												IF vectorR.type = $$TYPE_INVALID THEN vectorR.type = $$TYPE_COMPLEX
												symbol1.data = iData
												LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
											CASE ELSE
												exception = $$EXCEPTION_TYPES
										END SELECT
									NEXT
									
									eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
									
									Delete_Expression (expr)
									Delete_Float (symbol2.data)
									SYMBOL_Delete (parameter)
									SYMBOL_Delete (function)
								CASE ELSE
									exception = $$EXCEPTION_TYPES
							END SELECT
						ELSE
							exception = $$EXCEPTION_UNDEFINED
						END IF
					END IF
				END IF
			CASE $$OP_VECCOM
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_veccom (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_CONCAT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_concat (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_ADD
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_add (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_SUB
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sub (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_OR
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_or (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_XOR
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_xor (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_MUL
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_mul (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_DIV
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_div (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_UPSHIFT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_upshift (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_DOWNSHIFT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_downshift (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_AND
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_and (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_C
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_c (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_P
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_p (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_DOTPROD
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_dotprod (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_RAISE
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_raise (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_ROOT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_root (iData1, type1, iData2, type2)
				END IF
			CASE $$OP_NEGATE
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_negate (iData1, type1)
				END IF
			CASE $$OP_FACT
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_fact (iData1, type1)
				END IF
			CASE $$OP_NOT
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_not (iData1, type1)
				END IF
			CASE $$OP_SQR
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sqr (iData1, type1)
				END IF
			CASE $$OP_CBR
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cbr (iData1, type1)
				END IF
			CASE $$OP_SIN
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sin (iData1, type1)
				END IF
			CASE $$OP_COS
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cos (iData1, type1)
				END IF
			CASE $$OP_TAN
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_tan (iData1, type1)
				END IF
			CASE $$OP_COT
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cot (iData1, type1)
				END IF
			CASE $$OP_SEC
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sec (iData1, type1)
				END IF
			CASE $$OP_COSEC
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cosec (iData1, type1)
				END IF
			CASE $$OP_ASIN
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_asin (iData1, type1)
				END IF
			CASE $$OP_ACOS
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acos (iData1, type1)
				END IF
			CASE $$OP_ATAN
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_atan (iData1, type1)
				END IF
			CASE $$OP_ACOT
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acot (iData1, type1)
				END IF
			CASE $$OP_ASEC
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_asec (iData1, type1)
				END IF
			CASE $$OP_ACOSEC
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acosec (iData1, type1)
				END IF
			CASE $$OP_SINH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sinh (iData1, type1)
				END IF
			CASE $$OP_COSH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cosh (iData1, type1)
				END IF
			CASE $$OP_TANH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_tanh (iData1, type1)
				END IF
			CASE $$OP_COTH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_coth (iData1, type1)
				END IF
			CASE $$OP_SECH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_sech (iData1, type1)
				END IF
			CASE $$OP_COSECH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_cosech (iData1, type1)
				END IF
			CASE $$OP_ASINH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_asinh (iData1, type1)
				END IF
			CASE $$OP_ACOSH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acosh (iData1, type1)
				END IF
			CASE $$OP_ATANH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_atanh (iData1, type1)
				END IF
			CASE $$OP_ACOTH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acoth (iData1, type1)
				END IF
			CASE $$OP_ASECH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_asech (iData1, type1)
				END IF
			CASE $$OP_ACOSECH
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_acosech (iData1, type1)
				END IF
			CASE $$OP_INT
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_int (iData1, type1)
				END IF
			CASE $$OP_FRAC
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_frac (iData1, type1)
				END IF
			CASE $$OP_LOGE
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_ln (iData1, type1)
				END IF
			CASE $$OP_EXP
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_exp (iData1, type1)
				END IF
			CASE $$OP_RE
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_re (iData1, type1)
				END IF
			CASE $$OP_IM
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_im (iData1, type1)
				END IF
			CASE $$OP_CONJ
				IFF eval_pop (@type1, @iData1, @Delete1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					exception = op_conj (iData1, type1)
				END IF
			CASE $$OP_SHORTC_T
				IFF eval_peek (@type1, @iData1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					IFF type1 = $$TYPE_BOOL THEN
						exception = $$EXCEPTION_TYPES
					ELSE
						IF iData1 THEN 
							' skip untill we find our other half
							shortc_id = token.value
							DO WHILE LinkedList_Walk (hWalk, @iData)
								TOKEN_Get(iData, @token)
								IF (token.value = shortc_id)&&(token.op = $$OP_LOGOR) THEN EXIT DO
							LOOP
						END IF
					END IF
				END IF
			CASE $$OP_SHORTC_F
				IFF eval_peek (@type1, @iData1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					IFF type1 = $$TYPE_BOOL THEN
						exception = $$EXCEPTION_TYPES
					ELSE
						IFF iData1 THEN 
							' skip untill we find our other half
							shortc_id = token.value
							DO WHILE LinkedList_Walk (hWalk, @iData)
								TOKEN_Get(iData, @token)
								IF (token.value = shortc_id)&&(token.op = $$OP_LOGAND) THEN EXIT DO
							LOOP
						END IF
					END IF
				END IF
			CASE $$OP_LOGAND
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					IFF (type1 = $$TYPE_BOOL)&&(type2 = $$TYPE_BOOL) THEN
						exception = $$EXCEPTION_TYPES
					ELSE
						IF iData1 && iData2 THEN eval_push ($$TYPE_BOOL, $$TRUE) ELSE eval_push ($$TYPE_BOOL, $$FALSE)
					END IF
				END IF
			CASE $$OP_LOGOR
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					IFF (type1 = $$TYPE_BOOL)&&(type2 = $$TYPE_BOOL) THEN
						exception = $$EXCEPTION_TYPES
					ELSE
						IF iData1 || iData2 THEN eval_push ($$TYPE_BOOL, $$TRUE) ELSE eval_push ($$TYPE_BOOL, $$FALSE)
					END IF
				END IF
			CASE $$OP_RELEQ
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					eval_push($$TYPE_BOOL, op_eq (iData1, type1, iData2, type2, @exception))
				END IF
			CASE $$OP_RELNOTEQ
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					result = op_eq (iData1, type1, iData2, type2, @exception)
					IF result THEN result = $$FALSE ELSE result = $$TRUE
					eval_push($$TYPE_BOOL, result)
				END IF
			CASE $$OP_RELGT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					result = op_gt (iData1, type1, iData2, type2, @exception)
					eval_push($$TYPE_BOOL, result)
				END IF
			CASE $$OP_RELLTE
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					result = op_gt (iData1, type1, iData2, type2, @exception)
					IF result THEN result = $$FALSE ELSE result = $$TRUE
					eval_push($$TYPE_BOOL, result)
				END IF
			CASE $$OP_RELLT
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					eval_push($$TYPE_BOOL, op_lt (iData1, type1, iData2, type2, @exception))
				END IF
			CASE $$OP_RELGTE
				IFF (eval_pop (@type2, @iData2, @Delete2) && eval_pop (@type1, @iData1, @Delete1)) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					result = op_lt (iData1, type1, iData2, type2, @exception)
					IF result THEN result = $$FALSE ELSE result = $$TRUE
					eval_push($$TYPE_BOOL, result)
				END IF
			CASE $$CMD_STOR
				IFF eval_peek (@type1, @iData1) THEN
					exception = $$EXCEPTION_INTERNAL
				ELSE
					STRING_Get (token.value, @name$)
					iSymbol = SymbolTable_Retreive (name$, @symbol1)
					IF iSymbol THEN 
						' the symbol is already defined, delete the old version
						SymbolTable_Delete (name$)
					END IF
					
					' The new symbol is a copy of the TOS
					symbol1.type = type1
					symbol1.data = CopySymbol (iData1, type1)
					SELECT CASE type1
						CASE $$TYPE_BOOL
							symbol1.delete = FUNCADDR(0)
						CASE $$TYPE_NUMBER
							symbol1.delete = &Delete_Float ()
						CASE $$TYPE_COMPLEX
							symbol1.delete = &Delete_Complex ()
						CASE $$TYPE_VECTOR
							symbol1.delete = &Delete_Vector ()
						CASE $$TYPE_FUNCTION
							symbol1.delete = &Delete_Function ()
					END SELECT
					SymbolTable_Insert (name$, symbol1)
				END IF
				
			CASE $$CMD_CLEAR
				STRING_Get (token.value, @name$)
				SymbolTable_Delete (name$)
		END SELECT
		@Delete1 (iData1)
		@Delete2 (iData2)
		
		SELECT CASE exception
			CASE $$EXCEPTION_INTERNAL
				AddError ("[Internal error] Underflow detected")
				GOSUB CleanStack
				RETURN 1
			CASE $$EXCEPTION_TYPES
				AddError ("Mismatched data types")
				GOSUB CleanStack
				RETURN 1
			CASE $$EXCEPTION_PARAMETERS
				AddError ("Incorrect number of parameters")
				GOSUB CleanStack
				RETURN 1
			CASE $$EXCEPTION_UNDEFINED
				GOSUB CleanStack
				RETURN 1
		END SELECT
	LOOP
	LinkedList_EndWalk (hWalk)
	
	IFF eval_pop (@type, @result, FUNCADDR(0)) THEN
		exception = $$EXCEPTION_INTERNAL
		AddError ("[Internal error] Underflow detected")
		INC cErrors
	END IF
	
	IF type = $$TYPE_FUNCTION THEN
		' this is a closure and it's about to be deleted, so copy it
		result = CopySymbol (result, $$TYPE_FUNCTION)
	END IF
	
	RETURN cErrors
	
	SUB CleanStack
		DO WHILE eval_pop (@cs_type, @iData1, @Delete1)
			@Delete1 (iData1)
		LOOP
		
		DO WHILE Delete_RuntimeParams (0)
		LOOP
	END SUB
	SUB CallFunction
'		? spaces$(#cSpaces);"CALLING FUNCTION: ";expr.iHead,iData2
'		INC #cSpaces
		evaluate (expr, @exception, @iData, @type, iData2, iNextCase)
		eval_push (type, iData)
'		? spaces$(#cSpaces);"PUSHED: ", type, iData, FormatSymbol$ (iData, type)
'		DEC #cSpaces
		Delete_RuntimeParams (0)
	END SUB
END FUNCTION
'
' #######################
' #####  eval_push  #####
' #######################
' Pushes a symbol onto the stack
' type = the type of the symbol
' iData = the index/address for the symbol data
' returns $$TRUE on success or $$FALSE on fail
FUNCTION eval_push (type, iData)
	SHARED STACK evalStack
	SYMBOL symbol
	
	symbol.type = type
	symbol.data = iData
	SELECT CASE type
		CASE $$TYPE_NUMBER, $$TYPE_INT
			symbol.delete = &Delete_Float ()
		CASE $$TYPE_COMPLEX
			symbol.delete = &Delete_Complex ()
		CASE $$TYPE_VECTOR
			symbol.delete = &Delete_Vector ()
		CASE $$TYPE_FUNCTION
			symbol.delete = &Delete_Function ()
		CASE $$TYPE_BOOL
			symbol.delete = FUNCADDR(0)
	END SELECT
	
	ret = SYMBOL_New (symbol)

	RETURN Stack_Push (@evalStack, ret)
END FUNCTION
'
' ######################
' #####  eval_pop  #####
' ######################
' Pops a symbol off the stack
' type = the variable to store the type
' iData = the variable to store the index/address for the symbol data
' returns $$TRUE on success or $$FALSE on fail
FUNCTION eval_pop (@type, @iData, FUNCADDR Delete)
	SHARED STACK evalStack
	SYMBOL symbol
	
	IFF Stack_Pop (@evalStack, @ret) THEN RETURN $$FALSE
	IFF SYMBOL_Get (ret, @symbol) THEN RETURN $$FALSE
	type = symbol.type
	iData = symbol.data
	Delete = symbol.delete
	SYMBOL_Delete (ret)
	
	RETURN $$TRUE
END FUNCTION
'
' #######################
' #####  eval_peek  #####
' #######################
' Same as pop but doesn't remove the data from the stack
FUNCTION eval_peek (@type, @iData)
	SHARED STACK evalStack
	SYMBOL symbol
	
	IFF Stack_Peek (evalStack, @ret) THEN RETURN $$FALSE
	IFF SYMBOL_Get (ret, @symbol) THEN RETURN $$FALSE
	type = symbol.type
	iData = symbol.data
	
	RETURN $$TRUE
END FUNCTION
'
' ##########################
' #####  Delete_Float  #####
' ##########################
' Frees a floating point number correctly
' pNumber = the number to delete
' returns $$TRUE on success or $$FALSE on fail
FUNCTION Delete_Float (pNumber)
	'PRINT "Deallocating: ";pNumber
	__gmpf_clear (pNumber)
	IFZ LocalFree (pNumber) THEN RETURN $$TRUE ELSE RETURN $$FALSE
END FUNCTION
'
' ############################
' #####  Allocate_Float  #####
' ############################
' Allocates a new floating point number
' returns a pointer to the number
FUNCTION Allocate_Float ()
	ret = LocalAlloc ($$LMEM_FIXED, SIZE(mpf_t))
	'PRINT "Allocating: ";ret
	__gmpf_init (ret)
	RETURN ret
END FUNCTION


' The operator implementations
FUNCTION op_add (iData1, type1, iData2, type2)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	VECTOR vector1
	VECTOR vector2
	VECTOR vectorR
	SYMBOL symbol1
	SYMBOL symbol2
	FUNCADDR Delete (XLONG)
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					x = Allocate_Float ()
					__gmpf_add (x, iData1, iData2)
					eval_push ($$TYPE_NUMBER, x)
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_add (complexR.real, iData1, complex2.real)
					__gmpf_set (complexR.imaginary, complex2.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_add (complexR.real, complex1.real, iData2)
					__gmpf_set (complexR.imaginary, complex1.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_add (complexR.real, complex1.real, complex2.real)
					__gmpf_add (complexR.imaginary, complex1.imaginary, complex2.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_VECTOR
			IF type2 != $$TYPE_VECTOR THEN RETURN $$EXCEPTION_TYPES
			
			exception = $$EXCEPTION_NONE
			VECTOR_Get (iData1, @vector1)
			VECTOR_Get (iData2, @vector2)
			
			IF vector1.size != vector2.size THEN RETURN $$EXCEPTION_UNDEFINED 

			LinkedList_Init (@vectorR.values)
			vectorR.size = vector1.size

			hV1Walk = LinkedList_StartWalk (vector1.values)
			hV2Walk = LinkedList_StartWalk (vector2.values)
			DO WHILE LinkedList_Walk (hV1Walk, @iV1Data) && LinkedList_Walk (hV2Walk, @iV2Data)
				SYMBOL_Get (iV1Data, @symbol1)
				SYMBOL_Get (iV2Data, @symbol2)
				exception_t = op_add (symbol1.data, symbol1.type, symbol2.data, symbol2.type)
				IF exception_t = $$EXCEPTION_NONE THEN
					eval_pop (@type, @iData, @Delete)
					vectorR.type = type
					symbol1.type = type
					symbol1.data = CopySymbol (iData, type)
					symbol1.delete = Delete
					LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
					@Delete (iData)
				ELSE
					exception = exception_t
				END IF
			LOOP
			LinkedList_EndWalk (hV1Walk)
			LinkedList_EndWalk (hV2Walk)
			
			eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
			RETURN exception
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sub (iData1, type1, iData2, type2)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	VECTOR vector1
	VECTOR vector2
	VECTOR vectorR
	SYMBOL symbol1
	SYMBOL symbol2
	FUNCADDR Delete (XLONG)
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					x = Allocate_Float ()
					__gmpf_sub (x, iData1, iData2)
					eval_push ($$TYPE_NUMBER, x)
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_sub (complexR.real, iData1, complex2.real)
					__gmpf_neg (complexR.imaginary, complex2.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_sub (complexR.real, complex1.real, iData2)
					__gmpf_neg (complexR.imaginary, complex1.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_sub (complexR.real, complex1.real, complex2.real)
					__gmpf_sub (complexR.imaginary, complex1.imaginary, complex2.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
			
		CASE $$TYPE_VECTOR
			IF type2 != $$TYPE_VECTOR THEN RETURN $$EXCEPTION_TYPES
			
			exception = $$EXCEPTION_NONE
			VECTOR_Get (iData1, @vector1)
			VECTOR_Get (iData2, @vector2)
			IF vector1.size != vector2.size THEN RETURN $$EXCEPTION_UNDEFINED 
			
			LinkedList_Init (@vectorR.values)
			vectorR.size = vector1.size
			
			hV1Walk = LinkedList_StartWalk (vector1.values)
			hV2Walk = LinkedList_StartWalk (vector2.values)
			DO WHILE LinkedList_Walk (hV1Walk, @iV1Data) && LinkedList_Walk (hV2Walk, @iV2Data)
				SYMBOL_Get (iV1Data, @symbol1)
				SYMBOL_Get (iV2Data, @symbol2)
				exception_t = op_sub (symbol1.data, symbol1.type, symbol2.data, symbol2.type)
				IF exception_t = $$EXCEPTION_NONE THEN
					eval_pop (@type, @iData, @Delete)
					vectorR.type = type
					symbol1.type = type
					symbol1.data = CopySymbol (iData, type)
					symbol1.delete = Delete
					LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
					@Delete (iData)
				ELSE
					exception = exception_t
				END IF
			LOOP
			LinkedList_EndWalk (hV1Walk)
			LinkedList_EndWalk (hV2Walk)
			
			eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
			RETURN exception
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_mul (iData1, type1, iData2, type2)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	VECTOR vector1
	VECTOR vectorR
	SYMBOL symbol1
	FUNCADDR Delete (XLONG)
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					x = Allocate_Float ()
					__gmpf_mul (x, iData1, iData2)
					eval_push ($$TYPE_NUMBER, x)
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_mul (complexR.real, iData1, complex2.real)
					__gmpf_mul (complexR.imaginary, iData1, complex2.imaginary)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_VECTOR
					IF type1 = $$TYPE_NUMBER || type1 = $$TYPE_COMPLEX THEN
						SWAP iData1, iData2
						SWAP type1, type2
						GOSUB MulVector
						RETURN exception
					ELSE
						RETURN $$EXCEPTION_TYPES
					END IF
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_mul (complexR.real, complex1.real, iData2)
					__gmpf_mul (complexR.imaginary, complex1.imaginary, iData2)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					t0 = Allocate_Float ()
					
					__gmpf_mul (complexR.real, complex1.real, complex2.real)
					__gmpf_mul (t0, complex1.imaginary, complex2.imaginary)
					__gmpf_sub (complexR.real, complexR.real, t0)
					
					__gmpf_mul (complexR.imaginary, complex1.real, complex2.imaginary)
					__gmpf_mul (t0, complex1.imaginary, complex2.real)
					__gmpf_add (complexR.imaginary, complexR.imaginary, t0)
					
					Delete_Float (t0)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_VECTOR
					IF type1 = $$TYPE_NUMBER || type1 = $$TYPE_COMPLEX THEN
						SWAP iData1, iData2
						SWAP type1, type2
						GOSUB MulVector
						RETURN exception
					ELSE
						RETURN $$EXCEPTION_TYPES
					END IF
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_VECTOR
			IF type2 = $$TYPE_NUMBER || type2 = $$TYPE_COMPLEX THEN
				GOSUB MulVector
				RETURN exception
			ELSE
				RETURN $$EXCEPTION_TYPES
			END IF
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
	
	SUB MulVector
		exception = $$EXCEPTION_NONE
		VECTOR_Get (iData1, @vector1)
		
		LinkedList_Init (@vectorR.values)
		vectorR.size = vector1.size
		
		hV1Walk = LinkedList_StartWalk (vector1.values)
		DO WHILE LinkedList_Walk (hV1Walk, @iV1Data)
			SYMBOL_Get (iV1Data, @symbol1)
			exception_t = op_mul (symbol1.data, symbol1.type, iData2, type2)
			IF exception_t = $$EXCEPTION_NONE THEN
				eval_pop (@type, @iData, @Delete)
				vectorR.type = type
				symbol1.type = type
				symbol1.data = CopySymbol (iData, type)
				symbol1.delete = Delete
				LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
				@Delete (iData)
			ELSE
				exception = exception_t
			END IF
		LOOP
		LinkedList_EndWalk (hV1Walk)
		
		eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
	END SUB
END FUNCTION

FUNCTION op_div (iData1, type1, iData2, type2)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	VECTOR vector1
	VECTOR vectorR
	SYMBOL symbol1
	FUNCADDR Delete (XLONG)
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IF __gmpf_cmp_ui (iData2, 0) = 0 THEN RETURN $$EXCEPTION_UNDEFINED
					x = Allocate_Float ()
					__gmpf_div (x, iData1, iData2)
					eval_push ($$TYPE_NUMBER, x)
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					
					den = Allocate_Float ()
					t0 = Allocate_Float ()
					GOSUB ComputeDenominator
					
					__gmpf_mul (complexR.real, iData1, complex2.real)
					__gmpf_div (complexR.real, complexR.real, den)
					
					__gmpf_mul (complexR.imaginary, iData1, complex2.imaginary)
					__gmpf_neg (complexR.imaginary, complexR.imaginary)
					__gmpf_div (complexR.real, complexR.real, den)
					
					Delete_Float (den)
					Delete_Float (t0)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					__gmpf_div (complexR.real, complex1.real, iData2)
					__gmpf_div (complexR.imaginary, complex1.imaginary, iData2)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					complexR.real = Allocate_Float ()
					complexR.imaginary = Allocate_Float ()
					
					den = Allocate_Float ()
					t0 = Allocate_Float ()
					GOSUB ComputeDenominator
					
					__gmpf_mul (complexR.real, complex1.real, complex2.real)
					__gmpf_mul (t0, complex1.imaginary, complex2.imaginary)
					__gmpf_add (complexR.real, complexR.real, t0)
					__gmpf_div (complexR.real, complexR.real, den)
					
					__gmpf_mul (complexR.imaginary, complex1.imaginary, complex2.real)
					__gmpf_mul (t0, complex1.real, complex2.imaginary)
					__gmpf_sub (complexR.imaginary, complexR.imaginary, t0)
					__gmpf_div (complexR.imaginary, complexR.imaginary, den)
					
					Delete_Float (den)
					Delete_Float (t0)
					eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_VECTOR
			IF type2 = $$TYPE_NUMBER || type2 = $$TYPE_COMPLEX THEN
				exception = $$EXCEPTION_NONE
				VECTOR_Get (iData1, @vector1)
				
				LinkedList_Init (@vectorR.values)
				vectorR.size = vector1.size
				
				hV1Walk = LinkedList_StartWalk (vector1.values)
				DO WHILE LinkedList_Walk (hV1Walk, @iV1Data)
					SYMBOL_Get (iV1Data, @symbol1)
					exception_t = op_div (symbol1.data, symbol1.type, iData2, type2)
					IF exception_t = $$EXCEPTION_NONE THEN
						eval_pop (@type, @iData, @Delete)
						vectorR.type = type
						symbol1.type = type
						symbol1.data = CopySymbol (iData, type)
						symbol1.delete = Delete
						LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
						@Delete (iData)
					ELSE
						exception = exception_t
					END IF
				LOOP
				LinkedList_EndWalk (hV1Walk)
				
				eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
				
				RETURN exception
			ELSE
				RETURN $$EXCEPTION_TYPES
			END IF
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
	
	SUB ComputeDenominator
		__gmpf_mul (den, complex2.real, complex2.real)
		__gmpf_mul (t0, complex2.imaginary, complex2.imaginary)
		__gmpf_add (den, den, t0)
		
		IF __gmpf_cmp_ui (den, 0) = 0 THEN
			Delete_Float (den)
			Delete_Float (t0)
			RETURN $$EXCEPTION_UNDEFINED
		END IF
	END SUB
END FUNCTION

FUNCTION op_or (iData1, type1, iData2, type2)
	mpz_t int1R
	mpz_t int2R
	int1 = &int1R
	int2 = &int2R
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData1) && __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					__gmpz_init (int1)
					__gmpz_init (int2)
					__gmpz_set_f (int1, iData1)
					__gmpz_set_f (int2, iData2)
					
					x = Allocate_Float ()
					__gmpz_ior (int1, int1, int2)
					__gmpf_set_z (x, int1)
					eval_push ($$TYPE_NUMBER, x)
					
					__gmpz_clear (int1)
					__gmpz_clear (int2)
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_xor (iData1, type1, iData2, type2)
	mpz_t int1R
	mpz_t int2R
	int1 = &int1R
	int2 = &int2R
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData1) && __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					__gmpz_init (int1)
					__gmpz_init (int2)
					__gmpz_set_f (int1, iData1)
					__gmpz_set_f (int2, iData2)
					
					x = Allocate_Float ()
					__gmpz_xor (int1, int1, int2)
					__gmpf_set_z (x, int1)
					eval_push ($$TYPE_NUMBER, x)
					
					__gmpz_clear (int1)
					__gmpz_clear (int2)
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_upshift (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					x = Allocate_Float ()
					__gmpf_mul_2exp (x, iData1, __gmpf_get_ui (iData2))
					eval_push ($$TYPE_NUMBER, x)
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_downshift (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					x = Allocate_Float ()
					__gmpf_div_2exp (x, iData1, __gmpf_get_ui (iData2))
					eval_push ($$TYPE_NUMBER, x)
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_and (iData1, type1, iData2, type2)
	mpz_t int1R
	mpz_t int2R
	int1 = &int1R
	int2 = &int2R
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData1) && __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					__gmpz_init (int1)
					__gmpz_init (int2)
					__gmpz_set_f (int1, iData1)
					__gmpz_set_f (int2, iData2)
					
					x = Allocate_Float ()
					__gmpz_and (int1, int1, int2)
					__gmpf_set_z (x, int1)
					eval_push ($$TYPE_NUMBER, x)
					
					__gmpz_clear (int1)
					__gmpz_clear (int2)
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_c (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData1) && __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					IF (__gmpf_cmp_ui (iData1, 0) <= 0)||(__gmpf_cmp_ui (iData2, 0) < 0) THEN RETURN $$EXCEPTION_UNDEFINED
					IF (__gmpf_cmp_ui (iData2, 0) = 0) THEN
						x = Allocate_Float ()
						__gmpf_set_ui (x, 1)
						eval_push ($$TYPE_NUMBER, x)
					ELSE
						t0 = Allocate_Float ()
						x = Allocate_Float ()
						__gmpf_set (x, iData1)
						start = __gmpf_get_ui (iData1)-1
						end = start - __gmpf_get_ui (iData2) + 2
						FOR k = start TO end STEP -1
							__gmpf_mul_ui (x, x, k)
						NEXT
						__gmpf_set (t0, iData2)
						start = __gmpf_get_ui (iData2)-1
						FOR k = start TO 1 STEP -1
							__gmpf_mul_ui (t0, t0, k)
						NEXT
						__gmpf_div (x, x, t0)
						Delete_Float (t0)
						eval_push ($$TYPE_NUMBER, x)
					END IF
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_p (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					IFF __gmpf_integer_p (iData1) && __gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_UNDEFINED
					IF (__gmpf_cmp_ui (iData1, 0) <= 0)||(__gmpf_cmp_ui (iData2, 0) < 0) THEN RETURN $$EXCEPTION_UNDEFINED
					IF (__gmpf_cmp_ui (iData2, 0) = 0) THEN
						x = Allocate_Float ()
						__gmpf_set_ui (x, 1)
						eval_push ($$TYPE_NUMBER, x)
					ELSE
						x = Allocate_Float ()
						__gmpf_set (x, iData1)
						start = __gmpf_get_ui (iData1)-1
						end = start - __gmpf_get_ui (iData2) + 2
						FOR k = start TO end STEP -1
							__gmpf_mul_ui (x, x, k)
						NEXT
						eval_push ($$TYPE_NUMBER, x)
					END IF
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_raise (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					x = Allocate_Float ()
					undefined = TransLib_Power (x, iData1, iData2)
					eval_push ($$TYPE_NUMBER, x)
					IF undefined THEN RETURN $$EXCEPTION_UNDEFINED
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_root (iData1, type1, iData2, type2)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_NUMBER
					check = $$FALSE
					x = Allocate_Float ()
					IF __gmpf_cmp_ui (iData1, 2) THEN
						IF __gmpf_cmp_ui(iData1, 0) < 0 THEN check = $$TRUE ELSE __gmpf_sqrt(x, iData2)
					ELSE
						__gmpf_ui_div (x, 1, iData1)
						check = TransLib_Power (x, iData2, x)
					END IF
					eval_push ($$TYPE_NUMBER, x)
					IF check THEN RETURN $$EXCEPTION_UNDEFINED
				CASE ELSE
					RETURN $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_negate (iData1, type1)
	COMPLEX complex1
	COMPLEX complexR
	
	VECTOR vector1
	VECTOR vectorR
	SYMBOL symbol1
	FUNCADDR Delete (XLONG)
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_neg (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData1, @complex1)
			complexR.real = Allocate_Float ()
			complexR.imaginary = Allocate_Float ()
			__gmpf_neg (complexR.real, complex1.real)
			__gmpf_neg (complexR.imaginary, complex1.imaginary)
			eval_push ($$TYPE_COMPLEX, COMPLEX_New (complexR))
		CASE $$TYPE_VECTOR
			exception = $$EXCEPTION_NONE
			VECTOR_Get (iData1, @vector1)
			
			LinkedList_Init (@vectorR.values)
			vectorR.size = vector1.size
			
			hV1Walk = LinkedList_StartWalk (vector1.values)
			DO WHILE LinkedList_Walk (hV1Walk, @iV1Data)
				SYMBOL_Get (iV1Data, @symbol1)
				exception_t = op_negate (symbol1.data, symbol1.type)
				IF exception_t = $$EXCEPTION_NONE THEN
					eval_pop (@type, @iData, @Delete)
					vectorR.type = type
					symbol1.type = type
					symbol1.data = CopySymbol (iData, type)
					symbol1.delete = Delete
					LinkedList_Append (@vectorR.values, SYMBOL_New (symbol1))
					@Delete (iData)
				ELSE
					exception = exception_t
				END IF
			LOOP
			LinkedList_EndWalk (hV1Walk)
			
			eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
			
			RETURN exception
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_fact (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			IFF __gmpf_integer_p (iData1) THEN RETURN $$EXCEPTION_UNDEFINED
			SELECT CASE TRUE
				CASE __gmpf_cmp_ui (iData1, 0) < 0
					RETURN $$EXCEPTION_UNDEFINED
				CASE __gmpf_cmp_ui (iData1, 0) = 0
					x = Allocate_Float ()
					__gmpf_set_ui (x, 1)
				CASE __gmpf_cmp_ui (iData1, 1) = 0
					x = Allocate_Float ()
					__gmpf_set_ui (x, 1)
				CASE ELSE
					x = Allocate_Float ()
					__gmpf_set_ui (x, 1)
					intMax = __gmpf_get_si (iData1)
					FOR i = 2 TO intMax
						__gmpf_mul_ui (x, x, i)
					NEXT
			END SELECT
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_not (iData1, type1)
	mpz_t int1R
	int1 = &int1R
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			IFF __gmpf_integer_p (iData1) THEN RETURN $$EXCEPTION_UNDEFINED
			__gmpz_init (int1)
			__gmpz_set_f (int1, iData1)
			
			x = Allocate_Float ()
			__gmpz_com (int1, int1)
			__gmpf_set_z (x, int1)
			eval_push ($$TYPE_NUMBER, x)
			
			__gmpz_clear (int1)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sqr (iData1, type1)
	COMPLEX complexR
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			IF __gmpf_cmp_ui (iData1, 0) < 0 THEN
				complexR.real = Allocate_Float ()
				complexR.imaginary = Allocate_Float ()
				__gmpf_set_ui (complexR.real, 0)
				__gmpf_neg (iData1, iData1)
				__gmpf_sqrt (complexR.imaginary, iData1)
				eval_push ($$TYPE_COMPLEX, COMPLEX_New(complexR))
			ELSE
				x = Allocate_Float ()
				__gmpf_sqrt (x, iData1)
				eval_push ($$TYPE_NUMBER, x)
			END IF
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cbr (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			t0 = Allocate_Float ()
			x = Allocate_Float ()
			__gmpf_set_ui (t0, 1)
			__gmpf_div_ui (t0, t0, 3)
			TransLib_Power (x, iData1, t0)
			Delete_Float (t0)
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sin (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Sin (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cos (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Cos (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_tan (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Tan (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cot (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Cot (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sec (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Sec (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cosec (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Cosec (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_asin (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ASin (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acos (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_Cos (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_atan (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ATan (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acot (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ACot (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_asec (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ASec (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acosec (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ACosec (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sinh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Sinh (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cosh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Cosh (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_tanh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Tanh (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_coth (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Coth (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_sech (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Sech (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_cosech (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			ConvertAngle ($$TRUE, iData1)
			x = Allocate_Float ()
			check = TransLib_Cosech (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_asinh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ASinh (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acosh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ACosh (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_atanh (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ATanh (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acoth (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ACoth (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_asech (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ASech (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_acosech (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_ACosech (x, iData1)
			
			ConvertAngle ($$FALSE, x)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_int (iData1, type1)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_trunc (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_frac (iData1, type1)
	COMPLEX complex1
	COMPLEX complex2
	COMPLEX complexR
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_trunc (x, iData1)
			__gmpf_sub (x, iData1, x)
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_ln (iData1, type1)
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			check = TransLib_Ln (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
			IF check THEN RETURN $$EXCEPTION_UNDEFINED
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION
'
' ############################
' #####  Delete_Complex  #####
' ############################
' Deletes a complex number
' iData = the id of the complex number
' returns $$TRUE on success or $$FALSE on fail
FUNCTION Delete_Complex (iData)
	COMPLEX complex
	IFF COMPLEX_Get (iData, @complex) THEN RETURN $$FALSE
	Delete_Float (complex.real)
	Delete_Float (complex.imaginary)
	RETURN COMPLEX_Delete (iData)
END FUNCTION
'
' ########################
' #####  CopySymbol  #####
' ########################
' Copies data from a symbol
' iData = the data to copy
' type = the type of the data
' returns the copied data
FUNCTION CopySymbol (iData, type)
	COMPLEX complexR
	COMPLEX complex1
	VECTOR vectorR
	VECTOR vector1
	SYMBOL symbol
	FUNC func1
	FUNC funcR
	FUNCCASE case1
	FUNCCASE caseR
	TOKEN token
	LINKEDLIST stack
	
	SELECT CASE type
		CASE $$TYPE_BOOL
			RETURN iData
		CASE $$TYPE_NUMBER, $$TYPE_INT
			x = Allocate_Float ()
			__gmpf_set (x, iData)
			RETURN x
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData, @complex1)
			complexR.real = Allocate_Float ()
			complexR.imaginary = Allocate_Float ()
			__gmpf_set (complexR.real, complex1.real)
			__gmpf_set (complexR.imaginary, complex1.imaginary)
			RETURN COMPLEX_New (complexR)
		CASE $$TYPE_VECTOR
			VECTOR_Get (iData, @vector1)
			
			vectorR.type = vector1.type
			vectorR.size = vector1.size
			LinkedList_Init (@vectorR.values)
			
			hWalk = LinkedList_StartWalk (vector1.values)
			DO WHILE LinkedList_Walk (hWalk, @iV1Data)
				SYMBOL_Get (iV1Data, @symbol)
				symbol.data = CopySymbol (symbol.data, symbol.type)
				LinkedList_Append (@vectorR.values, SYMBOL_New (symbol))
			LOOP
			LinkedList_EndWalk (hWalk)
			
			RETURN VECTOR_New (vectorR)
		CASE $$TYPE_FUNCTION
			FUNC_Get (iData, @func1)
			
			STRING_Get (func1.definition, @def$)
			funcR.definition = STRING_New (def$)
			
			LinkedList_Init (@funcR.paramTypes)
			hWalk = LinkedList_StartWalk (func1.paramTypes)
			DO WHILE LinkedList_Walk (hWalk, @iData)
				LinkedList_Append (@funcR.paramTypes, iData)
			LOOP
			LinkedList_EndWalk (hWalk)
			
			LinkedList_Init (@funcR.cases)
			hWalk = LinkedList_StartWalk (func1.cases)
			DO WHILE LinkedList_Walk (hWalk, @iData)
				FUNCCASE_Get (iData, @case1)
				LinkedList_Init (@caseR.condExpr)
				LinkedList_Init (@caseR.realExpr)
				
				hInnerWalk = LinkedList_StartWalk (case1.condExpr)
				DO WHILE LinkedList_Walk (hInnerWalk, @iInnerData)
					TOKEN_Get (iInnerData, @token)
					GOSUB CopyToken
					LinkedList_Append (@caseR.condExpr, TOKEN_New (token))
				LOOP
				LinkedList_EndWalk (hInnerWalk)
				
				hInnerWalk = LinkedList_StartWalk (case1.realExpr)
				DO WHILE LinkedList_Walk (hInnerWalk, @iInnerData)
					TOKEN_Get (iInnerData, @token)
					GOSUB CopyToken
					LinkedList_Append (@caseR.realExpr, TOKEN_New (token))
				LOOP
				LinkedList_EndWalk (hInnerWalk)
				
				LinkedList_Append (@funcR.cases, FUNCCASE_New (caseR))
			LOOP
			LinkedList_EndWalk (hWalk)
			
			LinkedList_Init (@funcR.caseElse)
			hWalk = LinkedList_StartWalk (func1.caseElse)
			DO WHILE LinkedList_Walk (hWalk, @iData)
				TOKEN_Get (iData, @token)
				GOSUB CopyToken
				LinkedList_Append (@funcR.caseElse, TOKEN_New (token))
			LOOP
			LinkedList_EndWalk (hWalk)
			
			RETURN FUNC_New (funcR)
		CASE ELSE
			RETURN 0
	END SELECT
	
	RETURN 0
	
	SUB CopyToken
		SELECT CASE token.op
			CASE $$OP_LITERAL
				token.value = CopySymbol (token.value, $$TYPE_NUMBER)
			CASE $$OP_FUNCLIT
				token.value = CopySymbol (token.value, $$TYPE_FUNCTION)
			CASE $$OP_BOUNDPARAM
				SYMBOL_Get (token.value, @symbol)
				symbol.data = CopySymbol (symbol.data, symbol.type)
				token.value = SYMBOL_New (symbol)
		END SELECT
	END SUB
END FUNCTION

FUNCTION parse_vexpr (LINKEDLIST out, @exprType)
	TOKEN token
	SYMBOL symbol
	cErrors = 0
	
	exprType = $$TYPE_VECTOR
	cErrors = cErrors+parse_vterm (@out, @exprType)
	
	token$ = PeekToken$ ()
	DO WHILE token$
		token.value = 0
		SELECT CASE UCASE$(token$)
			CASE "("
				TakeToken ()
				cErrors = cErrors + parse_condition (@out, @subExprType)
				token.op = $$OP_VECCOM
				LinkedList_Append (@out, TOKEN_New(token))
				exprType = $$TYPE_NUMBER
				
				token$ = PeekToken$ ()
				IF token$ != ")" THEN
					AddError ("Expected ) in vector component expression")
				ELSE
					TakeToken()
				END IF
				
				EXIT DO
			CASE "."
				TakeToken ()
				cErrors = cErrors + parse_vterm (@out, @exprType)
				token.op = $$OP_DOTPROD
				LinkedList_Append (@out, TOKEN_New(token))
				exprType = $$TYPE_NUMBER
				EXIT DO
			CASE "&"
				TakeToken ()
				cErrors = cErrors + parse_vterm (@out, @exprType)
				token.op = $$OP_CONCAT
				LinkedList_Append (@out, TOKEN_New(token))
			CASE "[", "@", "ANS", "LASTANS", "MKV"
				cErrors = cErrors+parse_vterm (@out, @exprType)
				token.op = $$OP_DOTPROD
				LinkedList_Append (@out, TOKEN_New(token))
				exprType = $$TYPE_NUMBER
				EXIT DO
			CASE ELSE
				IF SearchCompileTimeParamStack (token$, @index, @level, @type) THEN
					IF type = $$TYPE_VECTOR THEN
						cErrors = cErrors+parse_vterm (@out, @exprType)
						token.op = $$OP_DOTPROD
						LinkedList_Append (@out, TOKEN_New(token))
						exprType = $$TYPE_NUMBER
					END IF
				ELSE
					iSymbol = SymbolTable_Retreive (token$, @symbol)
					IF iSymbol && symbol.type = $$TYPE_VECTOR THEN
						cErrors = cErrors+parse_vterm (@out, @exprType)
						token.op = $$OP_DOTPROD
						LinkedList_Append (@out, TOKEN_New(token))
						exprType = $$TYPE_NUMBER
					END IF
				END IF
				EXIT DO
		END SELECT
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
END FUNCTION

FUNCTION parse_vterm (LINKEDLIST out, @exprType)
	TOKEN token
	SYMBOL symbol
	cErrors = 0
	
	token$ = PeekToken$ ()
	SELECT CASE UCASE$(token$)
		CASE "["
			cElements = 0
			
			DO
				TakeToken ()
				token$ = PeekToken$ ()
				IF token$ = "]" THEN EXIT DO
				INC cElements
				parse_expression (@out, @secExprType)
				IF secExprType != $$TYPE_NUMBER THEN
					AddError ("Only vectors of numbers are allowed")
					INC cErrors
				END IF
				token$ = PeekToken$ ()
			LOOP WHILE token$ = ","
			TakeToken ()
			IF token$ != "]" THEN
				AddError ("Expected ]")
				INC cErrors
			END IF
			
			token.op = $$OP_MAKEVECTOR
			token.value = cElements
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "("
			TakeToken ()
			cErrors = cErrors+parse_vexpr (@out, @exprType)
			
			token$ = PeekToken$ ()
			IF token$ != ")" THEN
				AddError ("Expected ) in vector expression")
			ELSE
				TakeToken()
			END IF
		CASE "@", "ANS", "LASTANS"
			TakeToken ()
			token.op = $$OP_LASTANS
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "MKV"
			TakeToken ()
			IF ReadToken$ () = "(" THEN
				cErrors = cErrors + parse_expression (@out, @exprType)
				
				IF ReadToken$ () = "," THEN
					cErrors = cErrors + parse_expression (@out, @exprType)
					
					token.op = $$OP_GENVEC
					LinkedList_Append (@out, TOKEN_New (token))
				ELSE
					INC cErrors
					AddError ("Expected ,")
				END IF
				
				IF PeekToken$ () != ")" THEN
					AddError ("Expected )")
				ELSE
					TakeToken ()
				END IF
			ELSE
				INC cErrors
				AddError ("Expected (")
			END IF
		CASE ELSE
			IF SearchCompileTimeParamStack (token$, @index, @level, @type) THEN
				IF type = $$TYPE_VECTOR THEN
					TakeToken ()
					token.op = $$OP_PARAMETER
					token.value = index
					token.value2 = level
					LinkedList_Append (@out, TOKEN_New (token))
				ELSE
					AddError ("Expected a vector")
					INC cErrors
				END IF
			ELSE
				iSymbol = SymbolTable_Retreive (token$, @symbol)
				IF iSymbol THEN
					IF symbol.type = $$TYPE_VECTOR THEN
						TakeToken ()
						token.op = $$OP_VAR
						token.value = iSymbol
						LinkedList_Append (@out, TOKEN_New (token))
					ELSE
						AddError ("Expected a vector")
						INC cErrors
					END IF
				ELSE
					AddError ("Expected a vector")
					INC cErrors
				END IF
			END IF
	END SELECT
	
	RETURN cErrors
END FUNCTION
'
' ###########################
' #####  Delete_Vector  #####
' ###########################
' Safely deletes a vector
' iData = the vector to delete
' returns $$TRUE on success or $$FALSE on fail
FUNCTION Delete_Vector (iData)
	VECTOR vector
	SYMBOL symbol
	
	VECTOR_Get (iData, @vector)
	hWalk = LinkedList_StartWalk (vector.values)
	DO WHILE LinkedList_Walk (hWalk, @iVecData)
		SYMBOL_Get (iVecData, @symbol)
		@symbol.delete (symbol.data)
		SYMBOL_Delete (iVecData)
	LOOP
	LinkedList_EndWalk (hWalk)
	LinkedList_DeleteAll (@vector.values)
	LinkedList_Uninit (@vector.values)
	
	RETURN VECTOR_Delete (iData)
END FUNCTION

FUNCTION op_dotprod (iData1, type1, iData2, type2)
	VECTOR vector1
	VECTOR vector2
	SYMBOL symbol1
	SYMBOL symbol2
	FUNCADDR Delete (XLONG)
	FUNCADDR DeleteR (XLONG)
	
	IFF (type1 = $$TYPE_VECTOR)&&(type2 = $$TYPE_VECTOR) THEN RETURN $$EXCEPTION_TYPES

	exception = $$EXCEPTION_NONE
	VECTOR_Get (iData1, @vector1)
	VECTOR_Get (iData2, @vector2)
	IF vector1.size != vector2.size THEN RETURN $$EXCEPTION_UNDEFINED 
	
	eval_push ($$TYPE_NUMBER, Allocate_Float ())
	
	hV1Walk = LinkedList_StartWalk (vector1.values)
	hV2Walk = LinkedList_StartWalk (vector2.values)
	DO WHILE LinkedList_Walk (hV1Walk, @iV1Data) && LinkedList_Walk (hV2Walk, @iV2Data)
		SYMBOL_Get (iV1Data, @symbol1)
		SYMBOL_Get (iV2Data, @symbol2)
		
		' Take the conjugate of the second term
		op_conj (symbol2.data, symbol2.type)
		eval_pop (@type, @iData, @Delete)
		symbol2.type = type
		symbol2.data = iData
		
		exception_t = op_mul (symbol1.data, symbol1.type, symbol2.data, symbol2.type)
		
		@Delete (symbol2.data)
		IF exception_t = $$EXCEPTION_NONE THEN
			eval_pop (@type, @iData, @Delete)
			eval_pop (@typeR, @iDataR, @DeleteR)
			op_add (iDataR, typeR, iData, type)
			
			@Delete (iData)
			@DeleteR (iDataR)
		ELSE
			exception = exception_t
		END IF
		
	LOOP
	LinkedList_EndWalk (hV1Walk)
	LinkedList_EndWalk (hV2Walk)
	
	RETURN exception
END FUNCTION
'
' ############################
' #####  parse_fLiteral  #####
' ############################
FUNCTION parse_fLiteral (@out)
	SHARED in$
	SHARED pos
	SHARED LINKEDLIST	compileParamStack
	FUNC func
	PARAMETER param
	FUNCCASE	case
	LINKEDLIST compileTimeLocalParams
	
	cErrors = 0
	
	' We need to capture the exact definition the user provided so we can echo back
	' something meaningfull on request.
	startPos = pos
	IF ReadToken$ () != "{" THEN
		INC cErrors
		AddError ("Expected {")
	END IF
	
	LinkedList_Init (@compileTimeLocalParams)
	
	LinkedList_Init (@func.paramTypes)
	LinkedList_Init (@func.cases)
	
	DO
		token$ = ReadToken$ ()
		SELECT CASE token$
			CASE "bool"			: param.type = $$TYPE_BOOL
			CASE "int"			:	param.type = $$TYPE_INT
			CASE "real"			: param.type = $$TYPE_NUMBER
			CASE "complex"	: param.type = $$TYPE_COMPLEX
			CASE "vector"		: param.type = $$TYPE_VECTOR
			CASE "function"	: param.type = $$TYPE_FUNCTION
			CASE ELSE
				INC cErrors
				AddError ("Expected a type, \""+token$+"\" is not a valid type")
		END SELECT
		
		token$ = ReadToken$ ()
		param.name = STRING_New (token$)
		
		LinkedList_Append (@compileTimeLocalParams, PARAMETER_New (param))
		LinkedList_Append (@func.paramTypes, param.type)
		token$ = ReadToken$ ()
	LOOP WHILE token$ = ","
	
	LinkedList_Insert (@compileParamStack, 0, LINKEDLIST_New(compileTimeLocalParams))
	
	IF token$ != "|" THEN
		INC cErrors
		AddError ("Expected |")
	END IF
	
	funcType = $$TYPE_INVALID
	DO
		LinkedList_Init (@case.realExpr)
		cErrors = cErrors + parse_expression (@case.realExpr, @exprType)
		
		token$ = ReadToken$ ()
		SELECT CASE token$
			CASE ":"
				' Parse a conditional expression
				LinkedList_Init (@case.condExpr)
				cErrors = cErrors + parse_condition (@case.condExpr, @exprType)
				
				LinkedList_Append (@func.cases, FUNCCASE_New (case))
			CASE "}"
				' The function is over
				func.caseElse = case.realExpr
				EXIT DO
			CASE ELSE
				INC cErrors
				AddError ("Expected }")
		END SELECT
		token$ = ReadToken$ ()
	LOOP WHILE token$ = ";"
	
	' Remove the local parameters
	LinkedList_GetItem (compileParamStack, 0, @iData)
	LINKEDLIST_Get (iData, @compileTimeLocalParams)
	hWalk = LinkedList_StartWalk (compileTimeLocalParams)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		PARAMETER_Get (iData, @param)
		STRING_Delete (param.name)
		PARAMETER_Delete (iData)
	LOOP
	LinkedList_EndWalk (hWalk)
	LinkedList_Uninit (@compileTimeLocalParams)
	LinkedList_DeleteItem (@compileParamStack, 0)
	
	IF token$ != "}" THEN
		INC cErrors
		AddError ("Expected }")
	END IF
	
	' The definition field stores the unparsed version of the function
	func.definition = STRING_New (TRIM$(MID$(in$, startPos+1, pos-startPos)))
	func.isClosure = $$FALSE
	
	out = FUNC_New (func)
	
	RETURN cErrors
END FUNCTION
'
' #############################
' #####  parse_condition  #####
' #############################
FUNCTION parse_condition (LINKEDLIST out, @exprType)
	STATIC expr_uniq_id	' This id must be unique within an expression
	TOKEN token
	cErrors = 0
	
	cErrors = cErrors + parse_condTerm (@out, @exprType)
	
	token$ = PeekToken$ ()
	DO WHILE token$
		token.value = 0
		SELECT CASE token$
			CASE ""
			CASE ",", "]", ":", "}",";"
				EXIT DO
			CASE "&&"
				' insert short cicuit marker
				token.value = expr_uniq_id
				INC expr_uniq_id
				token.op = $$OP_SHORTC_F
				LinkedList_Append (@out, TOKEN_New (token))
				TakeToken ()
				cErrors = cErrors + parse_condTerm (@out, @exprType)
				token.op = $$OP_LOGAND
				LinkedList_Append (@out, TOKEN_New (token))
			CASE "||"
				' insert short cicuit marker
				token.value = expr_uniq_id
				INC expr_uniq_id
				token.op = $$OP_SHORTC_T
				LinkedList_Append (@out, TOKEN_New (token))
				TakeToken ()
				cErrors = cErrors + parse_condTerm (@out, @exprType)
				token.op = $$OP_LOGOR
				LinkedList_Append (@out, TOKEN_New (token))
			CASE ")"
				RETURN cErrors
			CASE ELSE
				TakeToken ()
				AddError ("Syntax error, expected an operator.  Not \""+token$+"\"")
				INC cErrors
		END SELECT
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
END FUNCTION
'
' ############################
' #####  parse_condTerm  #####
' ############################
FUNCTION parse_condTerm (LINKEDLIST out, @exprType)
	TOKEN token
	cErrors = 0
	
'	? "Start expression"
	cErrors = cErrors + parse_nexpr (@out, @exprType)
'	? "End expression"
	
	token.value = 0
	token$ = PeekToken$ ()
'	? "TOKEN: \"";token$;"\""
	SELECT CASE token$
		CASE "="
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELEQ
			LinkedList_Append (@out, TOKEN_New (token))
		CASE ">"
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELGT
			LinkedList_Append (@out, TOKEN_New (token))
		CASE ">="
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELGTE
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "<"
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELLT
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "<="
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELLTE
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "<>", "><"
			TakeToken ()
			parse_nexpr (@out, @exprType)
			token.op = $$OP_RELNOTEQ
			LinkedList_Append (@out, TOKEN_New (token))
		CASE ELSE
			' do nothing, just let the next level handle it
	END SELECT
	
	RETURN cErrors
END FUNCTION
'
' #########################
' #####  parse_fexpr  #####
' #########################
FUNCTION parse_fexpr (LINKEDLIST out, @exprType)
	TOKEN token
	SYMBOL symbol
	cErrors = 0
	
	token$ = PeekToken$ ()
	SELECT CASE UCASE$(token$)
		CASE "{"
			token.op = $$OP_FUNCLIT
			cErrors = cErrors + parse_fLiteral (@value)
			token.value = value
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "SELF"
			TakeToken ()
			token.op = $$OP_RECURSE
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE "@", "ANS", "LASTANS"
			TakeToken ()
			token.op = $$OP_LASTANS
			token.value = 0
			LinkedList_Append (@out, TOKEN_New (token))
		CASE ELSE
			IF SearchCompileTimeParamStack (token$, @index, @level, @type) THEN
				TakeToken ()
				token.op = $$OP_PARAMETER
				token.value = index
				token.value2 = level
				LinkedList_Append (@out, TOKEN_New (token))
			ELSE
				iSymbol = SymbolTable_Retreive (token$, @symbol)
				TakeToken ()
				token.op = $$OP_VAR
				token.value = iSymbol
				LinkedList_Append (@out, TOKEN_New (token))
			END IF
	END SELECT
	
	token.op = $$OP_FUNCCALL
	token$ = PeekToken$ ()
	DO WHILE token$ = "("
		' Process a function call
		
		token.value = 0
		DO
			TakeToken ()
			INC token.value
			
			cErrors = cErrors + parse_expression (@out, @exprType)
			
			token$ = PeekToken$ ()
		LOOP WHILE token$ = ","
		IF token$ != ")" THEN
			AddError ("Expected ) at end of function call")
		ELSE
			TakeToken ()
		END IF
		
		LinkedList_Append (@out, TOKEN_New (token))
		
		token$ = PeekToken$ ()
	LOOP
	
	RETURN cErrors
END FUNCTION
'
' #########################################
' #####  SearchCompileTimeParamStack  #####
' #########################################
' Searches the compile time parameter stack for a parameter
' name$ = the name of the parameter
' @index = the variable to store the 0 based index of the parameter
' @level = the lexical level of the parameter, 0 for this function, 1 for the parent etc
' @type = the type of the parameter
' returns $$TRUE if the parameter is found otherwise $$FALSE
FUNCTION SearchCompileTimeParamStack (name$, @index, @level, @type)
	SHARED LINKEDLIST	compileParamStack
	LINKEDLIST levelList
	PARAMETER param
	
	level = -1
	done = $$FALSE
	hLevelWalk = LinkedList_StartWalk (compileParamStack)
	DO WHILE LinkedList_Walk (hLevelWalk, @iLevelData) && !done
		LINKEDLIST_Get (iLevelData, @levelList)
		INC level
		
		index = -1
		hParamWalk = LinkedList_StartWalk (levelList)
		DO WHILE LinkedList_Walk (hParamWalk, @iParamData) && !done
			INC index
			
			PARAMETER_Get (iParamData, @param)
			STRING_Get (param.name, @paramName$)
			IF UCASE$(paramName$) = UCASE$(name$) THEN
				type = param.type
				done = $$TRUE
			END IF
		LOOP
		LinkedList_EndWalk (hParamWalk)
		
	LOOP
	LinkedList_EndWalk (hLevelWalk)
	
	RETURN done
END FUNCTION
'
' #############################
' #####  Delete_Function  #####
' #############################
' Deletes a function
' iFunc = the id of the function
' returns $$TRUE on success or $$FALSE on fail
FUNCTION Delete_Function (iFunc)
	FUNC func
	FUNCCASE case
	
	FUNC_Get (iFunc, @func)
	STRING_Delete (func.definition)
	LinkedList_Uninit (@func.paramTypes)
	hWalk = LinkedList_StartWalk (func.cases)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		FUNCCASE_Get (iData, @case)
		Delete_Expression (case.condExpr)
		Delete_Expression (case.realExpr)
		FUNCCASE_Delete (iData)
	LOOP
	LinkedList_EndWalk (hWalk)
	
	LinkedList_DeleteAll (@func.cases)
	LinkedList_Uninit (@func.cases)
	
	Delete_Expression (func.caseElse)
	
	RETURN FUNC_Delete (iFunc)
END FUNCTION
'
' ###############################
' #####  Delete_Expression  #####
' ###############################
FUNCTION Delete_Expression (LINKEDLIST list)
	TOKEN token
	SYMBOL symbol
	
	hWalk = LinkedList_StartWalk (list)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		TOKEN_Get (iData, @token)
		SELECT CASE token.op
			CASE $$OP_LITERAL
				Delete_Float (token.value)
			CASE $$OP_FUNCLIT
				Delete_Function (token.value)
			CASE $$OP_BOUNDPARAM
				SYMBOL_Get (token.value, @symbol)
				@symbol.delete (symbol.data)
				SYMBOL_Delete (token.value)
		END SELECT
		TOKEN_Delete (iData)
	LOOP
	LinkedList_EndWalk (hWalk)
	
	LinkedList_Uninit (@list)
END FUNCTION
'
' ########################
' #####  op_closure  #####
' ########################
' Performs a closure operation on function iFunc
' level = the lexical level of the closure
' returns the id of the new function
FUNCTION op_closure (iFunc, level)
	SHARED LINKEDLIST	runtimeParamStack
	
	FUNC func
	FUNCCASE case
	TOKEN token
	LINKEDLIST stack
	SYMBOL symbol
	
	FUNC_Get (iFunc, @func)
	func.isClosure = $$TRUE
	FUNC_Update (iFunc, func)
	
	' Look at the cases
	hWalk = LinkedList_StartWalk (func.cases)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		FUNCCASE_Get (iData, @case)
		hInnerWalk = LinkedList_StartWalk (case.realExpr)
		DO WHILE LinkedList_Walk (hInnerWalk, @iInnerData)
			GOSUB Closure
		LOOP
		LinkedList_EndWalk (hInnerWalk)
		
		' Bound parameters can occur in conditions too!
		hInnerWalk = LinkedList_StartWalk (case.condExpr)
		DO WHILE LinkedList_Walk (hInnerWalk, @iInnerData)
			GOSUB Closure
		LOOP
		LinkedList_EndWalk (hInnerWalk)
	LOOP
	LinkedList_EndWalk (hWalk)
	
	' Look at the else case
	hWalk = LinkedList_StartWalk (func.caseElse)
	DO WHILE LinkedList_Walk (hWalk, @iInnerData)
		GOSUB Closure
	LOOP
	LinkedList_EndWalk (hWalk)
	
	RETURN iFunc
	
	SUB Closure
		TOKEN_Get (iInnerData, @token)
		SELECT CASE token.op
			CASE $$OP_FUNCLIT
				token.value = op_closure (token.value, level+1)
			CASE $$OP_PARAMETER
				IF token.value2 = level THEN
					' Bind the parameter
					LinkedList_GetItem (runtimeParamStack, 0, @iData)
					LINKEDLIST_Get (iData, @stack)
					LinkedList_GetItem (stack, token.value, @iData)
					SYMBOL_Get (iData, @symbol)
					symbol.data = CopySymbol (symbol.data, symbol.type)
					token.op = $$OP_BOUNDPARAM
					token.value = SYMBOL_New (symbol)
				END IF
		END SELECT
		TOKEN_Update (iInnerData, token)
	END SUB
END FUNCTION

FUNCTION op_eq (iData1, type1, iData2, type2, @exception)
	COMPLEX complex1
	COMPLEX complex2
	VECTOR vector1
	VECTOR vector2
	SYMBOL symbol1
	SYMBOL symbol2
	
	SELECT CASE type1
		CASE $$TYPE_INT, $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					IF TransLib_IsEq (iData1, iData2) THEN RETURN $$TRUE
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					IF __gmpf_cmp_ui (complex2.imaginary, 0) = 0 THEN
						IF TransLib_IsEq (iData1, complex2.real) THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					IF __gmpf_cmp_ui (complex1.imaginary, 0) = 0 THEN
						IF TransLib_IsEq (complex1.real, iData2) THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					IF (TransLib_IsEq (complex1.real, complex2.real))&& _
						(TransLib_IsEq (complex1.imaginary, complex2.imaginary)) THEN RETURN $$TRUE
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_VECTOR
			IF type2 = $$TYPE_VECTOR THEN
				VECTOR_Get (type1, @vector1)
				VECTOR_Get (type2, @vector2)
				
				result = $$TRUE
				hV1Walk = LinkedList_StartWalk (vector1.values)
				hV2Walk = LinkedList_StartWalk (vector2.values)
				DO WHILE LinkedList_Walk (hV1Walk, @iV1Data) && LinkedList_Walk (hV2Walk, @iV2Data)
					SYMBOL_Get (iV1Data, @symbol1)
					SYMBOL_Get (iV2Data, @symbol2)
					result = result && op_eq (symbol1.data, symbol1.type, symbol2.data, symbol2.type, @exception)
				LOOP
				LinkedList_EndWalk (hV1Walk)
				LinkedList_EndWalk (hV2Walk)
				
				RETURN result
			ELSE
				exception = $$EXCEPTION_TYPES
			END IF
		CASE ELSE
			exception = $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$FALSE
END FUNCTION

FUNCTION op_gt (iData1, type1, iData2, type2, @exception)
	COMPLEX complex1
	COMPLEX complex2

	SELECT CASE type1
		CASE $$TYPE_INT, $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					IF __gmpf_cmp (iData1, iData2) > 0 THEN RETURN $$TRUE
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					IF __gmpf_cmp_ui (complex2.imaginary, 0) = 0 THEN
						IF __gmpf_cmp (iData1, complex2.real) > 0 THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					IF __gmpf_cmp_ui (complex1.imaginary, 0) = 0 THEN
						IF __gmpf_cmp (complex1.real, iData2) > 0 THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					IF (__gmpf_cmp (complex1.real, complex2.real) > 0)&& _
						(__gmpf_cmp (complex1.imaginary, complex2.imaginary) > 0) THEN RETURN $$TRUE
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			exception = $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$FALSE
END FUNCTION

FUNCTION op_lt (iData1, type1, iData2, type2, @exception)
	COMPLEX complex1
	COMPLEX complex2

	SELECT CASE type1
		CASE $$TYPE_INT, $$TYPE_NUMBER
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					IF __gmpf_cmp (iData1, iData2) < 0 THEN RETURN $$TRUE
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData2, @complex2)
					IF __gmpf_cmp_ui (complex2.imaginary, 0) = 0 THEN
						IF __gmpf_cmp (iData1, complex2.real) < 0 THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE $$TYPE_COMPLEX
			SELECT CASE type2
				CASE $$TYPE_INT, $$TYPE_NUMBER
					COMPLEX_Get (iData1, @complex1)
					IF __gmpf_cmp_ui (complex1.imaginary, 0) = 0 THEN
						IF __gmpf_cmp (complex1.real, iData2) < 0 THEN RETURN $$TRUE
					ELSE
						exception = $$EXCEPTION_TYPES
					END IF
				CASE $$TYPE_COMPLEX
					COMPLEX_Get (iData1, @complex1)
					COMPLEX_Get (iData2, @complex2)
					IF (__gmpf_cmp (complex1.real, complex2.real) > 0)&& _
						(__gmpf_cmp (complex1.imaginary, complex2.imaginary) < 0) THEN RETURN $$TRUE
				CASE ELSE
					exception = $$EXCEPTION_TYPES
			END SELECT
		CASE ELSE
			exception = $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$FALSE
END FUNCTION

FUNCTION op_exp (iData1, type1)
	COMPLEX complex1
	COMPLEX complexR
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			TransLib_Exp (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData1, @complex1)
			complexR.real = Allocate_Float ()
			complexR.imaginary = Allocate_Float ()
			t0 = Allocate_Float ()
			TransLib_Exp (t0, complex1.real)
			TransLib_Cos (complexR.real, complex1.imaginary)
			TransLib_Sin (complexR.imaginary, complex1.imaginary)
			__gmpf_mul (complexR.real, complexR.real, t0)
			__gmpf_mul (complexR.imaginary, complexR.imaginary, t0)
			Delete_Float (t0)
			eval_push ($$TYPE_COMPLEX, COMPLEX_New (complexR))
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_re (iData1, type1)
	COMPLEX complex1
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_set (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData1, @complex1)
			x = Allocate_Float ()
			__gmpf_set (x, complex1.real)
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_im (iData1, type1)
	COMPLEX complex1
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_set_ui (x, 0)
			eval_push ($$TYPE_NUMBER, x)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData1, @complex1)
			x = Allocate_Float ()
			__gmpf_set (x, complex1.imaginary)
			eval_push ($$TYPE_NUMBER, x)
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

FUNCTION op_conj (iData1, type1)
	COMPLEX complex1
	COMPLEX complexR
	
	SELECT CASE type1
		CASE $$TYPE_NUMBER
			x = Allocate_Float ()
			__gmpf_set (x, iData1)
			eval_push ($$TYPE_NUMBER, x)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get (iData1, @complex1)
			complexR.real = Allocate_Float ()
			complexR.imaginary = Allocate_Float ()
			
			__gmpf_set (complexR.real, complex1.real)
			__gmpf_set (complexR.imaginary, complex1.imaginary)
			__gmpf_neg (complexR.imaginary, complexR.imaginary)
			
			eval_push ($$TYPE_COMPLEX, COMPLEX_New (complexR))
		CASE ELSE
			RETURN $$EXCEPTION_TYPES
	END SELECT
	
	RETURN $$EXCEPTION_NONE
END FUNCTION
'
' ##################################
' #####  Delete_RuntimeParams  #####
' ##################################
' Deletes the runtime parameter info for a particular function
' level = the level of the function to clear params for - 0 for current function
' returns $$TRUE on success or $$FALSE on error
FUNCTION Delete_RuntimeParams (level)
	SHARED LINKEDLIST runtimeParamStack
	SYMBOL symbol1
	FUNC func
	LINKEDLIST stack
	
	IFF LinkedList_GetItem (runtimeParamStack, level, @iStack) THEN RETURN $$FALSE
	IFF LINKEDLIST_Get (iStack, @stack) THEN RETURN $$FALSE
	hFWalk = LinkedList_StartWalk (stack)
	DO WHILE LinkedList_Walk (hFWalk, @iData)
		SYMBOL_Get (iData, @symbol1)
		IF symbol1.type = $$TYPE_FUNCTION THEN
			FUNC_Get (symbol1.data, @func)
			IF func.isClosure THEN
				@symbol1.delete (symbol1.data)
				SYMBOL_Delete (iData)
			ELSE
				SYMBOL_Delete (iData)
			END IF
		ELSE
			@symbol1.delete (symbol1.data)
			SYMBOL_Delete (iData)
		END IF
	LOOP
	LinkedList_EndWalk (hFWalk)
	LinkedList_Uninit (@stack)
	LINKEDLIST_Delete (iStack)
	
	LinkedList_DeleteItem (@runtimeParamStack, level)
	
	RETURN $$TRUE
END FUNCTION
'
' #########################
' #####  GetFuncCase  #####
' #########################
' Gets the function case to execute
' func = the function to get the case for
' iExpr = the variable to store the id of the expression
' iCase = the variable to store the index of the case - -1 for case else
' returns any exceptions encountered
FUNCTION GetFuncCase (FUNC func, LINKEDLIST expr, @iCase)
	FUNCCASE case
	
	exception = $$EXCEPTION_NONE
	
	valueFound = $$FALSE
	hFWalk = LinkedList_StartWalk (func.cases)
	iCase = -1
	DO WHILE LinkedList_Walk (hFWalk, @iData) && !valueFound
		INC iCase
		FUNCCASE_Get (iData, @case)
		
		evaluate (case.condExpr, @exception, @iData, @type, 0, -1)
		IF type != $$TYPE_BOOL THEN
'			PRINT exception, iData, type
			exception = $$EXCEPTION_TYPES
			LinkedList_EndWalk (hFWalk)
			EXIT DO
		ELSE
			IF iData THEN
				expr = case.realExpr
				valueFound = $$TRUE
				LinkedList_EndWalk (hFWalk)
				EXIT DO
			END IF
		END IF
	LOOP
	LinkedList_EndWalk (hFWalk)
	
	IF !valueFound && exception = $$EXCEPTION_NONE THEN
		IF func.caseElse.cItems THEN
			expr = func.caseElse
			iCase = -1
		ELSE
			exception = $$EXCEPTION_UNDEFINED
		END IF
	END IF
	
	RETURN exception
END FUNCTION
'
' ###########################
' #####  FormatSymbol$  #####
' ###########################
' Formats a symbol for output
' symbol = the symbol to format
' returns the formatted symbol
FUNCTION FormatSymbol$ (iData, type)
	COMPLEX complex
	VECTOR vector
	FUNC func
	SYMBOL symbol

	ret$ = ""
	SELECT CASE type
		CASE $$TYPE_BOOL
			IF iData THEN RETURN "1" ELSE RETURN "0"
		CASE $$TYPE_NUMBER
			ret$ = FormatFloat$ (iData)
		CASE $$TYPE_COMPLEX
			COMPLEX_Get(iData, @complex)
			IF __gmpf_cmp_ui (complex.real, 0) = 0 THEN
				SELECT CASE TRUE
					CASE __gmpf_cmp_si (complex.imaginary, 0) = 0
						ret$ = "0"
					CASE __gmpf_cmp_si (complex.imaginary, 1) = 0
						ret$ = "i"
					CASE __gmpf_cmp_si (complex.imaginary, -1) = 0
						ret$ = "-i"
					CASE ELSE
						ret$ = FormatFloat$ (complex.imaginary)+"i"
				END SELECT
			ELSE
				ret$ = FormatFloat$ (complex.real)
				SELECT CASE TRUE
					CASE __gmpf_cmp_si (complex.imaginary, 0) = 0
					CASE __gmpf_cmp_si (complex.imaginary, 1) = 0
						ret$ = ret$+" + i"
					CASE __gmpf_cmp_si (complex.imaginary, -1) = 0
						ret$ = ret$+" - i"
					CASE __gmpf_cmp_si (complex.imaginary, 0) < 0
						ret$ = ret$+" - "+LCLIP$(FormatFloat$ (complex.imaginary))+"i"
					CASE ELSE
						ret$ = ret$+" + "+FormatFloat$ (complex.imaginary)+"i"
				END SELECT
			END IF
		CASE $$TYPE_VECTOR
			VECTOR_Get (iData, @vector)
			hVecWalk = LinkedList_StartWalk (vector.values)
			ret$ = "["
			DO WHILE LinkedList_Walk (hVecWalk, @iVecData)
				SYMBOL_Get (iVecData, @symbol)
				ret$ = ret$+FormatSymbol$ (symbol.data, symbol.type)+", "
			LOOP
			IF ret$ = "[" THEN ret$ = "[]" ELSE ret$ = RCLIP$(ret$,2)+"]"
			LinkedList_EndWalk (hVecWalk)
		CASE $$TYPE_FUNCTION
			FUNC_Get (iData, @func)
			STRING_Get (func.definition, @ret$)
	END SELECT
	
	RETURN ret$
END FUNCTION
'
' ##################################
' #####  calculate_getSymbols  #####
' ##################################
' Retreives a list of symbols
' returns a string containing all the defined symbols and their definitions
FUNCTION calculate_getSymbols$ ()
	SHARED CALCSTATE calcState
	SHARED ASSOCARRAY symbolTable
	SYMBOL symbol
	
	SYMBOL_Get (calcState.iLastAns, @symbol)
	ret$ = "@ = "+FormatSymbol$ (symbol.data, symbol.type)
	traverse = BinTree_StartTraversal (symbolTable.tree, $$ADT_PREORDER)
	DO WHILE BinTree_Traverse (traverse, @iData, @iKey)
		SYMBOL_Get (iData, @symbol)
		STRING_Get (iKey, @key$)
		ret$ = ret$+"\n"+key$+" = "+FormatSymbol$ (symbol.data, symbol.type)
	LOOP
	BinTree_EndTraversal (traverse)
	
	RETURN ret$
END FUNCTION
'
' ##########################
' #####  ConvertAngle  #####
' ##########################
' Converts an angle to or from radians based on the current angle mode
' inout = $$TRUE if the conversion is for input or $$FALSE for output conversion
' iAngle = the gmpf to convert
' returns $$TRUE on success or $$FALSE on fail
FUNCTION ConvertAngle (in, iAngle)
	SHARED CALCSTATE calcState
	SHARED CONSTANTS const
	
	IF in THEN
		SELECT CASE calcState.angleMode
			CASE $$ANGLE_DEGREES
				__gmpf_div (iAngle, iAngle, const.pRadToDeg)
			CASE $$ANGLE_GRADS
				__gmpf_div (iAngle, iAngle, const.pRadToGrad)
		END SELECT
	ELSE
		SELECT CASE calcState.angleMode
			CASE $$ANGLE_DEGREES
				__gmpf_mul (iAngle, iAngle, const.pRadToDeg)
			CASE $$ANGLE_GRADS
				__gmpf_mul (iAngle, iAngle, const.pRadToGrad)
		END SELECT
	END IF
	
	RETURN $$TRUE
END FUNCTION
'
' ####################################
' #####  calculate_getPrecision  #####
' ####################################
'
'
'
FUNCTION calculate_getPrecision ()
	SHARED CONSTANTS const
	
	RETURN const.cBitsPrec
END FUNCTION
'
' ###################################
' #####  calculate_getRounding  #####
' ###################################
'
'
'
FUNCTION calculate_getRounding (@roundParam)
	SHARED CALCSTATE calcState
	
	roundParam = calcState.roundParameter
	RETURN calcState.roundMode
END FUNCTION
'
' ##################################
' #####  calculate_getOutMode  #####
' ##################################
'
'
'
FUNCTION calculate_getOutMode ()
	SHARED CALCSTATE calcState
	
	RETURN calcState.outMode
END FUNCTION
'
' ################################
' #####  calculate_getAngle  #####
' ################################
'
'
'
FUNCTION calculate_getAngle ()
	SHARED CALCSTATE calcState
	
	RETURN calcState.angleMode
END FUNCTION
'
' ###########################
' #####  StringToFloat  #####
' ###########################
' Converts a string to gmpf
' stri$ = the string
' float = the float
' returns $$TRUE on success or $$FALSE on fail
FUNCTION StringToFloat (stri$, float)
	mpz_t int1R
	
	int1 = &int1R
	
	SELECT CASE MID$(stri$, 1, 2)
		CASE "0x"
			rstri$ = MID$(stri$, 3)
			ret = __gmpz_init_set_str (int1, &rstri$, 16)
			__gmpf_set_z (float, int1)
			__gmpz_clear (int1)
		CASE "0o"
			rstri$ = MID$(stri$, 3)
			ret = __gmpz_init_set_str (int1, &rstri$, 8)
			__gmpf_set_z (float, int1)
			__gmpz_clear (int1)
		CASE "0b"
			rstri$ = MID$(stri$, 3)
			ret = __gmpz_init_set_str (int1, &rstri$, 2)
			__gmpf_set_z (float, int1)
			__gmpz_clear (int1)
		CASE ELSE
			RETURN __gmpf_set_str (float, &stri$, 10)
	END SELECT
	
	RETURN ret
END FUNCTION

FUNCTION op_veccom (iData1, type1, iData2, type2)
	VECTOR vector
	SYMBOL symbol
	
	IF type1 != $$TYPE_VECTOR THEN
'		PRINT "VECCOM TYPE: ";type1
		' treat this as a multiplication
		RETURN op_mul (iData1, type1, iData2, type2)
	END IF

	IF type2 != $$TYPE_INT THEN
		IF type2 = $$TYPE_NUMBER THEN
			IF !__gmpf_integer_p (iData2) THEN RETURN $$EXCEPTION_TYPES
		ELSE
			IF type2 = $$TYPE_VECTOR THEN
				RETURN op_dotprod (iData1, type1, iData2, type2)
			ELSE
				RETURN $$EXCEPTION_TYPES
			END IF
		END IF
	END IF
	
	VECTOR_Get (iData1, @vector)
	
	comp = __gmpf_get_si (iData2)
	IF comp < 1 || comp > vector.size THEN RETURN $$EXCEPTION_UNDEFINED
	
	' Get the item
	LinkedList_GetItem (vector.values, comp-1, @iData)
	SYMBOL_Get (iData, @symbol)
	
	' Push it onto the stack
	eval_push (symbol.type, CopySymbol (symbol.data, symbol.type))
	
	RETURN $$EXCEPTION_NONE
END FUNCTION

CFUNCTION gmp_malloc (aSize)
	ret = GlobalAlloc ($$GMEM_FIXED, aSize)
	IFZ ret THEN
		XstSystemErrorNumberToName (GetLastError (), @sysError$)
		msg$ = "Memmory allocation error: \""+sysError$+"\", aSize = "+STRING$(aSize)
		MessageBoxA (0, &msg$, &"Error", $$MB_ICONSTOP|$$MB_OK)
		QUIT(0)
	END IF
	
	RETURN ret
END FUNCTION
CFUNCTION gmp_realloc (ptr, oldSize, newSize)
	ret = GlobalReAlloc (ptr, newSize, 0)
	IFZ ret THEN
		XstSystemErrorNumberToName (GetLastError (), @sysError$)
		msg$ = "Memmory reallocation error: \""+sysError$+"\", ptr = "+STRING$(ptr)+", oldSize = "+STRING$(oldSize)+", newSize = "+STRING$(newSize)
		MessageBoxA (0, &msg$, &"Error", $$MB_ICONSTOP|$$MB_OK)
		QUIT(0)
	END IF
	
	RETURN ret
END FUNCTION
CFUNCTION gmp_free (ptr, aSize)
	GlobalFree (ptr)
END FUNCTION
'
' ########################################
' #####  calculate_getDependantVars  #####
' ########################################
' Retreives a list of all the variables in an expression
' iExpr = the handle of the expression
' vars$[] = an array to store the variable names
' returns the number of variables iExpr depends on
FUNCTION calculate_getDependantVars (iExpr, vars[])
	LINKEDLIST expr
	TOKEN token
	
	DIM vars[7]
	cVars = 0
	LINKEDLIST_Get(iExpr, @expr)
	hWalk = LinkedList_StartWalk (expr)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		TOKEN_Get(iData, @token)
		IF token.op = $$OP_VAR THEN
			IF cVars > UBOUND(vars[]) THEN REDIM vars[((UBOUND(vars[])+1)<<1)-1]
			vars[cVars] = token.value
			INC cVars
		END IF
	LOOP
	LinkedList_EndWalk (hWalk)
	
	REDIM vars[cVars-1]
	RETURN cVars
END FUNCTION
'
' ################################
' #####  calculate_getVarID  #####
' ################################
' Gets the internal id of a variable
' var$ = the variable to get the id of
' returns the id of the variable or 0 if it is undefined
FUNCTION calculate_getVarID (var$)
	SYMBOL symbol
	RETURN SymbolTable_Retreive (var$, @symbol)
END FUNCTION
'
' ###################################
' #####  calculate_fastExecute  #####
' ###################################
' Performs a fast execution of an expression
' iExpr = the expression to execute
' returns $$TRUE if there was an error, otherwise $$FALSE
FUNCTION calculate_fastExecute (iExpr)
	SHARED CALCSTATE calcState
	SHARED STACK evalStack
	SHARED errors$
	
	LINKEDLIST expression
	
	SYMBOL symbol
	
	FUNCADDR Delete(XLONG)
	FUNCADDR Delete2(XLONG)
	
	errors$ = ""
	resultString$ = ""
	exception = $$EXCEPTION_NONE
	
	' Push a dummy value onto the stack to handle commands that naturally underflow
	x = Allocate_Float ()
	__gmpf_set_ui (x, 0)
	eval_push ($$TYPE_NUMBER, x)
	
	IF LINKEDLIST_Get (iExpr, @expression) THEN
		cErrors = evaluate (expression, @exception, @iData, @type, 0, -1)
		
		' pop off the dummy, preventing a memmory leak
		IF eval_pop (0, @iData2, @Delete2) THEN
			@Delete2(iData2)
		END IF
		
		IF exception = $$EXCEPTION_NONE
			SELECT CASE type
				CASE $$TYPE_NUMBER, $$TYPE_INT
					Delete = &Delete_Float ()
				CASE $$TYPE_COMPLEX
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Complex ()
				CASE $$TYPE_VECTOR
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Vector ()
				CASE $$TYPE_FUNCTION
					exception = $$EXCEPTION_NOTNUM
					Delete = &Delete_Function ()
			END SELECT
		END IF
	ELSE
		RETURN $$TRUE
	END IF
	
	IF cErrors THEN
		RETURN $$TRUE
	ELSE
		' Replace the last answer
		SYMBOL_Get (calcState.iLastAns, @symbol)
		@symbol.delete (symbol.data)
		SYMBOL_Delete (calcState.iLastAns)
		symbol.type = type
		symbol.data = CopySymbol (iData, type) 
		symbol.delete = Delete
		calcState.iLastAns = SYMBOL_New (symbol)
	END IF
	@Delete(iData)
END FUNCTION
'
' #################################
' #####  calculate_getVector  #####
' #################################
' Gets @ as a vector
' vec = an array of pointers to mpf_t's
' returns the number of components in the vector, 0 on error
FUNCTION calculate_getVector (vec[])
	SHARED CALCSTATE calcState
	SYMBOL symbol
	VECTOR vector
	
	SYMBOL_Get (calcState.iLastAns, @symbol)
	IF symbol.type != $$TYPE_VECTOR THEN RETURN 0
	VECTOR_Get (symbol.data, @vector)
	
	i = 0
	hWalk = LinkedList_StartWalk (vector.values)
	DO WHILE LinkedList_Walk (hWalk, @iData) && i <= UBOUND(vec[])
		SYMBOL_Get (iData, @symbol)
		IF symbol.type = $$TYPE_INT || symbol.type = $$TYPE_NUMBER THEN
			__gmpf_set (vec[i], symbol.data)
		ELSE
			LinkedList_EndWalk (hWalk)
			RETURN 0
		END IF
		INC i
	LOOP
	LinkedList_EndWalk (hWalk)
	
	RETURN vector.size
END FUNCTION
'
' ###############################
' #####  calculate_getReal  #####
' ###############################
' Gets @ as a real
' pMpf = a pointer to an mpf_t
' returns $$TRUE on success or $$FALSE on fail
FUNCTION calculate_getReal (pMpf)
	SHARED CALCSTATE calcState
	SYMBOL symbol
	VECTOR vector
	
	SYMBOL_Get (calcState.iLastAns, @symbol)
	IF symbol.type = $$TYPE_INT || symbol.type = $$TYPE_NUMBER THEN
		__gmpf_set (pMpf, symbol.data)
		RETURN $$TRUE
	END IF
	
	RETURN $$FALSE
END FUNCTION

FUNCTION op_concat (iData1, type1, iData2, type2)
	VECTOR vector1
	VECTOR vector2
	VECTOR vectorR
	SYMBOL symbol
	SYMBOL symbolR
	
	IF !(type1 = $$TYPE_VECTOR && type2 = $$TYPE_VECTOR) THEN RETURN $$EXCEPTION_TYPES
	IFF VECTOR_Get(iData1, @vector1) THEN RETURN $$EXCEPTION_INTERNAL
	IFF VECTOR_Get(iData2, @vector2) THEN RETURN $$EXCEPTION_INTERNAL
	
	vectorR.size = vector1.size+vector2.size
	IF vector1.type = $$TYPE_COMPLEX || vector2.type = $$TYPE_COMPLEX THEN
		vectorR.type = $$TYPE_COMPLEX
	ELSE
		vectorR.type = $$TYPE_NUMBER
	END IF
	
	LinkedList_Init (@vectorR.values)
	
	hWalk = LinkedList_StartWalk (vector1.values)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		SYMBOL_Get(iData, @symbol)
		symbolR.type = symbol.type
		symbolR.delete = symbol.delete
		symbolR.data = CopySymbol (symbol.data, symbol.type)
		LinkedList_Append (@vectorR.values, SYMBOL_New(symbolR))
	LOOP
	LinkedList_EndWalk (hWalk)
	
	hWalk = LinkedList_StartWalk (vector2.values)
	DO WHILE LinkedList_Walk (hWalk, @iData)
		SYMBOL_Get(iData, @symbol)
		symbolR.type = symbol.type
		symbolR.delete = symbol.delete
		symbolR.data = CopySymbol (symbol.data, symbol.type)
		LinkedList_Append (@vectorR.values, SYMBOL_New(symbolR))
	LOOP
	LinkedList_EndWalk (hWalk)
	
	eval_push ($$TYPE_VECTOR, VECTOR_New(vectorR))
	
	RETURN $$EXCEPTION_NONE
END FUNCTION
'
' #####################
' #####  spaces$  #####
' #####################
'
'
'
FUNCTION spaces$ (x)
	FOR i = 1 TO x
		ret$ = ret$+" "
	NEXT
	RETURN ret$
END FUNCTION

DefineAccess(LINKEDLIST)
DefineAccess(TOKEN)
DefineAccess(VECTOR)
DefineAccess(FUNC)
DefineAccess(COMPLEX)
DefineAccess(SYMBOL)
DefineAccess(PARAMETER)
DefineAccess(FUNCCASE)
END PROGRAM
