'
' ####################
' #####  PROLOG  #####
' ####################
'
' An WinX GUI program
'
PROGRAM "graph"
VERSION	"0.0000"
CONSOLE
'
	IMPORT	"xst"   		' standard library	: required by most programs
	IMPORT	"xma"
	IMPORT  "xsx"				' extended std library
'	IMPORT	"xio"				' console io library
	IMPORT	"gdi32"     ' gdi32.dll
	IMPORT  "user32"    ' user32.dll
	IMPORT  "kernel32"  ' kernel32.dll
	IMPORT  "comctl32"	' comctl32.dll			: common controls library
'	IMPORT	"comdlg32"  ' comdlg32.dll	    : common dialog library
'	IMPORT	"xma"   		' math library			: Sin/Asin/Sinh/Asinh/Log/Exp/Sqrt...
'	IMPORT	"xcm"				' complex math library
'	IMPORT  "msvcrt"		' msvcrt.dll				: C function library
'	IMPORT  "shell32"   ' shell32.dll
	IMPORT  "opengl32"
	IMPORT  "glu32"

	IMPORT	"adt"
	IMPORT	"calculate"
	IMPORT	"gmp"
	IMPORT	"WinX"			' The Xwin GUI library
	
'
TYPE GRAPHSTATE
	' GUI variables
	XLONG	.hMain
	XLONG	.hDisplay
	XLONG	.hToolbar
	XLONG	.hRC
	XLONG	.viewWidth
	XLONG	.viewHeight
	XLONG	.hSetup
	XLONG	.hCourierFont
	XLONG	.lvWndProc
	
	' Graphing variables
	XLONG	.gmpf_result[2]
	XLONG	.axes
	XLONG	.initX
	XLONG	.initY
	XLONG	.initZ
	XLONG	.initT
	XLONG	.incX
	XLONG	.incY
	XLONG	.incZ
	XLONG	.incT
END TYPE
TYPE GRAPH
	XLONG	.type
	XLONG	.expr
	XLONG	.lineColour
	XLONG	.surfaceColour
	LINKEDLIST	.data
END TYPE
TYPE GRAPHPOINT	
	DOUBLE	.x
	DOUBLE	.y
	DOUBLE	.z
END TYPE

$$GRAPH_X	= 0x01
$$GRAPH_Y = 0x02
$$GRAPH_Z = 0x04
$$GRAPH_T = 0x08
$$TYPE_YX = 1
$$TYPE_XY = 2
$$TYPE_XYT = 3
$$TYPE_XYZ = 4
$$TYPE_YZX = 5
$$TYPE_ZXY = 6
$$TYPE_XYZT = 8
$$TYPE_XYZTS = 9

$$CMD_DRAW = 100
$$CMD_SAVE = 101
$$CMD_OPEN = 102
$$CMD_HELP = 103

$$BTN_ADD			= 200
$$BTN_REMOVE	= 201
$$LV_EXPRS		= 202
$$ED_X_FROM		= 203
$$ED_X_TO			= 204
$$ED_X_PITCH	= 205
$$ED_Y_FROM		= 206
$$ED_Y_TO			= 207
$$ED_Y_PITCH	= 208
$$ED_Z_FROM		= 209
$$ED_Z_TO			= 210
$$ED_Z_PITCH	= 211
$$ED_T_FROM		= 212
$$ED_T_TO			= 213
$$ED_T_PITCH	= 214

EXPORT
DECLARE FUNCTION Graph_Init ()
DECLARE FUNCTION Graph_ShowDialog ()
DECLARE FUNCTION Graph_KillDialog ()
END EXPORT

DECLARE FUNCTION initApp ()
DECLARE FUNCTION onLVLabelEdit (control, code, item, text$)
DECLARE FUNCTION onLVItem (idControl, event, parameter)
DECLARE FUNCTION onCommand (id, code, hWnd)
DECLARE FUNCTION onPaint (hWnd, hDC)
DECLARE FUNCTION controlSizer (hWnd, w, h)
DECLARE FUNCTION onClose (hWnd)
DECLARE FUNCTION  SetWindowPixelFormat (hWnd)
DECLARE FUNCTION  CreateRenderingContext (hWnd, @hRC)
DECLARE FUNCTION  DeleteRC (hRC)
DECLARE FUNCTION Listview (hWnd, msg, wParam, lParam)
DECLARE FUNCTION DOUBLE ComputePoint1D (iExpr)
'
'
' ###########################
' #####  Graph_Init ()  #####
' ###########################
'
FUNCTION Graph_Init ()
	SHARED GRAPHSTATE graphState
	
	IF !LIBRARY(0) THEN
		IF WinX() THEN QUIT(0)
		IF ADT() THEN QUIT(0)
		calculate_init ()
	END IF
	
	IF initApp () THEN RETURN $$TRUE
	
	IF !LIBRARY(0) THEN
		'quit if either of these fail
		IF Graph_ShowDialog () THEN QUIT(0)
		
		WinXDoEvents (0)
	END IF

END FUNCTION
'
' #####################
' #####  initApp  #####
' #####################
' 
'
'
FUNCTION initApp ()
	SHARED GRAPHSTATE graphState
	
	graphState.gmpf_result[0] = LocalAlloc ($$LMEM_FIXED, SIZE(mpf_t)) : __gmpf_init(graphState.gmpf_result[0])
	graphState.gmpf_result[1] = LocalAlloc ($$LMEM_FIXED, SIZE(mpf_t)) : __gmpf_init(graphState.gmpf_result[1])
	graphState.gmpf_result[2] = LocalAlloc ($$LMEM_FIXED, SIZE(mpf_t)) : __gmpf_init(graphState.gmpf_result[2])
	
	RETURN 0
END FUNCTION
'
' ##############################
' #####  Graph_ShowDialog  #####
' ##############################
'
'
'
FUNCTION Graph_ShowDialog ()
	SHARED GRAPHSTATE graphState
	LOGFONT lFont
	
	' Create the graphing window
	graphState.hMain = WinXNewWindow (0, "Graphing Window", -1, -1, 314, 400, $$XWSS_APP, 0, LoadIconA (GetModuleHandleA (0), 100), 0)
	WinXSetMinSize (graphState.hMain, 314, 400)
	WinXDrawImage (graphState.hMain, hBmp, 0, 48, 128, 32, 0, 0, $$FALSE)
	graphState.hToolbar = WinXNewToolbar (32, 32, 4, LoadBitmapA (GetModuleHandleA (0), 200), 0, 0, 0xFF00FF, $$TRUE, $$FALSE)
	WinXToolbar_AddButton (graphState.hToolbar, $$CMD_DRAW, 0, "Draw Graphs", $$FALSE, $$FALSE)
	WinXToolbar_AddButton (graphState.hToolbar, $$CMD_SAVE, 1, "Save Graphs", $$FALSE, $$FALSE)
	WinXToolbar_AddButton (graphState.hToolbar, $$CMD_OPEN, 2, "Open Graphs", $$FALSE, $$FALSE)
	WinXToolbar_AddButton (graphState.hToolbar, $$CMD_HELP, 3, "Graphing help", $$FALSE, $$FALSE)
	WinXSetWindowToolbar (graphState.hMain, graphState.hToolbar)
	WinXAddStatusBar (graphState.hMain, "", 0)

	' We're building the window upside down
	hMainSeries = WinXNewAutoSizerSeries ($$DIR_VERT|$$DIR_REVERSE) 
	WinXAutoSizer_SetSimpleInfo (hMainSeries, WinXAutoSizer_GetMainSeries (graphState.hMain), 0, 1, $$SIZER_SERIES)
	
	' create the setup panel
	graphState.hSetup = WinXNewChildWindow (graphState.hMain, "Setup Panel", $$WS_CLIPSIBLINGS, $$WS_EX_CLIENTEDGE, 0)
	WinXAutoSizer_SetSimpleInfo (graphState.hSetup, hMainSeries, 0, 152, $$SIZER_SPLITTER)
	WinXSplitter_SetProperties (hMainSeries, graphState.hSetup, 152, 300, $$TRUE)

	hH = WinXNewAutoSizerSeries ($$DIR_HORIZ)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, "x = ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 20, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "-10", $$ES_AUTOHSCROLL, $$ED_X_FROM), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, " to ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 0, 15, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "10", $$ES_AUTOHSCROLL, $$ED_X_TO), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, ", pitch ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 30, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "1", $$ES_AUTOHSCROLL, $$ED_X_PITCH), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (hH, WinXAutoSizer_GetMainSeries (graphState.hSetup), 2, 20, $$SIZER_SERIES)
	hH = WinXNewAutoSizerSeries ($$DIR_HORIZ)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, "y = ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 20, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "-10", $$ES_AUTOHSCROLL, $$ED_Y_FROM), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, " to ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 0, 15, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "10", $$ES_AUTOHSCROLL, $$ED_Y_TO), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, ", pitch ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 30, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "1", $$ES_AUTOHSCROLL, $$ED_Y_PITCH), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (hH, WinXAutoSizer_GetMainSeries (graphState.hSetup), 2, 20, $$SIZER_SERIES)
	hH = WinXNewAutoSizerSeries ($$DIR_HORIZ)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, "z = ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 20, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_Z_FROM), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, " to ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 0, 15, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_Z_TO), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, ", pitch ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 30, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_Z_PITCH), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (hH, WinXAutoSizer_GetMainSeries (graphState.hSetup), 2, 20, $$SIZER_SERIES)
	hH = WinXNewAutoSizerSeries ($$DIR_HORIZ)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, "t = ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 20, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_T_FROM), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, " to ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 0, 15, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_T_TO), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddStatic (graphState.hSetup, ", pitch ", 0, $$SS_LEFT|$$SS_CENTERIMAGE, 0), hH, 2, 30, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddEdit (graphState.hSetup, "", $$ES_AUTOHSCROLL, $$ED_T_PITCH), hH, 0, 80, 0)
	WinXAutoSizer_SetSimpleInfo (hH, WinXAutoSizer_GetMainSeries (graphState.hSetup), 2, 20, $$SIZER_SERIES)
	
	hFunctionsH = WinXNewAutoSizerSeries ($$DIR_HORIZ|$$DIR_REVERSE)
	hFunctionsV = WinXNewAutoSizerSeries ($$DIR_VERT)
	WinXAutoSizer_SetSimpleInfo (hFunctionsH, WinXAutoSizer_GetMainSeries (graphState.hSetup), 2, 1, $$SIZER_SERIES|$$SIZER_SIZERELREST)
	WinXAutoSizer_SetSimpleInfo (hFunctionsV, hFunctionsH, 0, 25, $$SIZER_SERIES)
	WinXAutoSizer_SetSimpleInfo (WinXAddButton (graphState.hSetup, "+", 0, $$BTN_ADD), hFunctionsV, 0, 25, 0)
	WinXAutoSizer_SetSimpleInfo (WinXAddButton (graphState.hSetup, "-", 0, $$BTN_REMOVE), hFunctionsV, 0, 25, 0)
	hLV = WinXAddListView (graphState.hSetup, 0, 0, $$TRUE, $$LVS_REPORT, $$LV_EXPRS)
	
	WinXSetStyle (hLV, $$WS_BORDER, 0, 0, 0)
	SendMessageA (hLV, $$LVM_SETEXTENDEDLISTVIEWSTYLE, $$LVS_EX_CHECKBOXES, $$LVS_EX_CHECKBOXES)
	
	lFont = WinXDraw_MakeLogFont ("Courier New", 16, 0)
	graphState.hCourierFont = CreateFontIndirectA (&lFont)
	WinXSetFont (hLV, graphState.hCourierFont)
	
	WinXListView_AddColumn (hLV, 0, 150, "Expressions", 0)
	
	graphState.lvWndProc = GetWindowLongA (hLV, $$GWL_WNDPROC)
	SetWindowLongA (hLV, $$GWL_WNDPROC, &Listview())
	
	WinXAutoSizer_SetSimpleInfo (hLV, hFunctionsH, 0, 1, $$SIZER_SIZERELREST)
	
	' Create the viewing pane
	graphState.hDisplay = WinXNewChildWindow (graphState.hMain, "Display", 0, $$WS_EX_CLIENTEDGE, 100)
	SetWindowPixelFormat (graphState.hDisplay)
	CreateRenderingContext (graphState.hDisplay, @hRC)
	graphState.hRC = hRC
	glClearColor (1.0, 1.0, 1.0, 1.0)
	glEnable ($$GL_BLEND)
'	glEnable ($$GL_LINE_SMOOTH)
	glBlendFunc($$GL_SRC_ALPHA, $$GL_ONE_MINUS_SRC_ALPHA)
	
	WinXAutoSizer_SetSimpleInfo (graphState.hDisplay, hMainSeries, 0, 1, $$SIZER_SIZERELREST)
	
	'remember to register callbacks
	WinXRegOnCommand (graphState.hMain, &onCommand())
	WinXRegOnCommand (graphState.hSetup, &onCommand())
	WinXRegOnLabelEdit (graphState.hSetup, &onLVLabelEdit())
	WinXRegOnItem (graphState.hSetup, &onLVItem ())
	WinXRegOnPaint (graphState.hDisplay, &onPaint())
	WinXRegControlSizer (graphState.hDisplay, &controlSizer())
	WinXRegOnClose (graphState.hMain, &onClose())
	
	WinXDisplay (graphState.hMain)
	
	RETURN 0
END FUNCTION
'
' ##############################
' #####  Graph_KillDialog  #####
' ##############################
' Destroys the dialog, deallocating all resources in the process
FUNCTION Graph_KillDialog ()
	SHARED GRAPHSTATE graphState
	
	__gmpf_clear(graphState.gmpf_result[0]) : LocalFree (graphState.gmpf_result[0])
	__gmpf_clear(graphState.gmpf_result[1]) : LocalFree (graphState.gmpf_result[1])
	__gmpf_clear(graphState.gmpf_result[2]) : LocalFree (graphState.gmpf_result[2])
	
	SetWindowLongA (GetDlgItem (graphState.hSetup, $$LV_EXPRS), $$GWL_WNDPROC, graphState.lvWndProc)

	DestroyWindow (graphState.hDisplay)
	DeleteRC (graphState.hRC)
	
	DestroyWindow (graphState.hMain)
	
	DeleteObject (graphState.hCourierFont)
	
	IF !LIBRARY(0) THEN PostQuitMessage (0)
END FUNCTION

'
' ###########################
' #####  onLVLabelEdit  #####
' ###########################
FUNCTION onCommand (id, code, hWnd)
	LVITEM lvi
	
	SHARED GRAPHSTATE graphState
	
	SELECT CASE id
		CASE $$CMD_DRAW
		CASE $$CMD_SAVE
		CASE $$CMD_OPEN
		CASE $$CMD_HELP
		CASE $$BTN_ADD
			hLV = GetDlgItem (graphState.hSetup, $$LV_EXPRS)
			SetFocus (hLV)
			WinXListView_AddItem (hLV, -1, "enter expression", -1)
			cItems = SendMessageA (hLV, $$LVM_GETITEMCOUNT, 0, 0)
			
			lvi.stateMask = $$LVIS_STATEIMAGEMASK
			lvi.state = 2<<12
			SendMessageA (hLV, $$LVM_SETITEMSTATE, cItems-1, &lvi)
			SendMessageA (hLV, $$LVM_EDITLABEL, cItems-1, 0)
		CASE $$BTN_REMOVE
			hLV = GetDlgItem (graphState.hSetup, $$LV_EXPRS)
			PRINT WinXListView_GetSelection (hLV, @iItems[])
			FOR i = UBOUND(iItems[]) TO 0 STEP -1
				WinXListView_DeleteItem (hLV, iItems[i])
			NEXT
	END SELECT
	
END FUNCTION

'
' ###########################
' #####  onLVLabelEdit  #####
' ###########################
FUNCTION onLVLabelEdit (control, code, item, text$)
	STATIC oldLabel$
	
	SELECT CASE code
		CASE $$EDIT_START
			oldLabel$ = text$
			RETURN $$TRUE
		CASE $$EDIT_DONE
			IF text$!=oldLabel$ THEN
				
			END IF
			
			RETURN $$TRUE
	END SELECT
END FUNCTION
'
' ######################
' #####  onLVItem  #####
' ######################
FUNCTION onLVItem (idControl, event, parameter)
	SHARED GRAPHSTATE graphState
	LVITEM lvi

	hLV = GetDlgItem (graphState.hSetup, $$LV_EXPRS)
	SELECT CASE event
		CASE $$NM_DBLCLK
			WinXGetMousePos (hLV, @x, @y)
			iItem = WinXListView_GetItemFromPoint (hLV, x, y)
			state = SendMessageA (hLV, $$LVM_GETITEMSTATE, iItem, $$LVIS_STATEIMAGEMASK)>>12
			lvi.stateMask = $$LVIS_STATEIMAGEMASK
			IF state = 1 THEN lvi.state = 2<<12 ELSE lvi.state = 1<<12
			SendMessageA (hLV, $$LVM_SETITEMSTATE, iItem, &lvi)
		CASE $$NM_RCLICK
	END SELECT
END FUNCTION

FUNCTION onPaint (hWnd, hDC)
	SHARED GRAPHSTATE graphState

	glClear ($$GL_COLOR_BUFFER_BIT)

	glBegin ($$GL_LINES)
		glColor4f (0.0, 0.0, 0.0, 1.0)
		glVertex2f (0, -graphState.viewHeight\2)
		glVertex2f (0, graphState.viewHeight\2)
		glVertex2f (-graphState.viewWidth\2, 0)
		glVertex2f (graphState.viewWidth\2, 0)
	glEnd ()

'	Code to draw functions will go here
	
	glFlush ()
END FUNCTION

FUNCTION controlSizer (hWnd, w, h)
	SHARED GRAPHSTATE graphState
	
	graphState.viewWidth = w
	graphState.viewHeight = h

	glViewport (0, 0, graphState.viewWidth, graphState.viewHeight)
	glMatrixMode ($$GL_PROJECTION)
	glLoadIdentity ()
	gluOrtho2D (0.0, graphState.viewWidth, 0.0, graphState.viewHeight)
	
	glMatrixMode ($$GL_MODELVIEW)
	glLoadIdentity ()
	glTranslated (graphState.viewWidth\2, graphState.viewHeight\2, 0)
END FUNCTION

FUNCTION onClose (hWnd)
	Graph_KillDialog ()
END FUNCTION
'
'
' #####################################
' #####  SetWindowPixelFormat ()  #####
' #####################################
'	Set the pixel format for OpenGL
FUNCTION  SetWindowPixelFormat (hWnd)

	PIXELFORMATDESCRIPTOR pixelDesc

	pixelDesc.nSize = SIZE(PIXELFORMATDESCRIPTOR)
	pixelDesc.nVersion = 1
	pixelDesc.dwFlags = $$PFD_DRAW_TO_WINDOW | $$PFD_DRAW_TO_BITMAP | $$PFD_SUPPORT_OPENGL | $$PFD_SUPPORT_GDI | $$PFD_STEREO_DONTCARE
	pixelDesc.iPixelType = $$PFD_TYPE_RGBA
	pixelDesc.cColorBits = 32
	pixelDesc.cRedBits = 8
	pixelDesc.cRedShift = 16
	pixelDesc.cGreenBits = 8
	pixelDesc.cGreenShift = 8
	pixelDesc.cBlueBits = 8
	pixelDesc.cBlueShift = 0
	pixelDesc.cAlphaBits = 0
	pixelDesc.cAlphaShift = 0
	pixelDesc.cAccumBits = 64
	pixelDesc.cAccumRedBits = 16
	pixelDesc.cAccumGreenBits = 16
	pixelDesc.cAccumBlueBits = 16
	pixelDesc.cAccumAlphaBits = 0
	pixelDesc.cDepthBits = 32
	pixelDesc.cStencilBits = 8
	pixelDesc.cAuxBuffers = 0
	pixelDesc.iLayerType = $$PFD_MAIN_PLANE
	pixelDesc.bReserved = 0
	pixelDesc.dwLayerMask = 0
	pixelDesc.dwVisibleMask = 0
	pixelDesc.dwDamageMask = 0

	hDC = GetDC (hWnd)
	pixelFormat = ChoosePixelFormat (hDC, &pixelDesc)

	IFZ pixelFormat THEN 					' choose a default index.
		pixelFormat = 1
		IFZ DescribePixelFormat (hDC, pixelFormat, SIZE(PIXELFORMATDESCRIPTOR), &pixelDesc) THEN
			ReleaseDC (hWnd, hDC)
			RETURN $$FALSE
		END IF
	END IF

	IFZ SetPixelFormat (hDC, pixelFormat, &pixelDesc) THEN
		ReleaseDC (hWnd, hDC)
		RETURN $$FALSE
	END IF

	ReleaseDC (hWnd, hDC)
	RETURN $$TRUE

END FUNCTION
'
' #######################################
' #####  CreateRenderingContext ()  #####
' #######################################
' Create the OpenGL rendering context and make it current
FUNCTION  CreateRenderingContext (hWnd, @hRC)

	hDC = GetDC (hWnd)
	hRC = wglCreateContext (hDC)
	IF hRC == NULL THEN
		ReleaseDC (hWnd, hDC)
		RETURN $$FALSE
	END IF

	IFZ wglMakeCurrent (hDC, hRC) THEN
		ReleaseDC (hWnd, hDC)
		RETURN $$FALSE
	END IF

	ReleaseDC (hWnd, hDC)
	RETURN $$TRUE

END FUNCTION
'
' #########################
' #####  DeleteRC ()  #####
' #########################
' Make the OpenGL rendering context not current and then delete it.
FUNCTION  DeleteRC (hRC)

	IF (wglGetCurrentContext () != NULL) THEN
		wglMakeCurrent (NULL, NULL)   ' make the rendering context not current
	END IF

	IFZ hRC THEN RETURN

	wglDeleteContext (hRC)					' delete current rendering context

END FUNCTION

FUNCTION Listview (hWnd, msg, wParam, lParam)
	SHARED GRAPHSTATE graphState
	
	SELECT CASE msg
		CASE $$WM_SIZE
			SendMessageA (hWnd, $$LVM_ENSUREVISIBLE, 0, 0)
			ret = CallWindowProcA (graphState.lvWndProc, hWnd, msg, wParam, lParam)
			SendMessageA (hWnd, $$LVM_SETCOLUMNWIDTH, 0, LOWORD(lParam))
			RETURN ret
	END SELECT
	
	RETURN CallWindowProcA (graphState.lvWndProc, hWnd, msg, wParam, lParam)
END FUNCTION
'
' ###############################
' #####  ComputePoint1D ()  #####
' ###############################
'	compute a 1D point from and expression
' iExpr = the expression to calculate from
' returns the result of the computation or $$NNAN on error
FUNCTION DOUBLE ComputePoint1D (iExpr)
	SHARED GRAPHSTATE graphState
	
'	calculate_execute (iExpr, graphState.gmpf_result, @exception, "")
'	IF exception = $$EXCEPTION_NONE THEN RETURN __gmpf_get_d(graphState.gmpf_result) ELSE RETURN $$NNAN
END FUNCTION

END PROGRAM